using System;
using System.Collections.Generic;
using System.DirectoryServices.Protocols;
using System.Threading;
using InfiniTec.DirectoryServices;
using InfiniTec.Threading;

namespace InfiniTec.Exchange.ActiveDirectory
{
    internal class ActiveDirectoryResolver : IActiveDirectoryResolver
    {
        private Item _ConfigurationNamingContext;
        private object _ConfigurationNamingContextLock = new object();
        private static string _DefaultSmtpDomain;
        private static object _DefaultSmtpDomainLock = new object();
        private Connection _Connection;

        public ActiveDirectoryResolver(Connection connection)
        {
            _Connection = connection;
        }

        public AsyncOperation<string> GetMailboxResolveOperation(ActiveDirectoryUser user)
        {
            return new AsyncOperation<string>(ResolveMailboxInternal(user));
        }

        private IEnumerable<OperationAction> ResolveMailboxInternal(ActiveDirectoryUser user)
        {
            Item homeMDB;
            IList<Item> httpVirtualServer;
            Item exServer;
            Searcher searcher;
            Item defaultSmtpVirtualServer;

            string exServerDn;
            string alias;
            string domain;
            string uri;
            string[] serverBinding;
            

            if (user == null) throw new ArgumentNullException("user", "User must not be null");

            if (UserItemNeedsUpdate(user))
            {
                yield return new NestedOperation(user.GetRefreshOperation(), false);
            }

            homeMDB = new Item(user.Properties.GetProperty<string>(WellknownProperties.User.HomeMDB).Value,
                               _Connection.ActiveDirectoryConnection);
            homeMDB.PropertiesToLoad.Add(WellknownProperties.User.OwningServer);

            using (AsyncOperation operation = homeMDB.GetRefreshOperation())
            {
                yield return new NestedOperation(operation, false, typeof (DirectoryOperationException));
                if (operation.Status == OperationStatus.Error)
                {
                    DirectoryOperationException ex;

                    ex = (DirectoryOperationException) operation.Exception;
                    
                    if (ex.Response.ResultCode == ResultCode.NoSuchObject)
                    {
                        throw new MailboxUrlResolutionFailedException(Strings.MailboxUrlResolutionFailedException, ex);
                    }
                    else throw ex;
                }
            }

            exServerDn = homeMDB.Properties.GetProperty<string>(WellknownProperties.User.OwningServer).Value;
            exServer = new Item(exServerDn, _Connection.ActiveDirectoryConnection);
            exServer.PropertiesToLoad.Add(WellknownProperties.ExchangeServer.NetworkAddress);
            exServer.PropertiesToLoad.Add(WellknownProperties.ExchangeServer.VersionNumber);

            yield return new NestedOperation(exServer.GetRefreshOperation(), false);

            searcher =
                new Searcher(string.Concat("CN=http, CN=Protocols, ", exServerDn), _Connection.ActiveDirectoryConnection,
                             WellknownProperties.User.ServerBindings,
                             WellknownProperties.User.DefaultDomain);

            searcher.Constraints = "(objectClass=protocolCfgHttpServer)";
            searcher.Scope = System.DirectoryServices.Protocols.SearchScope.OneLevel;

            yield return new NestedOperation(searcher.GetFindOperation(SearchMode.FindAll, false), false);

            httpVirtualServer = searcher.SearchResult;

            defaultSmtpVirtualServer = GetDefaultHttpVirtualServer(httpVirtualServer);

            if (defaultSmtpVirtualServer == null)
                throw new DefaultSmtpServerNotFoundException("Unable to find default http virtual server for user " +
                                                             user.NT4AccountName);

            if (_DefaultSmtpDomain == null)
            {
                lock (_DefaultSmtpDomainLock)
                {
                    if (_DefaultSmtpDomain == null)
                    {
                        yield return new NestedOperation(GetDefaultSmtpDomain(), false);
                    }
                }
            }

            uri = "";
            // Iterate through the proxy-address of the user to find the SMTP addresses
            foreach (
                string proxyAddress in
                    user.Properties.GetMultivaluedProperty<string>(WellknownProperties.User.ProxyAddresses).Values)
            {
                // Check wether the found address is a SMTP address and if the 
                // domain-part equals the defaultSmtpDomain.

                if (!proxyAddress.StartsWith("smtp:", StringComparison.OrdinalIgnoreCase) ||
                    (proxyAddress.IndexOf(_DefaultSmtpDomain, StringComparison.OrdinalIgnoreCase) == -1))
                    continue;

                alias = proxyAddress.Substring(5, proxyAddress.IndexOf("@") - 5);
                serverBinding = defaultSmtpVirtualServer.Properties.GetMultivaluedProperty<string>(
                    WellknownProperties.User.ServerBindings).Values[0].Split(':');

                uri = BuildUriFromServerBinding(alias, exServer, serverBinding);
            }
            if (uri == "")
            {
                //None of the smtp-addresses matched the defaultSmtp Domain so check the
                //other http virtual server if one is responsible for the domain of the mail address.

                foreach (
                    string proxyAddress in
                        user.Properties.GetMultivaluedProperty<string>(WellknownProperties.User.ProxyAddresses).Values)
                {
                    // Check wether the found address is a SMTP address and if the 
                    // domain-part equals the defaultSmtpDomain.

                    if (!proxyAddress.StartsWith("smtp:", StringComparison.OrdinalIgnoreCase))
                        continue;

                    alias = proxyAddress.Substring(5, proxyAddress.IndexOf("@") - 5);
                    domain = proxyAddress.Substring(proxyAddress.IndexOf("@") + 1);

                    foreach (Item httpServer in httpVirtualServer)
                    {
                        // If the virtual http server directory entry does not contain
                        // the property msExchDefaultDomain, it is the default smtp host.
                        // That node has been handled before, so ignore it here
                        if (!httpServer.Properties.Contains(WellknownProperties.User.DefaultDomain))
                        {
                            continue;
                        }
                        // Check if the value of the property msExchDefaultDomain equals
                        // the desired domain.
                        if (httpServer.Properties.GetProperty<string>(WellknownProperties.User.DefaultDomain).Value !=
                            domain)
                        {
                            continue;
                        }

                        serverBinding =
                            httpServer.Properties.GetProperty<string>(WellknownProperties.User.DefaultDomain).Value.
                                Split(':');

                        uri = BuildUriFromServerBinding(alias, exServer, serverBinding);
                    }
                }
            }
            yield return new OperationResult<string>(uri);
        }

        private static string BuildUriFromServerBinding(string alias, Item exServer, string[] serverBinding)
        {
            string uri;
            if (serverBinding[2] != "")
                uri = string.Format("http://{0}:{1}/exchange/{2}", serverBinding[2], serverBinding[1], alias);
            else if (serverBinding[0] != "")
                uri = string.Format("http://{0}:{1}/exchange/{2}", serverBinding[0], serverBinding[1], alias);
            else
            {
                uri =
                    string.Format("http://{0}:{1}/exchange/{2}", GetExchangeServerTcpIpAddress(exServer),
                                  serverBinding[1], alias);
            }
            return uri;
        }

        private static bool UserItemNeedsUpdate(ActiveDirectoryUser user)
        {
            bool result = false;

            if (!user.Properties.Contains(WellknownProperties.User.HomeMDB))
            {
                user.PropertiesToLoad.Add(WellknownProperties.User.HomeMDB);
                result = true;
            }

            if (!user.Properties.Contains(WellknownProperties.User.ProxyAddresses))
            {
                user.PropertiesToLoad.Add(WellknownProperties.User.ProxyAddresses);
                result = true;
            }

            return result;
        }

        /// <summary>
        /// Searches the specified httpServer collection for the default http server.
        /// </summary>
        /// <param name="searchResult">A list of virtual httpServer </param>
        /// <returns>The <see cref="Item"/> with the default httpServer.</returns>
        private static Item GetDefaultHttpVirtualServer(IList<Item> searchResult)
        {
            int i;
            i = 0;

            while ((i < searchResult.Count) && (searchResult[i].Properties.Contains("msExchDefaultDomain"))) i++;

            return (i < searchResult.Count) ? searchResult[i] : null;
        }

        /// <summary>
        /// Returns the TCP/IP address for the specified server object
        /// </summary>
        /// <param name="serverObject"></param>
        /// <returns></returns>
        private static string GetExchangeServerTcpIpAddress(Item serverObject)
        {
            foreach (
                string address in
                    serverObject.Properties.GetMultivaluedProperty<string>(
                        WellknownProperties.ExchangeServer.NetworkAddress).Values)
            {
                if (address.StartsWith("ncacn_ip_tcp:", StringComparison.OrdinalIgnoreCase))
                {
                    return address.Substring(13);
                }
            }
            throw new ServerTcpIpAddressNotFoundException(
                string.Format(Strings.ServerTcpIpAddressNotFoundException, serverObject.DistinguishedName));
        }

        private AsyncOperation GetDefaultSmtpDomain()
        {
            return new AsyncOperation(GetDefaultSmtpDomainInternal());
        }

        /// <summary>
        /// Resolves the default SMTP for the Exchange organization organization.
        /// </summary>
        private IEnumerable<OperationAction> GetDefaultSmtpDomainInternal()
        {
            Item item;
            int index;

            using (AsyncOperation<Item> op = GetOrganizationItem())
            {
                yield return new NestedOperation(op, false);
                item = op.StatusInformation;
            }

            item = new Item("CN=Default Policy,CN=Recipient Policies, " +
                            item.DistinguishedName, _Connection.ActiveDirectoryConnection);
            item.PropertiesToLoad.Add(WellknownProperties.RecipientPolicy.GatewayProxy);

            yield return new NestedOperation(item.GetRefreshOperation(), false);

            foreach (
                string gateway in
                    item.Properties.GetMultivaluedProperty<string>(WellknownProperties.RecipientPolicy.GatewayProxy).
                        Values)
            {
                index = gateway.IndexOf("@");
                if (!gateway.StartsWith("SMTP:", StringComparison.Ordinal) || index == -1)
                    continue;

                _DefaultSmtpDomain = gateway.Substring(index);

                yield break;
            }

            throw new DefaultSmtpDomainNotFoundException(Strings.DefaultSmtpDomainNotFoundException);
        }

        private AsyncOperation<Item> GetOrganizationItem(params string[] propertiesToLoad)
        {
            return new AsyncOperation<Item>(GetOrganizationInternal(propertiesToLoad));
        }

        private IEnumerable<OperationAction> GetOrganizationInternal(params string[] propertiesToLoad)
        {
            Searcher searcher;

            if (_ConfigurationNamingContext == null)
            {
                lock (_ConfigurationNamingContextLock)
                {
                    if (_ConfigurationNamingContext == null)
                    {
                        using (
                            AsyncOperation<int, Item> operation =
                                ActiveDirectoryEntry.GetConfigurationNamingContext(_Connection.ActiveDirectoryConnection)
                            )
                        {
                            yield return new NestedOperation(operation, false);
                            _ConfigurationNamingContext = operation.Result;
                        }
                    }
                }
            }
            searcher = new Searcher("CN=Microsoft Exchange, CN=Services, " +
                                    _ConfigurationNamingContext.DistinguishedName,
                                    _Connection.ActiveDirectoryConnection, propertiesToLoad);

            searcher.Scope = System.DirectoryServices.Protocols.SearchScope.OneLevel;
            searcher.Constraints = "(objectClass=msExchOrganizationContainer)";
            searcher.PropertiesToLoad.Add(WellknownProperties.Common.DistinguishedName);

            yield return new NestedOperation(searcher.GetFindOperation(SearchMode.FindAll, false), false);

            if (searcher.SearchResult.Count == 0)
                throw new ExchangeOrganizationNotfoundException(Strings.NoExchangeOrganizationFoundException);

            yield return new OperationResult<Item>(searcher.SearchResult[0]);
        }

        public AsyncOperation<AddressListContainer> GetFindAddressListOperation()
        {
            return new AsyncOperation<AddressListContainer>(GetAddressListsInternal());
        }

        private IEnumerable<OperationAction> GetAddressListsInternal()
        {
            string organizationDN;
            AddressListContainer result;

            using (AsyncOperation<Item> op = GetOrganizationItem())
            {
                yield return new NestedOperation(op, false);
                organizationDN = op.StatusInformation.DistinguishedName;
            }

            result = new AddressListContainer();

            using (
                AsyncOperation<AddressList[]> op =
                    GetAddressLists("CN=All Global Address Lists, CN=Address Lists Container, " + organizationDN))
            {
                yield return new NestedOperation(op, false);
                result.GlobalAddressLists = op.StatusInformation;
            }

            using (
                AsyncOperation<AddressList[]> op =
                    GetAddressLists("CN=All Address Lists, CN=Address Lists Container, " + organizationDN))
            {
                yield return new NestedOperation(op, false);
                result.AllAddressLists = op.StatusInformation;
            }

            yield return new OperationResult<AddressListContainer>(result);
        }

        private AsyncOperation<AddressList[]> GetAddressLists(string searchRoot)
        {
            return new AsyncOperation<AddressList[]>(GetAddressListsInternal(searchRoot));
        }

        private IEnumerable<OperationAction> GetAddressListsInternal(string searchRoot)
        {
            Searcher searcher;
            List<AddressList> result;

            searcher = new Searcher(searchRoot, _Connection.ActiveDirectoryConnection, WellknownProperties.Common.Cn);

            searcher.Scope = System.DirectoryServices.Protocols.SearchScope.OneLevel;

            yield return new NestedOperation(searcher.GetFindOperation(SearchMode.FindAll, false), false);

            result = new List<AddressList>();

            foreach (Item item in searcher.SearchResult)
            {
                result.Add(
                    new AddressList(
                        item.Properties.GetProperty<string>(WellknownProperties.Common.Cn).Value,
                        item.DistinguishedName, _Connection));
            }

            yield return new OperationResult<AddressList[]>(result.ToArray());
        }
    }
}