using System.Collections.Generic;
using System.ComponentModel;
using InfiniTec.DirectoryServices;
using InfiniTec.Threading;
using AsyncOperation=InfiniTec.Threading.AsyncOperation;

namespace InfiniTec.Exchange
{
    /// <summary>
    /// This class provides access to the members of an Exchange address list.
    /// </summary>
    public class AddressList
    {
        /// <summary>
        /// This event is fired when an asynchronous search for the members of the address
        /// list finishes.
        /// </summary>
        public event AsyncCompletedEventHandler FindMembersCompleted;

        private string _Name;
        private string _DistinguishedName;
        private IList<IAddressListEntry> _Members;
        private Connection _Connection;
        private SearchMode _SearchMode;
        private SearchToken _SearchToken;
        private AsyncOperationContext _AsyncOperationContext;

        /// <summary>
        /// Gets or sets the <see cref="SearchMode"/>.
        /// </summary>
        /// <value>The search mode.</value>
        public SearchMode SearchMode
        {
            get { return _SearchMode; }
            set { _SearchMode = value; }
        }

        /// <summary>
        /// Gets the name.
        /// </summary>
        /// <value>The name.</value>
        public string Name
        {
            get { return _Name; }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="AddressList"/> class.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <param name="distinguishedName">Name of the distinguished.</param>
        /// <param name="connection">The connection.</param>
        internal AddressList(string name, string distinguishedName, Connection connection)
        {
            _Name = name;
            _DistinguishedName = distinguishedName;
            _Connection = connection;
            _AsyncOperationContext = new AsyncOperationContext();
        }


        /// <summary>
        /// Gets the members of this address list
        /// </summary>
        /// <value>The members of this address list.</value>
        public IList<IAddressListEntry> Members
        {
            get
            {
                if (_Members == null) throw new NotInitializedException(Strings.InstanceNotInitializedException);

                return _Members;
            }
        }

        /// <summary>
        /// Synchronously finds the members of this address list.
        /// </summary>
        public void FindMembers()
        {
            using (AsyncOperation op = GetFindMembersOperation())
            {
                op.Run();
                op.WaitForCompletion();
            }
        }

        /// <summary>
        /// Asynchronously finds the members of this address list.
        /// </summary>
        public void FindMembersAsync()
        {
            AsyncOperation op;

            _AsyncOperationContext.InitiatePendingAsyncOperation();

            op = GetFindMembersOperation();
            _AsyncOperationContext.RegisterPendingAsyncOperation(op);

            op.Completed += AsyncFind_Completed;
            op.Error += AsyncFind_Error;

            op.Run();
        }

        private void AsyncFind_Error(object sender, ExceptionEventArgs e)
        {
            _AsyncOperationContext.UnregisterPendingAsyncOperation();

            OnFindMembersCompleted(new AsyncCompletedEventArgs(e.Exception, false, null));
        }

        /// <summary>
        /// Raises the <see cref="FindMembersCompleted"/> event.
        /// </summary>
        /// <param name="e">The <see cref="System.ComponentModel.AsyncCompletedEventArgs"/> instance containing the event data.</param>
        protected virtual void OnFindMembersCompleted(AsyncCompletedEventArgs e)
        {
            if (FindMembersCompleted != null) FindMembersCompleted(this, e);
        }

        private void AsyncFind_Completed(object sender, CompletedEventArgs e)
        {
            AsyncOperation op;

            _AsyncOperationContext.UnregisterPendingAsyncOperation();
            op = (AsyncOperation) sender;
            OnFindMembersCompleted(new AsyncCompletedEventArgs(null,
                                                        op.Status != OperationStatus.Completed, null));
        }

        /// <summary>
        /// Gets the <see cref="AsyncOperation"/> which can be used to find
        /// the members of this address list.
        /// </summary>
        /// <returns></returns>
        public AsyncOperation GetFindMembersOperation()
        {
            return new AsyncOperation(FindInternal());
        }

        private IEnumerable<OperationAction> FindInternal()
        {
            PrincipalResolver resolver;
            List<IAddressListEntry> result;

            if (_SearchMode == SearchMode.FindPage && _SearchToken != null)
            {
                resolver = new PrincipalResolver(_SearchToken, _Connection.ActiveDirectoryConnection);
            }
            else
            {
                resolver = new PrincipalResolver("", _Connection.ActiveDirectoryConnection);
                resolver.Filter = string.Format("{0}={1}", ActiveDirectory.WellknownProperties.User.ShowInAddressBook,
                                                _DistinguishedName);
                resolver.FilterType = FilterType.SimpleLdapFilter;
                resolver.ResolveTypes = ResolveTypes.Groups | ResolveTypes.User | ResolveTypes.Contacts;
                resolver.PropertiesToLoad.Add(ActiveDirectory.WellknownProperties.User.HomeMDB);
                resolver.PropertiesToLoad.Add(ActiveDirectory.WellknownProperties.User.ProxyAddresses);
                resolver.SearchMode = _SearchMode;
            }

            using (AsyncOperation op = resolver.GetFindOperation())
            {
                yield return new NestedOperation(op, false);
            }

            if (SearchMode == SearchMode.FindPage)
            {
                _SearchToken = resolver.SearchToken;
            }

            result = new List<IAddressListEntry>();

            foreach (ActiveDirectoryPrincipal entry in resolver.SearchResult)
            {
                if (entry is ActiveDirectoryUser)
                {
                    result.Add(new User((ActiveDirectoryUser) entry, _Connection));
                }
                else if (entry is ActiveDirectoryContact)
                {
                    result.Add(new Contact((ActiveDirectoryContact) entry));
                }
            }

            _Members = result.AsReadOnly();
        }
    }
}