using System.Collections.Generic;
using System.ComponentModel;
using InfiniTec.Exchange.ActiveDirectory;
using InfiniTec.Threading;
using AsyncOperation=InfiniTec.Threading.AsyncOperation;

namespace InfiniTec.Exchange
{
    /// <summary>
    /// Provides access to the Exchange address lists.
    /// </summary>
    public class AdressListManager
    {
        /// <summary>
        /// Fired, when when an asynchronous refresh operation completes.
        /// </summary>
        public event AsyncCompletedEventHandler RefreshCompleted;

        private Connection _Connection;
        private Dictionary<string, AddressList> _GlobalAddressLists;
        private Dictionary<string, AddressList> _AllAddressLists;
        private AsyncOperationContext _AsyncOperationContext;

        /// <summary>
        /// Initializes a new instance of the <see cref="AdressListManager"/> class.
        /// </summary>
        /// <param name="connection">The connection.</param>
        public AdressListManager(Connection connection)
        {
            _Connection = connection;
            _AsyncOperationContext = new AsyncOperationContext();
        }

        /// <summary>
        /// Gets the global address lists.
        /// </summary>
        /// <value>The global address lists.</value>
        public IDictionary<string, AddressList> GlobalAddressLists
        {
            get
            {
                if (_GlobalAddressLists == null)
                    throw new NotInitializedException(Strings.InstanceNotInitializedException);

                return _GlobalAddressLists;
            }
        }

        /// <summary>
        /// Gets all other address lists.
        /// </summary>
        /// <value>All address lists.</value>
        public IDictionary<string, AddressList> AllAddressLists
        {
            get
            {
                if (_AllAddressLists == null)
                    throw new NotInitializedException(Strings.InstanceNotInitializedException);

                return _AllAddressLists;
            }
        }

        /// <summary>
        /// Refreshes this instance synchronously.
        /// </summary>
        public void Refresh()
        {
            using (AsyncOperation op = GetRefreshOperation())
            {
                op.Run();
                op.WaitForCompletion();
            }
        }

        /// <summary>
        /// Gets the asynchrnonous refresh operation for this instance.
        /// </summary>
        /// <returns></returns>
        public AsyncOperation GetRefreshOperation()
        {
            return new AsyncOperation(RefreshInternal());
        }

        private IEnumerable<OperationAction> RefreshInternal()
        {
            AddressListContainer result;

            using (
                AsyncOperation<AddressListContainer> op =
                    _Connection.ActiveDirectoryResolver.GetFindAddressListOperation())
            {
                yield return new NestedOperation(op, false);
                result = op.StatusInformation;
            }

            _GlobalAddressLists = new Dictionary<string, AddressList>();
            foreach (AddressList list in result.GlobalAddressLists)
            {
                _GlobalAddressLists.Add(list.Name, list);
            }


            _AllAddressLists = new Dictionary<string, AddressList>();
            foreach (AddressList list in result.AllAddressLists)
            {
                _AllAddressLists.Add(list.Name, list);
            }
        }

        /// <summary>
        /// Asynchronously refreshes this instance.
        /// </summary>
        public void RefreshAsync()
        {
            AsyncOperation op;

            _AsyncOperationContext.InitiatePendingAsyncOperation();

            op = GetRefreshOperation();
            op.Completed += Refresh_Completed;

            _AsyncOperationContext.RegisterPendingAsyncOperation(op);
        }

        private void Refresh_Completed(object sender, CompletedEventArgs e)
        {
            AsyncOperation op;

            _AsyncOperationContext.UnregisterPendingAsyncOperation();

            op = (AsyncOperation) sender;

            OnRefreshCompleted(new AsyncCompletedEventArgs(null, op.Status != OperationStatus.Completed, null));
        }

        /// <summary>
        /// Raises the <see cref="RefreshCompleted"/> event.
        /// </summary>
        /// <param name="e">The <see cref="System.ComponentModel.AsyncCompletedEventArgs"/> instance containing the event data.</param>
        protected virtual void OnRefreshCompleted(AsyncCompletedEventArgs e)
        {
            if (RefreshCompleted != null) RefreshCompleted(this, e);
        }
    }
}