using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.DirectoryServices.Protocols;
using System.Net;
using System.Runtime.CompilerServices;
using InfiniTec.DirectoryServices;
using InfiniTec.Exchange.ActiveDirectory;
using InfiniTec.Exchange.WebDav.Request;

namespace InfiniTec.Exchange
{
    /// <summary>
    /// This class contains information used to access the Exchange store.
    /// </summary>
    public class Connection
    {
        private static Connection _DefaultConnection = new Connection();
        private CookieCollection _FBACache;
        private string _ExchangeServer;
        private string _ActiveDirectoryServer;
        private ConnectionProtocol _ConnectionProtocol;
        private Credential _Credential;
        private List<IPAddress> _DnsEntries;
        private DirectoryServices.Connection _ActiveDirectoryConnection;
        private bool _UseGlobalCatalog;
        private IActiveDirectoryResolver _ActiveDirectoryResolver;

        /// <summary>
        /// Gets the default connection.
        /// </summary>
        /// <value>The default connection.</value>
        public static Connection DefaultConnection
        {
            get { return _DefaultConnection; }
        }

        internal IActiveDirectoryResolver ActiveDirectoryResolver
        {
            [MethodImpl(MethodImplOptions.Synchronized)]
            get
            {
                if (_ActiveDirectoryResolver == null) _ActiveDirectoryResolver = new ActiveDirectoryResolver(this);
                return _ActiveDirectoryResolver;
            }
        }

        internal CookieCollection FBACache
        {
            get { return _FBACache; }
            set { _FBACache = value; }
        }

        /// <summary>
        /// Gets or sets a value indicating whether to use global catalog for directory access.
        /// </summary>
        /// <value><c>true</c> if a global catalog should be used; otherwise, <c>false</c>.</value>
        public bool UseGlobalCatalog
        {
            get { return _UseGlobalCatalog; }
            set
            {
                if (_ActiveDirectoryConnection != null)
                    throw new InvalidOperationException(
                        "The Active Directory settings cannot be changed after the connection has been initialized.");

                _UseGlobalCatalog = value;
            }
        }

        /// <summary>
        /// Gets the active directory connection.
        /// </summary>
        /// <value>The active directory connection.</value>
        public DirectoryServices.Connection ActiveDirectoryConnection
        {
            get
            {
                EnsureActiveDirectoryConnection();
                return _ActiveDirectoryConnection;
            }
        }

        [MethodImpl(MethodImplOptions.Synchronized)]
        private void EnsureActiveDirectoryConnection()
        {
            if (_ActiveDirectoryConnection != null) return;

            _ActiveDirectoryConnection =
                new DirectoryServices.Connection(ActiveDirectoryServer, DirectoryIdentifierType.Server, false);

            _ActiveDirectoryConnection.DomainController = _ActiveDirectoryServer;

            ConfigureCredential();

            if (UseGlobalCatalog)
            {
                _ActiveDirectoryConnection.SessionOptions.LocatorFlag |= LocatorFlags.GCRequired;
            }
        }

        private void ConfigureCredential()
        {
            string username;
            string domain;
            int index;

            if (string.IsNullOrEmpty(Credential.Username))
            {
                _ActiveDirectoryConnection.Credential = CredentialCache.DefaultNetworkCredentials;
                _ActiveDirectoryConnection.AuthenticationType = AuthType.Negotiate;
                return;
            }

            index = Credential.Username.IndexOf('\\');

            if (index != -1)
            {
                domain = Credential.Username.Substring(0, index);
                username = Credential.Username.Substring(index + 1);
                _ActiveDirectoryConnection.Credential = new NetworkCredential(username, Credential.Password, domain);
            }
            else
            {
                _ActiveDirectoryConnection.Credential = new NetworkCredential(Credential.Username, Credential.Password);
            }

            switch (Credential.AuthenticationType)
            {
                case AuthenticationType.Basic:
                    _ActiveDirectoryConnection.AuthenticationType = AuthType.Basic;
                    break;
                case AuthenticationType.Ntlm:
                    _ActiveDirectoryConnection.AuthenticationType = AuthType.Sicily;
                    break;
                case AuthenticationType.Kerberos:
                    _ActiveDirectoryConnection.AuthenticationType = AuthType.Kerberos;
                    break;
                case AuthenticationType.Digest:
                    _ActiveDirectoryConnection.AuthenticationType = AuthType.Digest;
                    break;
                default:
                    _ActiveDirectoryConnection.AuthenticationType = AuthType.Basic;
                    break;
            }
        }

        private Uri BaseUrl
        {
            get
            {
                string prefix;

                switch (ConnectionProtocol)
                {
                    case ConnectionProtocol.Default:
                        prefix = "http";
                        break;
                    case ConnectionProtocol.Secure:
                        prefix = "https";
                        break;
                    default:
                        throw new NotSupportedException("The specified protocol type (" + ConnectionProtocol +
                                                        ") is not supported.");
                }
                return new Uri(string.Format("{0}://{1}", prefix, ExchangeServer));
            }
        }

        /// <summary>
        /// Gets or sets the DNS name of the active directory server to use. If left,
        /// a server will bee automatically selected. 
        /// </summary>
        /// <remarks>To use serverless binding, the program must be run from a computer within a forest.</remarks>
        /// <value>The active directory server.</value>
        public string ActiveDirectoryServer
        {
            [DebuggerStepThrough]
            get { return _ActiveDirectoryServer; }
            [DebuggerStepThrough]
            set
            {
                if (_ActiveDirectoryConnection != null)
                    throw new InvalidOperationException(
                        "The Active Directory settings cannot be changed after the connection has been initialized.");

                _ActiveDirectoryServer = value;
            }
        }

        /// <summary>
        /// Gets or sets the DNS name of the Exchange server.
        /// </summary>
        public string ExchangeServer
        {
            get { return _ExchangeServer; }
            set
            {
                _DnsEntries = GetIPAddresses(value);
                _ExchangeServer = value;
            }
        }

        private static List<IPAddress> GetIPAddresses(string value)
        {
            IPAddress address;
            List<IPAddress> result;

            result = new List<IPAddress>();

            if (IPAddress.TryParse(value, out address))
            {
                result.Add(address);
            }
            else
            {
                result.AddRange(Dns.GetHostAddresses(value));
            }

            return result;
        }

        /// <summary>
        /// Gets or sets the <see cref="ConnectionProtocol"/> to use
        /// for requests.
        /// </summary>
        public ConnectionProtocol ConnectionProtocol
        {
            get { return _ConnectionProtocol; }
            set { _ConnectionProtocol = value; }
        }

        /// <summary>
        /// Gets or sets the <see cref="Credential"/> used for requests.
        /// </summary>
        public Credential Credential
        {
            get { return _Credential; }
            set { _Credential = value; }
        }

        /// <summary>
        /// Creates a new <see cref="RequestOperation"/> configured with
        /// the specified <see cref="RequestSettings"/>.
        /// </summary>
        /// <param name="requestInfo">The settings for the request.</param>
        /// <returns>A <see cref="RequestOperation"/>.</returns>
        public RequestOperation CreateRequest(RequestSettings requestInfo)
        {
            return Request.Create(requestInfo, this);
        }

        /// <summary>
        /// Creates a full qualified url based on the current connection 
        /// settings and the provided relative path.
        /// </summary>
        /// <param name="path">The relative url on the Exchange server.</param>
        /// <returns>An <see cref="Uri"/> for the requested path.</returns>
        public Uri CreateUrl(string path)
        {
            if (path.StartsWith("/")) path = path.Substring(1);

            return new Uri(BaseUrl, path);
        }

        /// <summary>
        /// Checks the specified url it matches the specified <see cref="ConnectionProtocol"/> and
        /// server.
        /// </summary>
        /// <param name="url">The url to check.</param>
        /// <returns>True, if the url can be handled by the current <see cref="Connection"/> instance. False, otherwise.</returns>
        public bool IsValidUrl(Uri url)
        {
            List<IPAddress> addresses;

            if (string.IsNullOrEmpty(ExchangeServer)) return true;

            addresses = GetIPAddresses(url.Host);

            foreach (IPAddress serverAddress in _DnsEntries)
            {
                foreach (IPAddress urlAddress in addresses)
                {
                    if (serverAddress.Equals(urlAddress)) return true;
                }
            }

            return false;
        }

        /// <summary>
        /// Creates a new instance of this class using the specified settings.
        /// </summary>
        /// <param name="server">The DNS name or IP-Address of the Exchange server.</param>
        /// <param name="protocol">The <see cref="ConnectionProtocol"/> to use.</param>
        /// <param name="credential">The <see cref="Credential"/> to use.</param>
        public Connection(string server, ConnectionProtocol protocol, Credential credential)
        {
            ExchangeServer = server;
            ConnectionProtocol = protocol;
            Credential = credential;
        }


        /// <summary>
        /// Initializes a new instance of the <see cref="Connection"/> class.
        /// </summary>
        /// <param name="connectionProtocol">The <see cref="ConnectionProtocol"/> to use.</param>
        /// <param name="credential">The <see cref="Credential"/> to use.</param>
        public Connection(ConnectionProtocol connectionProtocol, Credential credential)
        {
            _ConnectionProtocol = connectionProtocol;
            _Credential = credential;
        }

        /// <summary>
        /// Creates a new instance of this class.
        /// </summary>
        public Connection()
        {
            Credential = new Credential();
        }
    }
}