using System;
using System.Diagnostics;
using System.Text.RegularExpressions;

namespace InfiniTec.Exchange
{
    /// <summary>
    /// This class defines an email address, along with a display name and the 
    /// type of the address.
    /// </summary>
    public class EmailAddress
    {
        private static Regex _EmailAddressValidator =
            new Regex(
                @"^((\""[^\""\f\n\r\t\v\b]+\"")|([\w\!\#\$\%\&\'\*\+\-\~\/\^\`\|\{\}]+(\.[\w\!\#\$\%\&\'\*\+\-\~\/\^\`\|\{\}]+)*))@((\[(((25[0-5])|(2[0-4][0-9])|([0-1]?[0-9]?[0-9]))\.((25[0-5])|(2[0-4][0-9])|([0-1]?[0-9]?[0-9]))\.((25[0-5])|(2[0-4][0-9])|([0-1]?[0-9]?[0-9]))\.((25[0-5])|(2[0-4][0-9])|([0-1]?[0-9]?[0-9])))\])|(((25[0-5])|(2[0-4][0-9])|([0-1]?[0-9]?[0-9]))\.((25[0-5])|(2[0-4][0-9])|([0-1]?[0-9]?[0-9]))\.((25[0-5])|(2[0-4][0-9])|([0-1]?[0-9]?[0-9]))\.((25[0-5])|(2[0-4][0-9])|([0-1]?[0-9]?[0-9])))|((([A-Za-z0-9\-])+\.)+[A-Za-z\-]+))$",
                RegexOptions.Compiled);

        private string _DisplayName;
        private string _Address;
        private string _AddressType;

        /// <summary>
        /// Gets or sets the type of the address (SMTP, EX).
        /// </summary>
        /// <value>The type of the address.</value>
        public string AddressType
        {
            [DebuggerStepThrough]
            get { return _AddressType; }
            [DebuggerStepThrough]
            set { _AddressType = value; }
        }

        /// <summary>
        /// Gets or sets the address.
        /// </summary>
        /// <value>The address.</value>
        public string Address
        {
            [DebuggerStepThrough]
            get { return _Address; }
            [DebuggerStepThrough]
            set { _Address = value; }
        }

        /// <summary>
        /// Gets or sets the display name.
        /// </summary>
        /// <value>The displayname.</value>
        public string DisplayName
        {
            [DebuggerStepThrough]
            get { return _DisplayName; }
            [DebuggerStepThrough]
            set { _DisplayName = value; }
        }

        /// <summary>
        /// Gets a value indicating whether this instance is valid SMTP address.
        /// </summary>
        /// <value>
        /// 	<see langword="true"/> if this instance is valid SMTP address; otherwise, <see langword="false"/>.
        /// </value>
        public bool IsValidSmtpAddress
        {
            get { return _EmailAddressValidator.IsMatch(Address); }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="EmailAddress"/> class.
        /// </summary>
        /// <param name="displayName">The display name.</param>
        /// <param name="address">The address.</param>
        /// <param name="addressType">The Type of the address.</param>
        public EmailAddress(string displayName, string address, string addressType)
        {
            _DisplayName = displayName;
            _Address = address;
            _AddressType = addressType;

            if (_AddressType.Equals("SMTP", StringComparison.InvariantCultureIgnoreCase) && !IsValidSmtpAddress)
            {
                _AddressType = "Unkown";
            }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="EmailAddress"/> class.
        /// </summary>
        /// <param name="displayName">The displayname.</param>
        /// <param name="address">The address.</param>
        public EmailAddress(string displayName, string address) : this(displayName, address, "SMTP")
        {
        }

        #region IEquatable<EmailAddress> Members

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object"></see> is equal to the current <see cref="T:System.Object"></see>.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object"></see> to compare with the current <see cref="T:System.Object"></see>.</param>
        /// <returns>
        /// true if the specified <see cref="T:System.Object"></see> is equal to the current <see cref="T:System.Object"></see>; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            if (ReferenceEquals(this, obj)) return true;
            EmailAddress emailAddress = obj as EmailAddress;
            if (emailAddress == null) return false;
            return Equals(_Address, emailAddress._Address) && Equals(_AddressType, emailAddress._AddressType);
        }

        /// <summary>
        /// Serves as a hash function for a particular type. <see cref="M:System.Object.GetHashCode"></see> is suitable for use in hashing algorithms and data structures like a hash table.
        /// </summary>
        /// <returns>
        /// A hash code for the current <see cref="T:System.Object"></see>.
        /// </returns>
        public override int GetHashCode()
        {
            return
                (_Address != null ? _Address.GetHashCode() : 0) +
                29*(_AddressType != null ? _AddressType.GetHashCode() : 0);
        }

        /// <summary>
        /// Implicit equality comparison between two <see cref="EmailAddress"/> instances
        /// </summary>
        /// <param name="x">The first email address to compare</param>
        /// <param name="y">The the second email address to compare</param>
        /// <returns>true, if both instances are considerd equal. False, otherwise</returns>
        public static bool operator ==(EmailAddress x, EmailAddress y)
        {
            if ((object) x == (object) y) return true;

            if ((object) x == null) return false;
            if ((object) y == null) return false;

            return x.Equals(y);
        }

        /// <summary>
        /// Implicit inequality comparison between two <see cref="EmailAddress"/> instances
        /// </summary>
        /// <param name="x">The first email address to compare</param>
        /// <param name="y">The the second email address to compare</param>
        /// <returns>true, if both instances are considerd different. False, otherwise</returns>
        public static bool operator !=(EmailAddress x, EmailAddress y)
        {
            if ((object) x == (object) y) return false;

            if ((object) x == null) return true;
            if ((object) y == null) return true;

            return !x.Equals(y);
        }

        #endregion
    }
}