using System.Collections.Generic;
using System.Xml;
using System.Xml.XPath;

namespace InfiniTec.Exchange
{
    /// <summary>
    /// Represents a multi-valued property of an Exchange item.
    /// </summary>
    /// <typeparam name="T">The type of the property</typeparam>
    public class MultiValuedProperty<T> : Property
    {
        private PropertyValueCollection<T> _Values;

        /// <summary>
        /// Returns a list containing the elements of the property.
        /// </summary>
        public IList<T> Values
        {
            get { return _Values; }
        }

        internal override sealed void Load(XPathNavigator navigator)
        {
            navigator.MoveToChild(XPathNodeType.Element);

            do
            {
                _Values.Add(DecodeValue<T>(navigator.InnerXml));
            } while (navigator.MoveToNext(XPathNodeType.Element));
        }

        internal override sealed void SaveTo(XmlWriter writer)
        {
            WriteEncodedStartElementForPropertyName(writer, this);

            if (_Values.Count > 0)
            {
                writer.WriteAttributeString("dt", Namespaces.DataType, "mv." + GetDataType<T>());

                foreach (T value in Values)
                {
                    writer.WriteElementString("v", Namespaces.Xml, EncodeValue(value));
                }
            }
            writer.WriteEndElement();
        }

        /// <summary>
        /// Creates a new instance of this class using the specified namespace and local
        /// name.
        /// </summary>
        /// <param name="ns">The namespace of the property</param>
        /// <param name="localName">The local name of the property</param>
        /// <param name="values">The initial values of the property</param>
        public MultiValuedProperty(string ns, string localName, T[] values) : base(ns, localName)
        {
            _Values = new PropertyValueCollection<T>(this);

            if (values == null) return;

            _Values.AddRange(values);

            SetChanged();
        }

        /// <summary>
        /// Creates a new instance of this class using the specified namespace and local
        /// name.
        /// </summary>
        /// <param name="fullQualifiedName">The full qualified name of the property</param>
        /// <param name="values">The initial values of the property</param>
        public MultiValuedProperty(string fullQualifiedName,
                                   T[] values) : base(fullQualifiedName)
        {
            _Values = new PropertyValueCollection<T>(this);

            if (values == null) return;

            _Values.AddRange(values);

            SetChanged();
        }

        /// <summary>
        /// Creates a new instance of this class using the specified namespace and local
        /// name.
        /// </summary>
        /// <param name="ns">The namespace of the property</param>
        /// <param name="localName">The local name of the property</param>
        public MultiValuedProperty(string ns, string localName)
            : base(ns, localName)
        {
            _Values = new PropertyValueCollection<T>(this);

            SetChanged();
        }

        /// <summary>
        /// Creates a new instance of this class using the specified namespace and local
        /// name.
        /// </summary>
        /// <param name="fullQualifiedName">The full qualified name of the property</param>
        public MultiValuedProperty(string fullQualifiedName)
            : base(fullQualifiedName)
        {
            _Values = new PropertyValueCollection<T>(this);

            SetChanged();
        }

        internal override bool IsPropertyValueNull()
        {
            return Values.Count == 0;
        }
    }
}