using System;
using System.Diagnostics;
using InfiniTec.Exchange.WebDav;

namespace InfiniTec.Exchange
{
    /// <summary>
    /// This class contains information about a physical address.
    /// </summary>
    public class PhysicalAddress : ICollectionEntry<PhysicalAddressType>
    {
        private string _Street;
        private string _City;
        private string _State;
        private string _Country;
        private string _PostalCode;
        private Item _Item;
        private PhysicalAddressType _AddressType;

        /// <summary>
        /// Initializes a new instance of the <see cref="PhysicalAddress"/> class.
        /// </summary>
        /// <param name="addressType">Type of the address.</param>
        public PhysicalAddress(PhysicalAddressType addressType)
        {
            _AddressType = addressType;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="PhysicalAddress"/> class.
        /// </summary>
        /// <param name="addressType">Type of the address.</param>
        /// <param name="item">The item.</param>
        public PhysicalAddress(PhysicalAddressType addressType, Item item)
        {
            _AddressType = addressType;
            _Item = item;

            Load();
        }

        private void Load()
        {
            switch (AddressType)
            {
                case PhysicalAddressType.Business:
                    _Street = _Item.GetPropertyValue<string>(WellknownProperties.Contact.BusinessStreet);
                    _City = _Item.GetPropertyValue<string>(WellknownProperties.Contact.BusinessCity);
                    _State = _Item.GetPropertyValue<string>(WellknownProperties.Contact.BusinessStateOrProvince);
                    _Country = _Item.GetPropertyValue<string>(WellknownProperties.Contact.BusinessCountry);
                    _PostalCode = _Item.GetPropertyValue<string>(WellknownProperties.Contact.BusinessPostalCode);
                    break;
                case PhysicalAddressType.Home:
                    _Street = _Item.GetPropertyValue<string>(WellknownProperties.Contact.HomeStreet);
                    _City = _Item.GetPropertyValue<string>(WellknownProperties.Contact.HomeAddressCity);
                    _State = _Item.GetPropertyValue<string>(WellknownProperties.Contact.HomeStateOrProvince);
                    _Country = _Item.GetPropertyValue<string>(WellknownProperties.Contact.HomeAddressCountry);
                    _PostalCode = _Item.GetPropertyValue<string>(WellknownProperties.Contact.HomePostalCode);
                    break;
                case PhysicalAddressType.Other:
                    _Street = _Item.GetPropertyValue<string>(WellknownProperties.Contact.OtherStreet);
                    _City = _Item.GetPropertyValue<string>(WellknownProperties.Contact.OtherCity);
                    _State = _Item.GetPropertyValue<string>(WellknownProperties.Contact.OtherStateOrProvince);
                    _Country = _Item.GetPropertyValue<string>(WellknownProperties.Contact.OtherCountry);
                    _PostalCode = _Item.GetPropertyValue<string>(WellknownProperties.Contact.OtherPostalCode);
                    break;
                default:
                    throw new NotSupportedException("The specified addresstype is not supported");
            }
        }

        /// <summary>
        /// Gets or sets the type of the address.
        /// </summary>
        /// <value>The type of the address.</value>
        public PhysicalAddressType AddressType
        {
            [DebuggerStepThrough]
            get { return _AddressType; }
            [DebuggerStepThrough]
            set { _AddressType = value; }
        }


        /// <summary>
        /// Gets or sets the postal code.
        /// </summary>
        /// <value>The postal code.</value>
        public string PostalCode
        {
            [DebuggerStepThrough]
            get { return _PostalCode; }
            [DebuggerStepThrough]
            set { _PostalCode = value; }
        }

        /// <summary>
        /// Gets or sets the country.
        /// </summary>
        /// <value>The country.</value>
        public string Country
        {
            [DebuggerStepThrough]
            get { return _Country; }
            [DebuggerStepThrough]
            set { _Country = value; }
        }

        /// <summary>
        /// Gets or sets the state.
        /// </summary>
        /// <value>The state.</value>
        public string State
        {
            [DebuggerStepThrough]
            get { return _State; }
            [DebuggerStepThrough]
            set { _State = value; }
        }

        /// <summary>
        /// Gets or sets the city.
        /// </summary>
        /// <value>The city.</value>
        public string City
        {
            [DebuggerStepThrough]
            get { return _City; }
            [DebuggerStepThrough]
            set { _City = value; }
        }

        /// <summary>
        /// Gets or sets the street.
        /// </summary>
        /// <value>The street.</value>
        public string Street
        {
            [DebuggerStepThrough]
            get { return _Street; }
            [DebuggerStepThrough]
            set { _Street = value; }
        }

        #region ICollectionEntry<PhysicalAddressType> Members

        void ICollectionEntry<PhysicalAddressType>.SetItem(Item item)
        {
            _Item = item;
        }

        void ICollectionEntry<PhysicalAddressType>.Save()
        {
            switch (AddressType)
            {
                case PhysicalAddressType.Business:
                    _Item.SetPropertyValue<string>(WellknownProperties.Contact.BusinessStreet, Street);
                    _Item.SetPropertyValue<string>(WellknownProperties.Contact.BusinessCity, City);
                    _Item.SetPropertyValue<string>(WellknownProperties.Contact.BusinessStateOrProvince, State);
                    _Item.SetPropertyValue<string>(WellknownProperties.Contact.BusinessCountry, Country);
                    _Item.SetPropertyValue<string>(WellknownProperties.Contact.BusinessPostalCode, PostalCode);
                    break;
                case PhysicalAddressType.Home:
                    _Item.SetPropertyValue<string>(WellknownProperties.Contact.HomeStreet, Street);
                    _Item.SetPropertyValue<string>(WellknownProperties.Contact.HomeAddressCity, City);
                    _Item.SetPropertyValue<string>(WellknownProperties.Contact.HomeStateOrProvince, State);
                    _Item.SetPropertyValue<string>(WellknownProperties.Contact.HomeAddressCountry, Country);
                    _Item.SetPropertyValue<string>(WellknownProperties.Contact.HomePostalCode, PostalCode);
                    break;
                case PhysicalAddressType.Other:
                    _Item.SetPropertyValue<string>(WellknownProperties.Contact.OtherStreet, Street);
                    _Item.SetPropertyValue<string>(WellknownProperties.Contact.OtherCity, City);
                    _Item.SetPropertyValue<string>(WellknownProperties.Contact.OtherStateOrProvince, State);
                    _Item.SetPropertyValue<string>(WellknownProperties.Contact.OtherCountry, Country);
                    _Item.SetPropertyValue<string>(WellknownProperties.Contact.OtherPostalCode, PostalCode);
                    break;
                default:
                    throw new NotSupportedException("The specified addresstype is not supported");
            }
        }

        void ICollectionEntry<PhysicalAddressType>.Delete()
        {
            switch (AddressType)
            {
                case PhysicalAddressType.Business:
                    _Item.SetPropertyValue<string>(WellknownProperties.Contact.BusinessStreet, null);
                    _Item.SetPropertyValue<string>(WellknownProperties.Contact.BusinessCity, null);
                    _Item.SetPropertyValue<string>(WellknownProperties.Contact.BusinessStateOrProvince, null);
                    _Item.SetPropertyValue<string>(WellknownProperties.Contact.BusinessCountry, null);
                    _Item.SetPropertyValue<string>(WellknownProperties.Contact.BusinessPostalCode, null);
                    break;
                case PhysicalAddressType.Home:
                    _Item.SetPropertyValue<string>(WellknownProperties.Contact.HomeStreet, null);
                    _Item.SetPropertyValue<string>(WellknownProperties.Contact.HomeAddressCity, null);
                    _Item.SetPropertyValue<string>(WellknownProperties.Contact.HomeStateOrProvince, null);
                    _Item.SetPropertyValue<string>(WellknownProperties.Contact.HomeAddressCountry, null);
                    _Item.SetPropertyValue<string>(WellknownProperties.Contact.HomePostalCode, null);
                    break;
                case PhysicalAddressType.Other:
                    _Item.SetPropertyValue<string>(WellknownProperties.Contact.OtherStreet, null);
                    _Item.SetPropertyValue<string>(WellknownProperties.Contact.OtherCity, null);
                    _Item.SetPropertyValue<string>(WellknownProperties.Contact.OtherStateOrProvince, null);
                    _Item.SetPropertyValue<string>(WellknownProperties.Contact.OtherCountry, null);
                    _Item.SetPropertyValue<string>(WellknownProperties.Contact.OtherPostalCode, null);
                    break;
                default:
                    throw new NotSupportedException("The specified addresstype is not supported");
            }
        }

        PhysicalAddressType ICollectionEntry<PhysicalAddressType>.GetKey()
        {
            return AddressType;
        }

        #endregion
    }
}