using System;
using System.Collections;
using System.Collections.Generic;
using System.Net;

namespace InfiniTec.Exchange
{
    /// <summary>
    /// This class manages the various properties of an Item.
    /// </summary>
    public class PropertyCollection : ICollection<Property>
    {
        private Dictionary<string, Property> _Properties;
        private List<string> _RemovedProperties;

        internal IList<string> RemovedProperties
        {
            get { return _RemovedProperties; }
        }

        /// <summary>
        /// Creates a new instance of this class.
        /// </summary>
        public PropertyCollection()
        {
            _Properties = new Dictionary<string, Property>();
            _RemovedProperties = new List<string>();
        }

        /// <summary>
        /// Adds the property to the collection.
        /// </summary>
        /// <param name="property">The property to add</param>
        public void Add(Property property)
        {
            if (_Properties.ContainsKey(property.FullQualifiedName))
            {
                if (_Properties[property.FullQualifiedName].Status.StatusCode == HttpStatusCode.OK)
                    throw
                        new ArgumentException("A property with the name " + property.FullQualifiedName +
                                              " already exists in the property collection");
                else _Properties.Remove(property.FullQualifiedName);
            }
            _Properties.Add(property.FullQualifiedName, property);
            if (_RemovedProperties.Contains(property.FullQualifiedName))
            {
                _RemovedProperties.Remove(property.FullQualifiedName);
            }
        }

        /// <summary>
        /// Adds the specified properties to the collection.
        /// </summary>
        /// <param name="properties">The properties to add</param>
        public void AddRange(IEnumerable<Property> properties)
        {
            foreach (Property property in properties)
            {
                Add(property);
            }
        }

        /// <summary>
        /// Creates a new <see cref="Property"/> and adds it to the collection.
        /// </summary>
        /// <typeparam name="T">The type of the property</typeparam>
        /// <param name="ns">The namespace of the property</param>
        /// <param name="name">The local name of the property</param>
        /// <param name="value">The value of the property</param>
        public void Add<T>(string ns, string name, T value)
        {
            Add(new Property<T>(ns, name, value));
        }

        /// <summary>
        /// Creates a new <see cref="Property"/> and adds it to the collection.
        /// </summary>
        /// <typeparam name="T">The type of the property</typeparam>
        /// <param name="fullQualifiedName">The name of the property, 
        /// including the namespace and the local name.</param>
        /// <param name="value">The value of the property</param>
        public void Add<T>(string fullQualifiedName, T value)
        {
            Add(new Property<T>(fullQualifiedName, value));
        }

        /// <summary>
        /// Creates a new <see cref="Property"/> and adds it to the collection.
        /// </summary>
        /// <typeparam name="T">The type of the property</typeparam>
        /// <param name="ns">The namespace of the property</param>
        /// <param name="name">The local name of the property</param>
        /// <param name="values">The value of the property</param>
        public void AddMultiValued<T>(string ns, string name, T[] values)
        {
            Add(new MultiValuedProperty<T>(ns, name, values));
        }

        /// <summary>
        /// Creates a new <see cref="Property"/> and adds it to the collection.
        /// </summary>
        /// <typeparam name="T">The type of the property</typeparam>
        /// <param name="fullQualifiedName">The name of the property, 
        /// including the namespace and the local name.</param>
        /// <param name="values">The value of the property</param>
        public void AddMultiValued<T>(string fullQualifiedName, T[] values)
        {
            Add(new MultiValuedProperty<T>(fullQualifiedName, values));
        }

        /// <summary>
        /// Returns true, if the collection contains a <see cref="Property"/>
        /// with the same name as the specified <see cref="Property"/>
        /// </summary>
        /// <param name="item">The <see cref="Property"/> to search for.</param>
        /// <returns>True, if the collection contains another 
        /// property with the same name.</returns>
        public bool Contains(Property item)
        {
            return Contains(item.FullQualifiedName);
        }

        /// <summary>
        /// Returns true, if the collection contains a <see cref="Property"/>
        /// with the specified name
        /// </summary>
        /// <param name="fullQualifiedName">The name of the property to search for.</param>
        /// <returns>True, if the collection contains another 
        /// property with the same name.</returns>
        public bool Contains(string fullQualifiedName)
        {
            Property property;

            if (!_Properties.TryGetValue(fullQualifiedName, out property)) return false;

            return property.Status.StatusCode != HttpStatusCode.NotFound;
        }

        /// <summary>
        /// Returns true, if the collection contains a <see cref="Property"/>
        /// with the specified name
        /// </summary>
        /// <param name="ns">The namespace of the property</param>
        /// <param name="localName">The name of the property to search for.</param>
        /// <returns>True, if the collection contains another 
        /// property with the same name.</returns>
        public bool Contains(string ns, string localName)
        {
            return Contains(ns + localName);
        }

        /// <summary>
        /// Copyies the contents of this collection to the specified
        /// array.
        /// </summary>
        /// <param name="array">The destination array.</param>
        /// <param name="arrayIndex"></param>
        public void CopyTo(Property[] array, int arrayIndex)
        {
            _Properties.Values.CopyTo(array, arrayIndex);
        }

        /// <summary>
        /// Removes the property with the same name from the collection.
        /// </summary>
        /// <param name="property">The property to remove</param>
        /// <returns>True, if the property was removed. False, otherwise.</returns>
        public bool Remove(Property property)
        {
            return Remove(property.FullQualifiedName);
        }

        /// <summary>
        /// Returns the <see cref="Property"/> with the specified name.
        /// </summary>
        /// <param name="fullQualifedName">The full qualified name of the property</param>
        /// <returns>The <see cref="Property"/> with the specified name.</returns>
        public Property this[string fullQualifedName]
        {
            get
            {
                Property property;

                if (!Contains(fullQualifedName))
                    throw new PropertyNotFoundException("The requested property was not found");

                property = _Properties[fullQualifedName];
                if (property.Status.StatusCode == HttpStatusCode.NotFound)
                {
                    throw property.Status.GetException();
                }
                return property;
            }
        }

        /// <summary>
        /// Returns the <see cref="Property"/> with the specified name.
        /// </summary>
        /// <param name="ns">The namespace of the property</param>
        /// <param name="localName">The local name of the property</param>
        /// <returns>The <see cref="Property"/> with the specified name.</returns>
        public Property this[string ns, string localName]
        {
            get { return this[ns + localName]; }
        }

        /// <summary>
        /// Returns the <see cref="Property"/> with the specified name.
        /// </summary>
        /// <remarks>The requested property must be a single-valued property</remarks>
        /// <typeparam name="T">The type of the property</typeparam>
        /// <param name="ns">The namespace of the property</param>
        /// <param name="localName">The local name of the property</param>
        /// <returns>The requested property</returns>
        public Property<T> GetProperty<T>(string ns, string localName)
        {
            return (Property<T>) this[ns, localName];
        }

        /// <summary>
        /// Returns the <see cref="Property"/> with the specified name.
        /// </summary>
        /// <remarks>The requested property must be a single-valued property</remarks>
        /// <typeparam name="T">The type of the property</typeparam>
        /// <param name="fullQualifedName">The full qualified name of the property</param>  
        /// <returns>The requested property</returns>
        public Property<T> GetProperty<T>(string fullQualifedName)
        {
            return (Property<T>) this[fullQualifedName];
        }

        /// <summary>
        /// Returns the <see cref="Property"/> with the specified name.
        /// </summary>
        /// <remarks>The requested property must be a single-valued property</remarks>
        /// <typeparam name="T">The type of the property</typeparam>
        /// <param name="fullQualifedName">The full qualified name of the property</param>  
        /// <returns>The requested property</returns>
        public MultiValuedProperty<T> GetMultivaluedProperty<T>(string fullQualifedName)
        {
            return (MultiValuedProperty<T>) this[fullQualifedName];
        }

        /// <summary>
        /// Returns the <see cref="Property"/> with the specified name.
        /// </summary>
        /// <remarks>The requested property must be a multi-valued property</remarks>
        /// <typeparam name="T">The type of the property</typeparam>
        /// <param name="ns">The namespace of the property</param>
        /// <param name="localName">The local name of the property</param>
        /// <returns>The requested property</returns>
        public MultiValuedProperty<T> GetMultivaluedProperty<T>(string ns, string localName)
        {
            return (MultiValuedProperty<T>) this[ns, localName];
        }

        /// <summary>
        /// Tries to get the property with the specified name.
        /// </summary>
        /// <typeparam name="T">The type of the property</typeparam>
        /// <param name="ns">The namespace of the property</param>
        /// <param name="localName">The local name of the property</param>
        /// <param name="property">When the method returns, contains the requested property, if found. Null, otherwise</param>
        /// <returns>True, if the property was found. False, otherwise</returns>
        public bool TryGetProperty<T>(string ns, string localName, out Property<T> property)
        {
            return TryGetProperty<T>(ns + localName, out property);
        }

        /// <summary>
        /// Tries to get the property with the specified name.
        /// </summary>
        /// <typeparam name="T">The type of the property</typeparam>
        /// <param name="fullQualifiedName">The full qualified name of the property</param>  
        /// <param name="property">When the method returns, contains the requested property, if found. Null, otherwise</param>
        /// <returns>True, if the property was found. False, otherwise</returns>
        public bool TryGetProperty<T>(string fullQualifiedName, out Property<T> property)
        {
            Property prop;

            property = null;
            if (!Contains(fullQualifiedName)) return false;

            prop = _Properties[fullQualifiedName];
            if (prop.Status.StatusCode != HttpStatusCode.OK)
                return false;

            property = (Property<T>) prop;

            return true;
        }

        /// <summary>
        /// Tries to get the multi-valued property with the specified name.
        /// </summary>
        /// <typeparam name="T">The type of the property</typeparam>
        /// <param name="fullQualifiedName">The full qualified name of the property</param>  
        /// <param name="property">When the method returns, contains the requested property, if found. Null, otherwise</param>
        /// <returns>True, if the property was found. False, otherwise</returns>
        public bool TryGetMultiValuedProperty<T>(string fullQualifiedName, out MultiValuedProperty<T> property)
        {
            Property prop;

            property = null;
            if (!Contains(fullQualifiedName)) return false;

            prop = _Properties[fullQualifiedName];
            if (prop.Status.StatusCode != HttpStatusCode.OK)
                return false;

            property = (MultiValuedProperty<T>) prop;

            return true;
        }

        /// <summary>
        /// Tries to get the property with the specified name.
        /// </summary>
        /// <typeparam name="T">The type of the property</typeparam>
        /// <param name="ns">The namespace of the property</param>
        /// <param name="localName">The local name of the property</param>
        /// <param name="property">When the method returns, contains the requested property, if found. Null, otherwise</param>
        /// <returns>True, if the property was found. False, otherwise</returns>
        public bool TryGetProperty<T>(string ns, string localName, out MultiValuedProperty<T> property)
        {
            return TryGetMultiValuedProperty<T>(ns + localName, out property);
        }

        /// <summary>
        /// Removes the <see cref="Property"/> with the specified name
        /// from the collection.
        /// </summary>
        /// <param name="fullQualifiedName">The full qualified name of the property</param>
        /// <returns>True, if the property was removed. False, otherwise.</returns>
        public bool Remove(string fullQualifiedName)
        {
            if (_Properties.Remove(fullQualifiedName))
            {
                _RemovedProperties.Add(fullQualifiedName);
                return true;
            }

            return false;
        }

        /// <summary>
        /// Removes the <see cref="Property"/> with the specified name
        /// from the collection.
        /// </summary>
        /// <param name="ns">The namespace of the property</param>
        /// <param name="localName">The local name of the property</param>
        /// <returns>True, if the property was removed. False, otherwise.</returns>
        public bool Remove(string ns, string localName)
        {
            return Remove(ns + localName);
        }


        /// <summary>
        /// Returns an <see cref="IEnumerator"/> which can be used to iterate
        /// through the collection of properties.
        /// </summary>
        /// <returns>An <see cref="IEnumerator"/> which can be used to iterate
        /// through the collection of properties</returns>
        IEnumerator<Property> IEnumerable<Property>.GetEnumerator()
        {
            return _Properties.Values.GetEnumerator();
        }

        #region ICollection<Property> Members

        /// <summary>
        /// Removes all properties from the collection.
        /// </summary>
        public void Clear()
        {
            _RemovedProperties.AddRange(_Properties.Keys);
            _Properties.Clear();
        }

        /// <summary>
        /// Returns the number of properties in the collection.
        /// </summary>
        public int Count
        {
            get { return _Properties.Count; }
        }

        /// <summary>
        /// Returns false;
        /// </summary>
        public bool IsReadOnly
        {
            get { return false; }
        }

        #endregion

        #region IEnumerable Members

        /// <summary>
        /// Returns an <see cref="IEnumerator"/> which can be used to iterate
        /// through the collection of properties.
        /// </summary>
        /// <returns>An <see cref="IEnumerator"/> which can be used to iterate
        /// through the collection of properties</returns>
        public IEnumerator GetEnumerator()
        {
            return _Properties.Values.GetEnumerator();
        }

        #endregion
    }
}