using System;
using System.Collections.Generic;
using System.Security.AccessControl;
using System.Security.Principal;
using InfiniTec.Security;

namespace InfiniTec.Exchange.Security
{
    internal class AceComparer : IComparer<CommonAce>
    {
        private Connection _Connection;

        public AceComparer(Connection connection)
        {
            if (connection == null) throw new ArgumentNullException("connection");
            _Connection = connection;
        }

        #region IComparer<CommonAce> Members

        public int Compare(CommonAce leftAce, CommonAce rightAce)
        {
            int result;

            if (leftAce == null && rightAce == null) return 0;
            if (leftAce == null) return -1;
            if (rightAce == null) return 1;

            result = CompareAccountType(leftAce.SecurityIdentifier, rightAce.SecurityIdentifier);

            if (result != 0) return result;

            // Groups and users must be sorted differently
            if (GetAccountType(leftAce.SecurityIdentifier) == SidAccountType.Group)
            {
                result = CompareAceType(leftAce.AceType, rightAce.AceType);
                if (result != 0) return result;

                result = leftAce.SecurityIdentifier.CompareTo(rightAce.SecurityIdentifier);
                if (result != 0) return result;

                result = CompareInheritanceFlags(leftAce, rightAce);
                if (result != 0) return result;
            }
            else
            {
                result = leftAce.SecurityIdentifier.CompareTo(rightAce.SecurityIdentifier);
                if (result != 0) return result;

                result = CompareInheritanceFlags(leftAce, rightAce);
                if (result != 0) return result;

                result = CompareAceType(leftAce.AceType, rightAce.AceType);
                if (result != 0) return result;
            }

            return 0;
        }

        private static int CompareInheritanceFlags(CommonAce leftAce, CommonAce rightAce)
        {
            if ((leftAce.AceFlags & AceFlags.InheritOnly) != AceFlags.InheritOnly &&
                (rightAce.AceFlags & AceFlags.InheritOnly) == AceFlags.InheritOnly)
                return -1;
            else if ((leftAce.AceFlags & AceFlags.InheritOnly) == AceFlags.InheritOnly &&
                     (rightAce.AceFlags & AceFlags.InheritOnly) != AceFlags.InheritOnly)
                return 1;
            else
            {
                if ((leftAce.InheritanceFlags & InheritanceFlags.ContainerInherit) ==
                    InheritanceFlags.ContainerInherit &&
                    (rightAce.InheritanceFlags & InheritanceFlags.ContainerInherit) !=
                    InheritanceFlags.ContainerInherit)
                    return -1;
                else if ((leftAce.InheritanceFlags & InheritanceFlags.ContainerInherit) ==
                         InheritanceFlags.ContainerInherit &&
                         (rightAce.InheritanceFlags & InheritanceFlags.ContainerInherit) !=
                         InheritanceFlags.ContainerInherit)
                    return 1;
            }

            return 0;

            //if (leftAce.InheritanceFlags == rightAce.InheritanceFlags) return 0;
            //if (leftAce.InheritanceFlags == InheritanceFlags.ObjectInherit) return -1;
            //else if (rightAce.InheritanceFlags == InheritanceFlags.ObjectInherit) return 1;
            //else return 0;
        }

        private int CompareAccountType(SecurityIdentifier leftSid, SecurityIdentifier rightSid)
        {
            SidAccountType leftType;
            SidAccountType rightType;


            leftType = GetAccountType(leftSid);
            rightType = GetAccountType(rightSid);

            return leftType.CompareTo(rightType);
        }

        private SidAccountType GetAccountType(SecurityIdentifier sid)
        {
            AccountType type;

            if (sid.IsWellKnown(WellKnownSidType.AnonymousSid)) return SidAccountType.Anonymous;
            if (sid.IsWellKnown(WellKnownSidType.WorldSid)) return SidAccountType.Everyone;

            type = AccountTypeResolver.ResolveAccountType(sid, _Connection);

            if (type == AccountType.Group || type == AccountType.WellknownGroup)
            {
                return SidAccountType.Group;
            }
            else
            {
                return SidAccountType.User;
            }
        }

        private static int CompareAceType(AceType left, AceType right)
        {
            if (left == right) return 0;

            if (left == AceType.AccessAllowed) return -1;

            return 1;
        }

        #endregion
    }
}