using System.Collections.Generic;
using System.Security.AccessControl;
using System.Security.Principal;

namespace InfiniTec.Exchange.Security
{
    /// <summary>
    /// Extends the <see cref="ItemAce"/> class for manipulation of access control entries
    /// on folders
    /// </summary>
    public class FolderAce : ItemAce
    {
        private CommonAce _SubFolderAllowedAce;
        private CommonAce _SubFolderDeniedAce;
        private CommonAce _FolderAllowedAce;
        private CommonAce _FolderDeniedAce;

        /// <summary>
        /// Returns the <see cref="AccessRole"/> for the current ace.
        /// </summary>
        public AccessRole Role
        {
            get
            {
                return RoleMapping.GetAccessRole(
                    ((_FolderAllowedAce != null)
                         ? (FolderAccessMasks) _FolderAllowedAce.AccessMask
                         : FolderAccessMasks.None),
                    ((ItemAllowedAce != null) ? (ItemAccessMasks) ItemAllowedAce.AccessMask : ItemAccessMasks.None));
            }
        }

        /// <summary>
        /// Returns the <see cref="ItemAccessMasks"/> containing the permissions for
        /// items in the folder.
        /// </summary>
        public override ItemAccessMasks ItemAccessMask
        {
            get { return ((ItemAllowedAce != null) ? (ItemAccessMasks) ItemAllowedAce.AccessMask : ItemAccessMasks.None); }
        }

        /// <summary>
        /// Returns the <see cref="FolderAccessMasks"/> containing the permissions for
        /// this folder
        /// </summary>
        public FolderAccessMasks FolderAccessMask
        {
            get
            {
                return
                    ((_FolderAllowedAce != null)
                         ? (FolderAccessMasks) _FolderAllowedAce.AccessMask
                         : FolderAccessMasks.None);
            }
        }

        /// <summary>
        /// Returns the permissions on this folder
        /// </summary>
        public override AccessMasks AccessMask
        {
            get { return GetFolderAccessMask() | GetItemAccessMask(ItemAllowedAce); }
        }


        private AccessMasks GetFolderAccessMask()
        {
            if (_FolderAllowedAce == null) return AccessMasks.None;

            switch ((FolderAccessMasks) _FolderAllowedAce.AccessMask)
            {
                case FolderAccessMasks.None:
                    return AccessMasks.None;
                case FolderAccessMasks.CreateItems:
                    return AccessMasks.CreateItems;
                case FolderAccessMasks.ReadItems:
                    return AccessMasks.Read;
                case FolderAccessMasks.CreateSubfolder:
                    return AccessMasks.CreateSubFolder;
                case FolderAccessMasks.FolderOwner:
                    return AccessMasks.FolderOwner;
                case FolderAccessMasks.FolderContact:
                    return AccessMasks.FolderContact;
                case FolderAccessMasks.FolderVisible:
                    return AccessMasks.FolderVisible;
                default:
                    return AccessMasks.Special;
            }
        }

        internal override void LoadAce(CommonAce ace)
        {
            if (ace.InheritanceFlags == InheritanceFlags.ContainerInherit &&
                (ace.AceFlags & AceFlags.InheritOnly) == AceFlags.InheritOnly)
            {
                switch (ace.AceType)
                {
                    case AceType.AccessAllowed:
                        _SubFolderAllowedAce = ace;
                        break;
                    case AceType.AccessDenied:
                        _SubFolderDeniedAce = ace;
                        break;
                }
            }
            else if (ace.InheritanceFlags == InheritanceFlags.ObjectInherit)
            {
                LoadAceInternal(ace);
            }
            else
            {
                switch (ace.AceType)
                {
                    case AceType.AccessAllowed:
                        _FolderAllowedAce = ace;
                        break;
                    case AceType.AccessDenied:
                        _FolderDeniedAce = ace;
                        break;
                }
            }
        }

        /// <summary>
        /// Creates a new instance of this class
        /// </summary>
        /// <param name="identity">The <see cref="SecurityIdentifier"/> to use</param>
        public FolderAce(SecurityIdentifier identity) : base(identity)
        {
        }

        internal override void SetAccess(AccessMasks accessMask)
        {
            ItemAccessMasks itemAccessMask;
            FolderAccessMasks folderAccessMask;

            itemAccessMask = DetermineItemAccessMask(accessMask);
            folderAccessMask = DetermineFolderAccessMask(accessMask);

            DangerousSetAccess(folderAccessMask, itemAccessMask);
        }

        internal void SetAccess(AccessRole role)
        {
            RoleMapping mapping;

            mapping = RoleMapping.GetRoleMapping(role);

            DangerousSetAccess(mapping.FolderAccessMask,
                               mapping.ItemAccessMask);
        }

        /// <summary>
        /// Sets the item-level permissions on this folder
        /// </summary>
        /// <param name="accessMask">The desired access mask</param>
        protected internal override void DangerousSetAccess(ItemAccessMasks accessMask)
        {
            DangerousSetAccessInternal(accessMask, AceFlags.InheritOnly | AceFlags.ObjectInherit);
        }

        /// <summary>
        /// Sets both, the folder and item level permissions on this ace.
        /// </summary>
        /// <param name="folderAccess">The folder access permissions</param>
        /// <param name="itemAccessMask">The item level access permissions</param>
        protected internal void DangerousSetAccess(FolderAccessMasks folderAccess,
                                                   ItemAccessMasks itemAccessMask)
        {
            DangerousSetAccess(itemAccessMask);

            DangerousSetAccessInternal((int) folderAccess, ref _FolderAllowedAce, AceFlags.ContainerInherit, true);

            //DangerousSetAccessInternal((int) folderAccess, ref _SubFolderAllowedAce,
            //                           AceFlags.ContainerInherit | AceFlags.InheritOnly | AceFlags.NoPropagateInherit, true);


            // The Everyone SID only needs an allow entry
            if (IdentityReference.IsWellKnown(WellKnownSidType.WorldSid)) return;

            DangerousSetAccessInternal((int) (~folderAccess & FolderAccessMasks.All), ref _FolderDeniedAce,
                                       AceFlags.ContainerInherit, false);
            //DangerousSetAccessInternal((int) (~folderAccess & FolderAccessMasks.All), ref _SubFolderDeniedAce,
            //                           AceFlags.ContainerInherit | AceFlags.InheritOnly | AceFlags.NoPropagateInherit, false);
        }


        private static FolderAccessMasks DetermineFolderAccessMask(AccessMasks accessMask)
        {
            FolderAccessMasks mask;

            mask = FolderAccessMasks.None;

            if ((accessMask & AccessMasks.CreateItems) == AccessMasks.CreateItems)
                mask |= FolderAccessMasks.CreateItems;
            if ((accessMask & AccessMasks.CreateSubFolder) == AccessMasks.CreateSubFolder)
                mask |= FolderAccessMasks.CreateSubfolder;
            if ((accessMask & AccessMasks.FolderContact) == AccessMasks.FolderContact)
                mask |= FolderAccessMasks.FolderContact;
            if ((accessMask & AccessMasks.FolderOwner) == AccessMasks.FolderOwner)
                mask |= FolderAccessMasks.FolderOwner;
            if ((accessMask & AccessMasks.FolderVisible) == AccessMasks.FolderVisible)
                mask |= FolderAccessMasks.FolderVisible;

            return mask;
        }

        internal override CommonAce[] GetAces()
        {
            List<CommonAce> entries;

            entries = new List<CommonAce>(base.GetAces());

            if (_FolderAllowedAce != null) entries.Add(_FolderAllowedAce);
            if (_FolderDeniedAce != null) entries.Add(_FolderDeniedAce);
            if (_SubFolderAllowedAce != null) entries.Add(_SubFolderAllowedAce);
            if (_SubFolderDeniedAce != null) entries.Add(_SubFolderDeniedAce);

            return entries.ToArray();
        }
    }
}