using System.Collections.Generic;
using System.Security.Principal;

namespace InfiniTec.Exchange.Security
{
    /// <summary>
    /// Provides access and manipulation of the security descriptor of an Exchange
    /// folder
    /// </summary>
    public class FolderSecurity : ExchangeSecurity
    {
        /// <summary>
        /// Creates a new instance of this class using the specified settings
        /// </summary>
        /// <param name="buffer">The buffer containing the security descriptor</param>
        /// <param name="offset">The offset of the security descriptor in the buffer</param>
        /// <param name="connection">The connection to use with this security descriptor</param>
        public FolderSecurity(byte[] buffer, int offset, Connection connection) :
            base(buffer, offset, connection)
        {
        }

        /// <summary>
        /// Returns a list of of <see cref="FolderAce"/> entries which make up the
        /// the discretionary access control list of the folder.
        /// </summary>
        /// <returns>The access control entries of the folder</returns>
        public IList<FolderAce> GetAccessControlEntries()
        {
            return GetAccessControlEntries<FolderAce>();
        }

        /// <summary>
        /// Creates a new <see cref="FolderAce"/>
        /// </summary>
        /// <param name="identity">The identity for the access control entry</param>
        /// <returns>A <see cref="FolderAce"/> instance</returns>
        protected override ItemAce CreateAce(SecurityIdentifier identity)
        {
            return new FolderAce(identity);
        }

        /// <summary>
        /// Sets the desired access for the specified identity
        /// </summary>
        /// <param name="identity">A <see cref="SecurityIdentifier"/> instance</param>
        /// <param name="role">A valid <see cref="AccessRole"/> value</param>
        public void SetAccess(SecurityIdentifier identity, AccessRole role)
        {
            ItemAce ace;

            UpdateControlFlagsForModification();

            if (!TryGetAce(identity, out ace))
            {
                ace = AddAce(identity);
            }

            ((FolderAce) ace).SetAccess(role);

            SortAcl();
        }

        /// <summary>
        /// Removes the access for the specified identity
        /// </summary>
        /// <param name="identity">A <see cref="SecurityIdentifier"/> instance.</param>
        public override void RemoveAccess(SecurityIdentifier identity)
        {
            // The Anonymous and Everyone permission must always be set. The best one can do is to set the accesslevel
            // to None.
            if (identity.IsWellKnown(WellKnownSidType.WorldSid) || identity.IsWellKnown(WellKnownSidType.AnonymousSid))
            {
                SetAccess(identity, AccessRole.None);
                return;
            }

            base.RemoveAccess(identity);
        }


        /// <summary>
        /// Sets the specified folder and item access mask for the identity.
        /// </summary>
        /// <param name="identity">A <see cref="SecurityIdentifier" /> instance</param>
        /// <param name="folderAccessMask">The desired <see cref="FolderAccessMasks"/>.</param>
        /// <param name="itemAccessMask">The desired <see cref="ItemAccessMasks"/>.</param>
        public void DangerousSetAccess(SecurityIdentifier identity, FolderAccessMasks folderAccessMask,
                                       ItemAccessMasks itemAccessMask)
        {
            ItemAce ace;

            if (!TryGetAce(identity, out ace))
            {
                ace = AddAce(identity);
            }
            ((FolderAce) ace).DangerousSetAccess(folderAccessMask, itemAccessMask);
        }
    }
}