using System;
using System.Collections.Generic;
using System.Security.AccessControl;
using System.Security.Principal;

namespace InfiniTec.Exchange.Security
{
    /// <summary>
    /// This class implements manipulation of access control entries of an Exchange
    /// item.
    /// </summary>
    public partial class ItemAce
    {
        internal event EventHandler<AceCreatedEventArgs> AceCreated;
        private CommonAce _ItemAllowedAce;
        private CommonAce _ItemDeniedAce;
        private SecurityIdentifier _IdentityReference;

        internal virtual void OnAceAcreated(AceCreatedEventArgs e)
        {
            if (AceCreated != null) AceCreated(this, e);
        }

        /// <summary>
        /// Returns the internal access control representing the deny permission
        /// of the item access control entry
        /// </summary>
        protected CommonAce ItemDeniedAce
        {
            get { return _ItemDeniedAce; }
        }

        /// <summary>
        /// Returns the internal access control representing the alllow permission
        /// of the item access control entry
        /// </summary>
        protected CommonAce ItemAllowedAce
        {
            get { return _ItemAllowedAce; }
        }

        /// <summary>
        /// Returns the <see cref="ItemAccessMasks"/> for the ace.
        /// </summary>
        public virtual ItemAccessMasks ItemAccessMask
        {
            get
            {
                return (ItemAllowedAce != null)
                           ? (ItemAccessMasks) ItemAllowedAce.AccessMask
                           :
                       ItemAccessMasks.None;
            }
        }

        /// <summary>
        /// Returns the <see cref="SecurityIdentifier"/> for the ace.
        /// </summary>
        public SecurityIdentifier IdentityReference
        {
            get { return _IdentityReference; }
        }

        /// <summary>
        /// Returns the <see cref="AccessMasks"/> for the item
        /// </summary>
        public virtual AccessMasks AccessMask
        {
            get { return GetItemAccessMask(ItemAllowedAce); }
        }

        /// <summary>
        /// Returns the <see cref="SecurityIdentifier"/> for the ace.
        /// </summary>
        /// <param name="ace">The ace to get the access mask for.</param>
        protected static AccessMasks GetItemAccessMask(CommonAce ace)
        {
            ItemAccessMasks mask;
            AccessMasks result;

            if (ace == null) return AccessMasks.None;

            mask = (ItemAccessMasks) ace.AccessMask;
            result = AccessMasks.None;

            if ((mask & ItemAccessMasks.ReadItem) == ItemAccessMasks.ReadItem) result |= AccessMasks.Read;
            if ((mask & ItemAccessMasks.EditOwn) == ItemAccessMasks.EditOwn) result |= AccessMasks.EditOwn;
            if ((mask & ItemAccessMasks.EditAny) == ItemAccessMasks.EditAny) result |= AccessMasks.EditAny;
            if ((mask & ItemAccessMasks.DeleteOwn) == ItemAccessMasks.DeleteOwn) result |= AccessMasks.DeleteOwn;
            if ((mask & ItemAccessMasks.DeleteAny) == ItemAccessMasks.DeleteAny) result |= AccessMasks.DeleteAny;

            return result;
        }

        internal virtual void LoadAce(CommonAce ace)
        {
            if ((ace.AceFlags & AceFlags.InheritOnly) == 0)
            {
                LoadAceInternal(ace);
            }
        }

        /// <summary>
        /// Loads the data from the internal ace
        /// </summary>
        /// <param name="ace">The ace to parse</param>
        protected void LoadAceInternal(CommonAce ace)
        {
            switch (ace.AceType)
            {
                case AceType.AccessAllowed:
                    _ItemAllowedAce = ace;
                    break;
                case AceType.AccessDenied:
                    _ItemDeniedAce = ace;
                    break;
            }
        }

        /// <summary>
        /// Create a new instance of this item for the specified identity
        /// </summary>
        /// <param name="identity">A <see cref="SecurityIdentifier"/> instance</param>
        public ItemAce(SecurityIdentifier identity)
        {
            _IdentityReference = identity;
        }

        internal virtual CommonAce[] GetAces()
        {
            List<CommonAce> entries;

            entries = new List<CommonAce>();
            if (_ItemAllowedAce != null) entries.Add(_ItemAllowedAce);
            if (_ItemDeniedAce != null) entries.Add(_ItemDeniedAce);

            return entries.ToArray();
        }
    }
}