using System.Security.AccessControl;
using System.Security.Principal;

namespace InfiniTec.Exchange.Security
{
    public partial class ItemAce
    {
        internal virtual void SetAccess(AccessMasks accessMask)
        {
            ItemAccessMasks mask;

            mask = DetermineItemAccessMask(accessMask);

            DangerousSetAccess(mask);
        }

        /// <summary>
        /// Translates the specified <see cref="AccessMasks"/> value to an
        /// <see cref="ItemAccessMasks"/> combination
        /// </summary>
        /// <param name="accessMask">The access mask to translate</param>
        /// <returns>The translated item access mask</returns>
        protected static ItemAccessMasks DetermineItemAccessMask(AccessMasks accessMask)
        {
            ItemAccessMasks mask;

            mask = ItemAccessMasks.None;

            if ((accessMask & AccessMasks.Read) == AccessMasks.Read) mask |= ItemAccessMasks.ReadItem;
            if ((accessMask & AccessMasks.EditOwn) == AccessMasks.EditOwn) mask |= ItemAccessMasks.EditOwn;
            if ((accessMask & AccessMasks.EditAny) == AccessMasks.EditAny) mask |= ItemAccessMasks.EditAny;
            if ((accessMask & AccessMasks.DeleteOwn) == AccessMasks.DeleteOwn) mask |= ItemAccessMasks.DeleteOwn;
            if ((accessMask & AccessMasks.DeleteAny) == AccessMasks.DeleteAny) mask |= ItemAccessMasks.DeleteAny;
            return mask;
        }

        /// <summary>
        /// Sets the item level permission on this ace (both allow and deny permissions)
        /// </summary>
        /// <param name="accessMask">The item-level access mask</param>
        protected internal virtual void DangerousSetAccess(ItemAccessMasks accessMask)
        {
            DangerousSetAccessInternal(accessMask, 0);
        }

        /// <summary>
        /// Sets the item level permission on this ace (both allow and deny permissions)
        /// </summary>
        /// <param name="accessMask">The item-level access mask</param>
        /// <param name="flags">A valid combination of <see cref="AceFlags"/></param>
        protected void DangerousSetAccessInternal(ItemAccessMasks accessMask, AceFlags flags)
        {
            DangerousSetAccessInternal((int) accessMask, ref _ItemAllowedAce,
                                       flags, true);

            // The Everyone SID only needs an allow entry
            if (!_IdentityReference.IsWellKnown(WellKnownSidType.WorldSid))
            {
                DangerousSetAccessInternal((int) (~accessMask & ItemAccessMasks.All),
                                           ref _ItemDeniedAce, flags, false);
            }
        }

        /// <summary>
        /// Sets the  low-level permission on this ace
        /// </summary>
        /// <param name="mask">The mask to set</param>
        /// <param name="ace">The <see cref="CommonAce"/> on which to set the ace</param>
        /// <param name="flags">The <see cref="AceFlags"/> to use</param>
        /// <param name="allowedAce">True, if the ace is an allowed ace. False, otherwise</param>
        protected void DangerousSetAccessInternal(int mask, ref CommonAce ace, AceFlags flags,
                                                  bool allowedAce)
        {
            if (ace == null)
            {
                ace = new CommonAce(flags, (allowedAce) ? AceQualifier.AccessAllowed : AceQualifier.AccessDenied, mask,
                                    IdentityReference, false, null);

                OnAceAcreated(new AceCreatedEventArgs(ace));
            }
            else
            {
                ace.AccessMask = mask;
                ace.AceFlags = flags & ~AceFlags.Inherited;
            }
        }
    }
}