using System.Collections.Generic;
using System.Security.AccessControl;
using System.Security.Principal;

namespace InfiniTec.Exchange.Security
{
    /// <summary>
    /// This class provides access to the security descriptors of items in an 
    /// Exchange store.
    /// </summary>
    public class ItemSecurity : ExchangeSecurity
    {
        /// <summary>
        /// Creates a new instance of this class using the specified settings.
        /// </summary>
        /// <param name="buffer">A buffer containing the binary security descriptor</param>
        /// <param name="offset">The offset where the binary descriptor begins</param>
        /// <param name="connection">The connection to use with the descriptor</param>
        public ItemSecurity(byte[] buffer, int offset, Connection connection) :
            base(buffer, offset, connection)
        {
        }

        /// <summary>
        /// Returns a list of <see cref="ItemAce"/>s which make up the 
        /// Dacl of the item
        /// </summary>
        /// <returns></returns>
        public IList<ItemAce> GetAccessControlEntries()
        {
            return GetAccessControlEntries<ItemAce>();
        }

        /// <summary>
        /// Removes the access for the specified identity
        /// </summary>
        /// <param name="identity">A <see cref="SecurityIdentifier"/> instance.</param>
        public override void RemoveAccess(SecurityIdentifier identity)
        {
            // The Anonymous and Everyone permission must always be set. The best one can do is to set the accesslevel
            // to None.
            if (identity.IsWellKnown(WellKnownSidType.WorldSid) || identity.IsWellKnown(WellKnownSidType.AnonymousSid))
            {
                SetAccess(identity, AccessMasks.None);
                return;
            }
            base.RemoveAccess(identity);
        }

        /// <summary>
        /// Sets the specified access mask for the identity.
        /// </summary>
        /// <param name="identity">A <see cref="SecurityIdentifier" /> instance</param>
        /// <param name="accessMask">The desired access mask</param>
        public void DangerousSetAccess(SecurityIdentifier identity, ItemAccessMasks accessMask)
        {
            ItemAce ace;

            if (!TryGetAce(identity, out ace))
            {
                ace = AddAce(identity);
            }
            ace.DangerousSetAccess(accessMask);
        }

        /// <summary>
        /// Updates the control flags for modification.
        /// </summary>
        protected override void UpdateControlFlagsForModification()
        {
            base.UpdateControlFlagsForModification();

            SetFlags(ControlFlags | ControlFlags.DiscretionaryAclProtected);
        }
    }
}