using System.Collections.Generic;
using System.Xml;
using System.Xml.XPath;

namespace InfiniTec.Exchange
{
    /// <summary>
    /// Represents a single-valued property of an Exchange item.
    /// </summary>
    /// <typeparam name="T">The type of the property</typeparam>
    public class Property<T> : Property
    {
        private T _Value;

        /// <summary>
        /// Gets or sets the value of the item
        /// </summary>
        public T Value
        {
            get { return _Value; }
            set
            {
                _Value = value;
                SetChanged();
            }
        }

        /// <summary>
        /// Loads the value of the property from the navigator
        /// </summary>
        /// <param name="navigator">A <see cref="XPathNavigator"/> containing the
        /// value.</param>
        internal override sealed void Load(XPathNavigator navigator)
        {
            Value = DecodeValue<T>(navigator.Value);
        }

        /// <summary>
        /// Creates a new instance of this class using the specified namespace and local
        /// name.
        /// </summary>
        /// <param name="ns">The namespace of the property</param>
        /// <param name="localName">The local name of the property</param>
        public Property(string ns, string localName) : base(ns, localName)
        {
        }

        /// <summary>
        /// Creates a new instance of this class using the specified namespace and local
        /// name.
        /// </summary>
        /// <param name="fullQualifiedName">The full qualified name of the property</param>
        /// <param name="value">The value of the property</param>
        public Property(string fullQualifiedName, T value) : this(fullQualifiedName)
        {
            Value = value;
            SetChanged();
        }

        /// <summary>
        /// Creates a new instance of this class using the specified namespace and local
        /// name.
        /// </summary>
        /// <param name="ns">The namespace of the property</param>
        /// <param name="localName">The local name of the property</param>
        /// <param name="value">The value of the property</param>
        public Property(string ns, string localName, T value) : this(ns, localName)
        {
            Value = value;
            SetChanged();
        }

        /// <summary>
        /// Creates a new instance of this class using the specified namespace and local
        /// name.
        /// </summary>
        /// <param name="fullQualifiedName">The full qualified name of the property</param>
        public Property(string fullQualifiedName) : base(fullQualifiedName)
        {
        }

        internal override sealed void SaveTo(XmlWriter writer)
        {
            // Workaround for a nasty Exchange bug. Properties starting with 0x are not xml-conform.
            // Additionally, Exchange needs an explicit namepsace prefix for these properties.
            // The name is replaced with a null prefix and a namespace. this is reversed before the request
            // is sent to the server.
            WriteEncodedStartElementForPropertyName(writer, this);

            writer.WriteAttributeString("dt", Namespaces.DataType, GetDataType<T>());
            writer.WriteString(EncodeValue(Value));

            writer.WriteEndElement();
        }

        internal override bool IsPropertyValueNull()
        {
            return EqualityComparer<T>.Default.Equals(Value, default(T));
        }
    }
}