using System.Collections.Generic;
using System.ComponentModel;
using InfiniTec.DirectoryServices;
using InfiniTec.Exchange.WebDav;
using InfiniTec.Threading;
using AsyncOperation=InfiniTec.Threading.AsyncOperation;

namespace InfiniTec.Exchange
{
    /// <summary>
    /// Provides access to the properties of an Active Directory user and to the 
    /// associated mailbox.
    /// </summary>
    public class User : IAddressListEntry
    {
        /// <summary>
        /// This event is fired, when an asynchronous refresh operation completes.
        /// </summary>
        public event AsyncCompletedEventHandler RefreshCompleted;

        private ActiveDirectoryUser _BaseItem;
        private Mailbox _Mailbox;
        private Connection _Connection;
        private AsyncOperationContext _AsyncOperationContext;

        /// <summary>
        /// Initializes a new instance of the <see cref="User"/> class.
        /// </summary>
        /// <param name="baseItem">The base item.</param>
        /// <param name="connection">The connection.</param>
        public User(ActiveDirectoryUser baseItem, Connection connection)
        {
            _BaseItem = baseItem;
            _Connection = connection;
            _AsyncOperationContext = new AsyncOperationContext();
        }

        /// <summary>
        /// Gets the mailbox object for this user.
        /// </summary>
        /// <value>The mailbox.</value>
        /// <remarks>
        /// The lookup of the mailbox url is a privileged operation.
        /// The current user must have read access to the section below
        /// CN=Microsoft Exchange,CN=Services,CN=Configuration,DC=primary_domain.
        /// </remarks>
        public Mailbox Mailbox
        {
            get
            {
                if (_Mailbox == null) throw new NotInitializedException(Strings.InstanceNotInitializedException);

                return _Mailbox;
            }
        }

        /// <summary>
        /// Gets or sets the displayname of the item.
        /// </summary>
        /// <value>The displayname of the item.</value>
        public string DisplayName
        {
            get { return _BaseItem.DisplayName; }
            set { _BaseItem.DisplayName = value; }
        }

        /// <summary>
        /// Gets or sets the country.
        /// </summary>
        /// <value>The country.</value>
        public string Country
        {
            get { return _BaseItem.Country; }
            set { _BaseItem.Country = value; }
        }

        /// <summary>
        /// Gets or sets the company.
        /// </summary>
        /// <value>The company.</value>
        public string Company
        {
            get { return _BaseItem.Company; }
            set { _BaseItem.Company = value; }
        }


        /// <summary>
        /// Gets the name of the account.
        /// </summary>
        /// <value>The name of the account.</value>
        public string NT4AccountName
        {
            get { return _BaseItem.NT4AccountName; }
        }

        /// <summary>
        /// Gets or sets the department.
        /// </summary>
        /// <value>The department.</value>
        public string Department
        {
            get { return _BaseItem.Department; }
            set { _BaseItem.Department = value; }
        }

        /// <summary>
        /// Gets or sets the description.
        /// </summary>
        /// <value>The description.</value>
        public string Description
        {
            get { return _BaseItem.Description; }
            set { _BaseItem.Description = value; }
        }

        /// <summary>
        /// Gets or sets the fax number.
        /// </summary>
        /// <value>The fax number.</value>
        public string FaxNumber
        {
            get { return _BaseItem.FaxNumber; }
            set { _BaseItem.FaxNumber = value; }
        }

        /// <summary>
        /// Gets or sets the first name.
        /// </summary>
        /// <value>The name of the first.</value>
        public string FirstName
        {
            get { return _BaseItem.FirstName; }
            set { _BaseItem.FirstName = value; }
        }

        /// <summary>
        /// Gets or sets the home phone.
        /// </summary>
        /// <value>The home phone.</value>
        public string HomePhone
        {
            get { return _BaseItem.HomePhone; }
            set { _BaseItem.HomePhone = value; }
        }

        /// <summary>
        /// Gets or sets the initials.
        /// </summary>
        /// <value>The initials.</value>
        public string Initials
        {
            get { return _BaseItem.Initials; }
            set { _BaseItem.Initials = value; }
        }

        /// <summary>
        /// Gets or sets the IP phone.
        /// </summary>
        /// <value>The IP phone.</value>
        public string IPPhone
        {
            get { return _BaseItem.IPPhone; }
            set { _BaseItem.IPPhone = value; }
        }

        /// <summary>
        /// Gets or sets the city.
        /// </summary>
        /// <value>The city.</value>
        public string City
        {
            get { return _BaseItem.City; }
            set { _BaseItem.City = value; }
        }

        /// <summary>
        /// Gets or sets the mail address.
        /// </summary>
        /// <value>The mail address.</value>
        public string MailAddress
        {
            get { return _BaseItem.MailAddress; }
            set { _BaseItem.MailAddress = value; }
        }

        /// <summary>
        /// Gets or sets the manager.
        /// </summary>
        /// <value>The manager.</value>
        public string Manager
        {
            get { return _BaseItem.Manager; }
            set { _BaseItem.Manager = value; }
        }

        /// <summary>
        /// Gets or sets the mobile phone.
        /// </summary>
        /// <value>The mobile phone.</value>
        public string MobilePhone
        {
            get { return _BaseItem.MobilePhone; }
            set { _BaseItem.MobilePhone = value; }
        }

        /// <summary>
        /// Gets the other fax numbers.
        /// </summary>
        /// <value>The other fax numbers.</value>
        public IList<string> OtherFaxNumbers
        {
            get { return _BaseItem.OtherFaxNumbers; }
        }

        /// <summary>
        /// Gets the other home numbers.
        /// </summary>
        /// <value>The other home numbers.</value>
        public IList<string> OtherHomeNumbers
        {
            get { return _BaseItem.OtherHomeNumbers; }
        }

        /// <summary>
        /// Gets the other IP phone numbers.
        /// </summary>
        /// <value>The other IP phone numbers.</value>
        public IList<string> OtherIPPhoneNumbers
        {
            get { return _BaseItem.OtherIPPhoneNumbers; }
        }

        /// <summary>
        /// Gets the other mobile phone numbers.
        /// </summary>
        /// <value>The other mobile phone numbers.</value>
        public IList<string> OtherMobilePhoneNumbers
        {
            get { return _BaseItem.OtherMobilePhoneNumbers; }
        }

        /// <summary>
        /// Gets the other pager numbers.
        /// </summary>
        /// <value>The other pager numbers.</value>
        public IList<string> OtherPagerNumbers
        {
            get { return _BaseItem.OtherPagerNumbers; }
        }

        /// <summary>
        /// Gets the other telephone numbers.
        /// </summary>
        /// <value>The other telephone numbers.</value>
        public IList<string> OtherTelephoneNumbers
        {
            get { return _BaseItem.OtherTelephoneNumbers; }
        }

        /// <summary>
        /// Gets or sets the pager number.
        /// </summary>
        /// <value>The pager number.</value>
        public string PagerNumber
        {
            get { return _BaseItem.PagerNumber; }
            set { _BaseItem.PagerNumber = value; }
        }

        /// <summary>
        /// Gets or sets the office.
        /// </summary>
        /// <value>The office.</value>
        public string Office
        {
            get { return _BaseItem.Office; }
            set { _BaseItem.Office = value; }
        }

        /// <summary>
        /// Gets or sets the postal code.
        /// </summary>
        /// <value>The postal code.</value>
        public string PostalCode
        {
            get { return _BaseItem.PostalCode; }
            set { _BaseItem.PostalCode = value; }
        }

        /// <summary>
        /// Gets or sets the last name
        /// </summary>
        /// <value>The name of the last.</value>
        public string LastName
        {
            get { return _BaseItem.LastName; }
            set { _BaseItem.LastName = value; }
        }

        /// <summary>
        /// Gets or sets the post office box.
        /// </summary>
        /// <value>The post office box.</value>
        public string PostOfficeBox
        {
            get { return _BaseItem.PostOfficeBox; }
            set { _BaseItem.PostOfficeBox = value; }
        }

        /// <summary>
        /// Gets or sets the state.
        /// </summary>
        /// <value>The state.</value>
        public string State
        {
            get { return _BaseItem.State; }
            set { _BaseItem.State = value; }
        }

        /// <summary>
        /// Gets or sets the street address.
        /// </summary>
        /// <value>The street address.</value>
        public string StreetAddress
        {
            get { return _BaseItem.StreetAddress; }
            set { _BaseItem.StreetAddress = value; }
        }

        /// <summary>
        /// Gets or sets the telephone number.
        /// </summary>
        /// <value>The telephone number.</value>
        public string TelephoneNumber
        {
            get { return _BaseItem.TelephoneNumber; }
            set { _BaseItem.TelephoneNumber = value; }
        }

        /// <summary>
        /// Gets or sets the title.
        /// </summary>
        /// <value>The title.</value>
        public string Title
        {
            get { return _BaseItem.Title; }
            set { _BaseItem.Title = value; }
        }

        /// <summary>
        /// Gets or sets the other web page.
        /// </summary>
        /// <value>The other web page.</value>
        public string OtherWebPage
        {
            get { return _BaseItem.OtherWebPage; }
            set { _BaseItem.OtherWebPage = value; }
        }

        /// <summary>
        /// Gets or sets the user principal name.
        /// </summary>
        /// <value>The name of the user principal.</value>
        public string UserPrincipalName
        {
            get { return _BaseItem.UserPrincipalName; }
            set { _BaseItem.UserPrincipalName = value; }
        }

        /// <summary>
        /// Gets or sets the web page.
        /// </summary>
        /// <value>The web page.</value>
        public string WebPage
        {
            get { return _BaseItem.WebPage; }
            set { _BaseItem.WebPage = value; }
        }

        #region IAddressListEntry Members

        string IAddressListEntry.Name
        {
            get { return !string.IsNullOrEmpty(_BaseItem.DisplayName) ? _BaseItem.DisplayName : _BaseItem.NT4AccountName; }
        }

        #endregion

        /// <summary>
        /// Synchronously refreshes the properties of the curren item.
        /// </summary>
        public void Refresh()
        {
            using (AsyncOperation operation = GetRefreshOperation())
            {
                operation.Run();
                operation.WaitForCompletion();
            }
        }

        /// <summary>
        /// Raises the <see cref="RefreshCompleted"/> event.
        /// </summary>
        /// <param name="e">The <see cref="System.ComponentModel.AsyncCompletedEventArgs"/> instance containing the event data.</param>
        protected virtual void OnRefreshCompleted(AsyncCompletedEventArgs e)
        {
            _AsyncOperationContext.UnregisterPendingAsyncOperation();
            if (RefreshCompleted != null) RefreshCompleted(this, e);
        }


        private void RefreshOperation_Completed(object sender, CompletedEventArgs e)
        {
            OnRefreshCompleted(new AsyncCompletedEventArgs(null, e.Status == OperationStatus.Aborted, null));
        }

        private void RefreshOperation_Error(object sender, ExceptionEventArgs e)
        {
            OnRefreshCompleted(new AsyncCompletedEventArgs(e.Exception, false, null));
        }

        /// <summary>
        /// Asynchronously refreshes the properties of the curren item.
        /// </summary>
        public void RefreshAsync()
        {
            AsyncOperation operation;

            _AsyncOperationContext.InitiatePendingAsyncOperation();

            operation = GetRefreshOperation();

            _AsyncOperationContext.RegisterPendingAsyncOperation(operation);

            operation.Completed += RefreshOperation_Completed;
            operation.Error += RefreshOperation_Error;
            operation.Run();
        }

        /// <summary>
        /// Gets the async operation operation used to refresh the item.
        /// </summary>
        /// <returns>An async operation which can be used to refresh the item.</returns>
        public AsyncOperation GetRefreshOperation()
        {
            return new AsyncOperation(RefreshInternal());
        }

        private IEnumerable<OperationAction> RefreshInternal()
        {
            using (AsyncOperation operation = _BaseItem.GetRefreshOperation())
            {
                yield return new NestedOperation(operation, false);
            }
            using (AsyncOperation<string> operation =
                    _Connection.ActiveDirectoryResolver.GetMailboxResolveOperation(_BaseItem))
            {
                yield return new NestedOperation(operation, true);
                _Mailbox = new Mailbox(operation.StatusInformation, _Connection);
            }
        }
    }
}