using System;
using System.Collections.Generic;
using System.ComponentModel;
using InfiniTec.DirectoryServices;
using InfiniTec.Threading;
using AsyncOperation=InfiniTec.Threading.AsyncOperation;

namespace InfiniTec.Exchange
{
    /// <summary>
    /// Resolves incomplete username to <see cref="User"/> objects using 
    /// ambiguous name resolution. Additionally, initials are also supported.
    /// </summary>
    public class UserResolver
    {
        /// <summary>
        /// Fired, when an asynchronous resolve operation completes.
        /// </summary>
        public event EventHandler ResolveCompleted;

        private AsyncOperationContext _AsyncOperationContext;
        private Connection _Connection;
        private IList<User> _SearchResult;
        private SearchMode _SearchMode;
        private string _SearchRoot;
        private SearchToken _SearchToken;
        private string _Filter;

        /// <summary>
        /// Gets or sets the <see cref="SearchMode"/>.
        /// </summary>
        /// <value>The search mode.</value>
        public SearchMode SearchMode
        {
            get { return _SearchMode; }
            set { _SearchMode = value; }
        }

        /// <summary>
        /// Gets or sets the filter to use for the search operation.
        /// </summary>
        /// <value>The filter.</value>
        public string Filter
        {
            get { return _Filter; }
            set { _Filter = value; }
        }

        /// <summary>
        /// Gets the search result.
        /// </summary>
        /// <value>The search result.</value>
        public IList<User> SearchResult
        {
            get
            {
                return _SearchResult;
            }
        }

        /// <summary>
        /// Gets or sets the search root. Leave blank to search the entire forest.
        /// </summary>
        /// <value>The search root.</value>
        public string SearchRoot
        {
            get { return _SearchRoot; }
            set { _SearchRoot = value; }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="UserResolver"/> class.
        /// </summary>
        /// <param name="connection">The connection.</param>
        public UserResolver(Connection connection)
        {
            _Connection = connection;
            _AsyncOperationContext = new AsyncOperationContext();
        }

        /// <summary>
        /// Synchronously Resolvees the properties of the curren item.
        /// </summary>
        public void Resolve()
        {
            using (AsyncOperation operation = GetResolveOperation())
            {
                operation.Run();
                operation.WaitForCompletion();
            }
        }

        /// <summary>
        /// Raises the <see cref="ResolveCompleted"/> event.
        /// </summary>
        /// <param name="e">The <see cref="System.ComponentModel.AsyncCompletedEventArgs"/> instance containing the event data.</param>
        protected virtual void OnResolveCompleted(AsyncCompletedEventArgs e)
        {
            _AsyncOperationContext.UnregisterPendingAsyncOperation();
            if (ResolveCompleted != null) ResolveCompleted(this, e);
        }

        private void ResolveOperation_Completed(object sender, CompletedEventArgs e)
        {
            OnResolveCompleted(new AsyncCompletedEventArgs(null, e.Status == OperationStatus.Aborted, null));
        }

        private void ResolveOperation_Error(object sender, ExceptionEventArgs e)
        {
            OnResolveCompleted(new AsyncCompletedEventArgs(e.Exception, false, null));
        }

        /// <summary>
        /// Asynchronously resolves the properties of the curren item.
        /// </summary>
        public void ResolveAsync()
        {
            AsyncOperation operation;

            _AsyncOperationContext.InitiatePendingAsyncOperation();

            operation = GetResolveOperation();

            _AsyncOperationContext.RegisterPendingAsyncOperation(operation);

            operation.Completed += ResolveOperation_Completed;
            operation.Error += ResolveOperation_Error;
            operation.Run();
        }

        /// <summary>
        /// Gets the async operation operation used to Resolve the item.
        /// </summary>
        /// <returns>An async operation which can be used to Resolve the item.</returns>
        public AsyncOperation GetResolveOperation()
        {
            return new AsyncOperation(ResolveInternal());
        }

        private IEnumerable<OperationAction> ResolveInternal()
        {
            PrincipalResolver resolver;

            if (_SearchMode == SearchMode.FindPage && _SearchToken != null)
            {
                resolver = new PrincipalResolver(_SearchToken, _Connection.ActiveDirectoryConnection);
            }
            else
            {
                _SearchResult = new List<User>();

                resolver = new PrincipalResolver(SearchRoot, _Connection.ActiveDirectoryConnection);
                resolver.FilterType = FilterType.Name;
                resolver.ResolveTypes = ResolveTypes.User;
                resolver.SearchMode = _SearchMode;
                resolver.Filter = _Filter;
            }

            using (AsyncOperation op = resolver.GetFindOperation())
            {
                yield return new NestedOperation(op, false);
            }

            if (SearchMode == SearchMode.FindPage)
            {
                _SearchToken = resolver.SearchToken;
            }

            foreach (ActiveDirectoryUser entry in resolver.SearchResult)
            {
                _SearchResult.Add(new User(entry, _Connection));
            }
        }
    }
}