using System;
using System.Collections.Generic;
using System.Runtime.CompilerServices;

namespace InfiniTec.Exchange.WebDav
{
    /// <summary>
    /// this class defines various properties used on Calendar items.
    /// </summary>
    [SchemaClass(typeof (WellknownProperties.Appointment))]
    public class CalendarItem : Item
    {
        private static string[] _RequiredFields;

        private ChangeNotifyCollection<EmailAddress> _RequiredAttendees;
        private ChangeNotifyCollection<EmailAddress> _OptionalAttendees;
        private ChangeNotifyCollection<EmailAddress> _Resources;
        private bool _PropertiesLoaded;

        /// <summary>
        /// Initializes the <see cref="CalendarItem"/> class.
        /// </summary>
        static CalendarItem()
        {
            _RequiredFields = SchemaHelper.ExtractFieldValues(typeof (WellknownProperties.Appointment));
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="CalendarItem"/> class.
        /// </summary>
        public CalendarItem()
        {
            Initialize();
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="CalendarItem"/> class.
        /// </summary>
        /// <param name="path">The path of the item.</param>
        /// <param name="connection">The connection.</param>
        public CalendarItem(string path, Connection connection)
            : base(path, connection)
        {
            Initialize();
        }

        private void Initialize()
        {
            AddRequiredFields(_RequiredFields);
        }

        [MethodImpl(MethodImplOptions.Synchronized)]
        private void EnsureAttendeeProperties()
        {
            if (_PropertiesLoaded) return;

            LoadAddressList(out _RequiredAttendees, WellknownProperties.Message.To);
            LoadAddressList(out _OptionalAttendees, WellknownProperties.Message.Cc);
            LoadAddressList(out _Resources, WellknownProperties.Message.Bcc);

            _PropertiesLoaded = true;
        }

        /// <summary>
        /// This method is called before the <see cref="Item.Refresh"/> or <see cref="Item.RefreshAsync"/>
        /// operation is called.
        /// </summary>
        protected override void PreRefresh()
        {
            _PropertiesLoaded = false;
            base.PreRefresh();
        }

        /// <summary>
        /// This method is called before the <see cref="Item.Save"/> or <see cref="Item.SaveAsync"/>
        /// operation is called.
        /// </summary>
        protected override void PreSave()
        {
            base.PreSave();

            CommitAddressList(_RequiredAttendees, WellknownProperties.Message.To);
            CommitAddressList(_OptionalAttendees, WellknownProperties.Message.Cc);
            CommitAddressList(_Resources, WellknownProperties.Message.Bcc);
        }

        /// <summary>
        /// Gets the default content class for this class.
        /// </summary>
        /// <param name="outlookMessageClass">The outlook message class (e.g. IPM.NOTE).</param>
        /// <param name="contentClass">The WebDAV content class (e.g. urn:content-classes:message).</param>
        protected override void GetDefaultContentClass(out string outlookMessageClass, out string contentClass)
        {
            outlookMessageClass = "IPM.Appointment";
            contentClass = "urn:content-classes:appointment";
        }

        /// <summary>
        /// Gets or sets the start date/time of the item.
        /// </summary>
        /// <value>The start date/time of the item.</value>
        public DateTime? Start
        {
            get { return GetNullablePropertyValue<DateTime>(WellknownProperties.Appointment.Starttime); }
            set { SetNullablePropertyValue(WellknownProperties.Appointment.Starttime, value); }
        }

        /// <summary>
        /// Gets or sets the end date/time of the item.
        /// </summary>
        /// <value>The end.</value>
        public DateTime? End
        {
            get { return GetNullablePropertyValue<DateTime>(WellknownProperties.Appointment.EndTime); }
            set { SetNullablePropertyValue(WellknownProperties.Appointment.EndTime, value); }
        }

        /// <summary>
        /// Gets or sets a value indicating whether this instance is an all day event.
        /// </summary>
        /// <value>
        /// 	<see langword="true"/> if this instance is an all day event; otherwise, <see langword="false"/>.
        /// </value>
        public bool? IsAllDayEvent
        {
            get { return GetNullablePropertyValue<bool>(WellknownProperties.Appointment.IsAllDavEvent); }
            set { SetNullablePropertyValue(WellknownProperties.Appointment.IsAllDavEvent, value); }
        }

        /// <summary>
        /// Gets or sets the free busy status.
        /// </summary>
        /// <value>The free busy status.</value>
        public FreeBusyType? FreeBusyStatus
        {
            get { return (FreeBusyType?) GetNullablePropertyValue<int>(WellknownProperties.Appointment.FreeBusyStatus); }
            set { SetNullablePropertyValue(WellknownProperties.Appointment.IsAllDavEvent, (int?) value); }
        }

        /// <summary>
        /// Gets or sets the location.
        /// </summary>
        /// <value>The location.</value>
        public string Location
        {
            get { return GetPropertyValue<string>(WellknownProperties.Appointment.Location); }
            set { SetPropertyValue(WellknownProperties.Appointment.Location, value); }
        }

        /// <summary>
        /// Gets or sets the when.
        /// </summary>
        /// <value>The when.</value>
        public string When
        {
            get { return GetPropertyValue<string>(WellknownProperties.Appointment.When); }
            set { SetPropertyValue(WellknownProperties.Appointment.When, value); }
        }

        /// <summary>
        /// Gets a value indicating whether this instance is meeting.
        /// </summary>
        /// <value>
        /// 	<see langword="true"/> if this instance is meeting; otherwise, <see langword="false"/>.
        /// </value>
        public bool? IsMeeting
        {
            get
            {
                return
                    CheckFlagValue(GetNullablePropertyValue<int>(WellknownProperties.Appointment.AppointmentState),
                                   (int) AppointmentState.Meeting);
            }
        }

        /// <summary>
        /// Gets a value indicating whether this instance is cancelled.
        /// </summary>
        /// <value>
        /// 	<see langword="true"/> if this instance is cancelled; otherwise, <see langword="false"/>.
        /// </value>
        public bool? IsCancelled
        {
            get
            {
                return
                    CheckFlagValue(GetNullablePropertyValue<int>(WellknownProperties.Appointment.AppointmentState),
                                   (int) AppointmentState.Cancelled);
            }
        }

        /// <summary>
        /// Gets a value indicating whether this instance is recurring.
        /// </summary>
        /// <value>
        /// 	<see langword="true"/> if this instance is recurring; otherwise, <see langword="false"/>.
        /// </value>
        public bool? IsRecurring
        {
            get { return GetNullablePropertyValue<bool>(WellknownProperties.Appointment.IsRecurring); }
        }

        /// <summary>
        /// Gets a value indicating whether this meeting request was sent.
        /// </summary>
        /// <value>
        /// 	<see langword="true"/> if [meeting request was sent]; otherwise, <see langword="false"/>.
        /// </value>
        public bool? MeetingRequestWasSent
        {
            get { return GetNullablePropertyValue<bool>(WellknownProperties.Appointment.MeetingRequestWasSent); }
        }

        /// <summary>
        /// Gets a value indicating whether a response has been requested.
        /// </summary>
        /// <value>
        /// 	<see langword="true"/> if a has been response requested; otherwise, <see langword="false"/>.
        /// </value>
        public bool? IsResponseRequested
        {
            get { return GetNullablePropertyValue<bool>(WellknownProperties.Appointment.IsResponseRequested); }
        }

        /// <summary>
        /// Gets the type of the calendar item.
        /// </summary>
        /// <value>The type of the calendar item.</value>
        public CalendarItemType? CalendarItemType
        {
            get { return (CalendarItemType) GetNullablePropertyValue<int>(WellknownProperties.Appointment.InstanceType); }
        }

        /// <summary>
        /// Gets the type of the response.
        /// </summary>
        /// <value>The type of the response.</value>
        public ResponseType? ResponseType
        {
            get { return (ResponseType?) GetNullablePropertyValue<int>(WellknownProperties.Appointment.ResponseType); }
        }

        /// <summary>
        /// Gets or sets the organizer.
        /// </summary>
        /// <value>The organizer.</value>
        public EmailAddress Organizer
        {
            get
            {
                return new EmailAddress(
                    GetPropertyValue<string>(WellknownProperties.Appointment.SentDisplayName),
                    GetPropertyValue<string>(WellknownProperties.Appointment.SentEmailAddress),
                    GetPropertyValue<string>(WellknownProperties.Appointment.SentEmailAddressType));
            }
            set
            {
                SetPropertyValue(WellknownProperties.Appointment.SentEmailAddress,
                                 (value != null) ? value.Address : null);
                SetPropertyValue(WellknownProperties.Appointment.SentDisplayName,
                                 (value != null) ? value.DisplayName : null);
                SetPropertyValue(WellknownProperties.Appointment.SentEmailAddressType,
                                 (value != null) ? value.AddressType : null);
            }
        }

        /// <summary>
        /// Gets the required attendees.
        /// </summary>
        /// <value>The required attendees.</value>
        public IList<EmailAddress> RequiredAttendees
        {
            get
            {
                EnsureAttendeeProperties();
                return _RequiredAttendees;
            }
        }

        /// <summary>
        /// Gets the optional attendees.
        /// </summary>
        /// <value>The optional attendees.</value>
        public IList<EmailAddress> OptionalAttendees
        {
            get
            {
                EnsureAttendeeProperties();
                return _OptionalAttendees;
            }
        }

        /// <summary>
        /// Gets the resources.
        /// </summary>
        /// <value>The resources.</value>
        public IList<EmailAddress> Resources
        {
            get
            {
                EnsureAttendeeProperties();
                return _Resources;
            }
        }

        /// <summary>
        /// Gets or sets the duration.
        /// </summary>
        /// <value>The duration.</value>
        public TimeSpan? Duration
        {
            get
            {
                int? result;

                result = GetNullablePropertyValue<int>(WellknownProperties.Appointment.Duration);
                return (result != null) ? (TimeSpan?) TimeSpan.FromSeconds(result.Value) : null;
            }
            set
            {
                SetNullablePropertyValue(WellknownProperties.Appointment.Duration,
                                         (value == null) ? null : (int?) value.Value.TotalSeconds);
            }
        }

        /// <summary>
        /// Gets or sets the time zone.
        /// </summary>
        /// <value>The time zone.</value>
        public string TimeZone
        {
            get { return GetPropertyValue<string>(WellknownProperties.Appointment.Timezone); }
            set { SetPropertyValue(WellknownProperties.Appointment.Timezone, value); }
        }

        /// <summary>
        /// Gets or sets the reply time.
        /// </summary>
        /// <value>The reply time.</value>
        public DateTime? ReplyTime
        {
            get { return GetNullablePropertyValue<DateTime>(WellknownProperties.Appointment.AppointmentReplyTime); }
            set { SetNullablePropertyValue(WellknownProperties.Appointment.AppointmentReplyTime, value); }
        }

        /// <summary>
        /// Gets or sets the sequence number.
        /// </summary>
        /// <value>The sequence number.</value>
        public int? SequenceNumber
        {
            get { return GetNullablePropertyValue<int>(WellknownProperties.Appointment.AppointmentSequenceNumber); }
            set { SetNullablePropertyValue(WellknownProperties.Appointment.AppointmentSequenceNumber, value); }
        }

        /// <summary>
        /// Gets or sets the state of the meeting.
        /// </summary>
        /// <value>The state of the meeting.</value>
        public AppointmentState? State
        {
            get
            {
                return
                    (AppointmentState?) GetNullablePropertyValue<int>(WellknownProperties.Appointment.AppointmentState);
            }
            set { SetNullablePropertyValue<int>(WellknownProperties.Appointment.AppointmentState, (int) value); }
        }

        /// <summary>
        /// Gets or sets the type of the conference.
        /// </summary>
        /// <value>The type of the conference.</value>
        public int? ConferenceType
        {
            get { return GetNullablePropertyValue<int>(WellknownProperties.Appointment.ConferenceType); }
            set { SetNullablePropertyValue(WellknownProperties.Appointment.ConferenceType, value); }
        }

        /// <summary>
        /// Gets or sets a value indicating whether it's allowed to propose a new time.
        /// </summary>
        /// <value>
        /// 	<see langword="true"/> if the proposal of a new time is allowed; otherwise, <see langword="false"/>.
        /// </value>
        public bool? AllowNewTimeProposal
        {
            get
            {
                bool? result;

                result = GetNullablePropertyValue<bool>(WellknownProperties.Appointment.DisallowNewTimeProposal);
                if (!result.HasValue) return null;
                else return !result.Value;
            }
            set
            {
                SetNullablePropertyValue(WellknownProperties.Appointment.DisallowNewTimeProposal,
                                         (value.HasValue) ? (bool?) !value.Value : null);
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether this instance is online meeting.
        /// </summary>
        /// <value>
        /// 	<see langword="true"/> if this instance is online meeting; otherwise, <see langword="false"/>.
        /// </value>
        public bool? IsOnlineMeeting
        {
            get { return GetNullablePropertyValue<bool>(WellknownProperties.Appointment.IsOnlineMeeting); }
            set { SetNullablePropertyValue(WellknownProperties.Appointment.IsOnlineMeeting, value); }
        }

        /// <summary>
        /// Gets or sets the meeting workspace URL.
        /// </summary>
        /// <value>The meeting workspace URL.</value>
        public string MeetingWorkspaceUrl
        {
            get { return GetPropertyValue<string>(WellknownProperties.Appointment.MeetingWorkspaceUrl); }
            set { SetPropertyValue(WellknownProperties.Appointment.MeetingWorkspaceUrl, value); }
        }

        /// <summary>
        /// Gets or sets the net show URL.
        /// </summary>
        /// <value>The net show URL.</value>
        public string NetShowUrl
        {
            get { return GetPropertyValue<string>(WellknownProperties.Appointment.NetShowUrl); }
            set { SetPropertyValue(WellknownProperties.Appointment.NetShowUrl, value); }
        }
    }
}