using System;
using System.Collections.Generic;

namespace InfiniTec.Exchange.WebDav
{
    /// <summary>
    /// This class specifies various properties on an Exchange Contact.
    /// </summary>
    [SchemaClass(typeof (WellknownProperties.Contact))]
    public class ContactItem : Item
    {
        private static string[] _RequiredFields;

        private ChangeNotifyDictionary<EmailAddressIdentifier, ContactEmailAddress> _EmailAddresses;
        private ChangeNotifyDictionary<PhysicalAddressType, PhysicalAddress> _PhysicalAddresses;
        private PhoneNumberDictionary _TelephoneNumbers;
        private CompleteName _CompleteName;

        /// <summary>
        /// Initializes the <see cref="ContactItem"/> class.
        /// </summary>
        static ContactItem()
        {
            _RequiredFields = SchemaHelper.ExtractFieldValues(typeof (WellknownProperties.Contact));
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ContactItem"/> class.
        /// </summary>
        public ContactItem() : base()
        {
            Initialize();
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ContactItem"/> class.
        /// </summary>
        /// <param name="path">The path of the contact item.</param>
        /// <param name="connection">The connection.</param>
        public ContactItem(string path, Connection connection) : base(path, connection)
        {
            Initialize();
        }

        private void Initialize()
        {
            AddRequiredFields(_RequiredFields);
            _CompleteName = new CompleteName(this);
        }

        /// <summary>
        /// Gets the default content class for this class.
        /// </summary>
        /// <param name="outlookMessageClass">The outlook message class (e.g. IPM.NOTE).</param>
        /// <param name="contentClass">The WebDAV content class (e.g. urn:content-classes:message).</param>
        protected override void GetDefaultContentClass(out string outlookMessageClass, out string contentClass)
        {
            outlookMessageClass = "IPM.Contact";
            contentClass = "urn:content-classes:person";
        }

        /// <summary>
        /// This method is called before the <see cref="M:Refresh"/> or <see cref="M:RefreshAsync"/>
        /// operation is called.
        /// </summary>
        protected override void PreRefresh()
        {
            _EmailAddresses = null;
            _PhysicalAddresses = null;
            _TelephoneNumbers = null;

            base.PreRefresh();
        }

        /// <summary>
        /// This method is called before the <see cref="M:Save"/> or <see cref="M:SaveAsync"/>
        /// operation is called.
        /// </summary>
        protected override void PreSave()
        {
            EnsureEmailAddresses();
            EnsurePhysicalAddresses();

            _EmailAddresses.Save();
            _PhysicalAddresses.Save();

            base.PreSave();
        }


        private void EnsureEmailAddresses()
        {
            if (_EmailAddresses != null) return;

            _EmailAddresses = new ChangeNotifyDictionary<EmailAddressIdentifier, ContactEmailAddress>(this);
            if (Properties.Contains(WellknownProperties.Contact.Email1EmailAddress))
            {
                _EmailAddresses.Add(new ContactEmailAddress(EmailAddressIdentifier.EmailAddress1, this));
            }
            if (Properties.Contains(WellknownProperties.Contact.Email2EmailAddress))
            {
                _EmailAddresses.Add(new ContactEmailAddress(EmailAddressIdentifier.EmailAddress2, this));
            }
            if (Properties.Contains(WellknownProperties.Contact.Email3EmailAddress))
            {
                _EmailAddresses.Add(new ContactEmailAddress(EmailAddressIdentifier.EmailAddress3, this));
            }
        }

        private void EnsurePhysicalAddresses()
        {
            if (_PhysicalAddresses != null) return;

            _PhysicalAddresses = new ChangeNotifyDictionary<PhysicalAddressType, PhysicalAddress>(this);

            _PhysicalAddresses.Add(new PhysicalAddress(PhysicalAddressType.Business, this));
            _PhysicalAddresses.Add(new PhysicalAddress(PhysicalAddressType.Home, this));
            _PhysicalAddresses.Add(new PhysicalAddress(PhysicalAddressType.Other, this));
        }

        private void EnsureTelephoneNumbers()
        {
            if (_TelephoneNumbers != null) return;

            _TelephoneNumbers = new PhoneNumberDictionary(this);
        }

        /// <summary>
        /// Gets or sets the file as string.
        /// </summary>
        /// <value>The file as string.</value>
        public string FileAs
        {
            get { return GetPropertyValue<string>(WellknownProperties.Contact.FileAs); }
            set { SetPropertyValue<string>(WellknownProperties.Contact.FileAs, value); }
        }

        /// <summary>
        /// Gets or sets the file as mapping.
        /// </summary>
        /// <value>The file as mapping.</value>
        public FileAsMapping? FileAsMapping
        {
            get { return (FileAsMapping?) GetNullablePropertyValue<int>(WellknownProperties.Contact.FileAsMapping); }
            set { SetNullablePropertyValue<int>(WellknownProperties.Contact.FileAsMapping, (int?) value); }
        }

        /// <summary>
        /// Gets or sets the display name.
        /// </summary>
        /// <value>The display name.</value>
        public string DisplayName
        {
            get { return GetPropertyValue<string>(WellknownProperties.Contact.DisplayName); }
            set { SetPropertyValue<string>(WellknownProperties.Contact.DisplayName, value); }
        }

        /// <summary>
        /// Gets or sets the given name.
        /// </summary>
        /// <value>The given name.</value>
        public string GivenName
        {
            get { return GetPropertyValue<string>(WellknownProperties.Contact.GivenName); }
            set { SetPropertyValue<string>(WellknownProperties.Contact.GivenName, value); }
        }

        /// <summary>
        /// Gets or sets the initials.
        /// </summary>
        /// <value>The initials.</value>
        public string Initials
        {
            get { return GetPropertyValue<string>(WellknownProperties.Contact.Initials); }
            set { SetPropertyValue<string>(WellknownProperties.Contact.Initials, value); }
        }

        /// <summary>
        /// Gets or sets the middle name.
        /// </summary>
        /// <value>The middle name.</value>
        public string MiddleName
        {
            get { return GetPropertyValue<string>(WellknownProperties.Contact.MiddleName); }
            set { SetPropertyValue<string>(WellknownProperties.Contact.MiddleName, value); }
        }

        /// <summary>
        /// Gets or sets the nickname.
        /// </summary>
        /// <value>The nickname.</value>
        public string Nickname
        {
            get { return GetPropertyValue<string>(WellknownProperties.Contact.NickName); }
            set { SetPropertyValue<string>(WellknownProperties.Contact.NickName, value); }
        }

        /// <summary>
        /// Gets the complete name.
        /// </summary>
        /// <value>The complete name.</value>
        public CompleteName CompleteName
        {
            get { return _CompleteName; }
        }

        /// <summary>
        /// Gets or sets the name of the company.
        /// </summary>
        /// <value>The name of the company.</value>
        public string CompanyName
        {
            get { return GetPropertyValue<string>(WellknownProperties.Contact.CompanyName); }
            set { SetPropertyValue<string>(WellknownProperties.Contact.CompanyName, value); }
        }

        /// <summary>
        /// Gets the email addresses.
        /// </summary>
        /// <value>The email addresses.</value>
        public IDictionary<EmailAddressIdentifier, ContactEmailAddress> EmailAddresses
        {
            get
            {
                EnsureEmailAddresses();
                return _EmailAddresses;
            }
        }

        /// <summary>
        /// Gets the physical addresses.
        /// </summary>
        /// <value>The physical addresses.</value>
        public IDictionary<PhysicalAddressType, PhysicalAddress> PhysicalAddresses
        {
            get
            {
                EnsurePhysicalAddresses();
                return _PhysicalAddresses;
            }
        }

        /// <summary>
        /// Gets the phone numbers.
        /// </summary>
        /// <value>The phone numbers.</value>
        public IDictionary<PhoneNumberType, string> PhoneNumbers
        {
            get
            {
                EnsureTelephoneNumbers();
                return _TelephoneNumbers;
            }
        }

        /// <summary>
        /// Gets or sets the name of the assistant.
        /// </summary>
        /// <value>The name of the assistant.</value>
        public string AssistantName
        {
            get { return GetPropertyValue<string>(WellknownProperties.Contact.AssistantName); }
            set { SetPropertyValue<string>(WellknownProperties.Contact.AssistantName, value); }
        }

        /// <summary>
        /// Gets or sets the birthday.
        /// </summary>
        /// <value>The birthday.</value>
        public DateTime? Birthday
        {
            get { return GetNullablePropertyValue<DateTime>(WellknownProperties.Contact.Birthday); }
            set { SetNullablePropertyValue<DateTime>(WellknownProperties.Contact.Birthday, value); }
        }

        /// <summary>
        /// Gets or sets the business home page.
        /// </summary>
        /// <value>The business home page.</value>
        public string BusinessHomePage
        {
            get { return GetPropertyValue<string>(WellknownProperties.Contact.BusinessHomepage); }
            set { SetPropertyValue<string>(WellknownProperties.Contact.BusinessHomepage, value); }
        }

        /// <summary>
        /// Gets the children.
        /// </summary>
        /// <value>The children.</value>
        public IList<string> Children
        {
            get
            {
                EnsureMultiValuedProperty<string>(WellknownProperties.Contact.Children);

                return GetMultivaluedPropertyValue<string>(WellknownProperties.Contact.Children);
            }
        }

        /// <summary>
        /// Gets the companies.
        /// </summary>
        /// <value>The companies.</value>
        public IList<string> Companies
        {
            get
            {
                EnsureMultiValuedProperty<string>(WellknownProperties.Contact.Companies);
                return GetMultivaluedPropertyValue<string>(WellknownProperties.Contact.Companies);
            }
        }

        /// <summary>
        /// Gets the contact source.
        /// </summary>
        /// <value>The contact source.</value>
        public ContactSource ContactSource
        {
            get { return ContactSource.Store; }
        }

        /// <summary>
        /// Gets or sets the department.
        /// </summary>
        /// <value>The department.</value>
        public string Department
        {
            get { return GetPropertyValue<string>(WellknownProperties.Contact.Department); }
            set { SetPropertyValue<string>(WellknownProperties.Contact.Department, value); }
        }

        /// <summary>
        /// Gets or sets the generation.
        /// </summary>
        /// <value>The generation.</value>
        public string Generation
        {
            get { return GetPropertyValue<string>(WellknownProperties.Contact.Generation); }
            set { SetPropertyValue<string>(WellknownProperties.Contact.Generation, value); }
        }

        /// <summary>
        /// Gets or sets the IM addresses.
        /// </summary>
        /// <value>The IM addresses.</value>
        public string IMAddresses
        {
            get { return GetPropertyValue<string>(WellknownProperties.Contact.IMAddress); }
            set { SetPropertyValue<string>(WellknownProperties.Contact.IMAddress, value); }
        }

        /// <summary>
        /// Gets or sets the job title.
        /// </summary>
        /// <value>The job title.</value>
        public string JobTitle
        {
            get { return GetPropertyValue<string>(WellknownProperties.Contact.JobTitle); }
            set { SetPropertyValue<string>(WellknownProperties.Contact.JobTitle, value); }
        }

        /// <summary>
        /// Gets or sets the manager.
        /// </summary>
        /// <value>The manager.</value>
        public string Manager
        {
            get { return GetPropertyValue<string>(WellknownProperties.Contact.Manager); }
            set { SetPropertyValue<string>(WellknownProperties.Contact.Manager, value); }
        }

        /// <summary>
        /// Gets or sets the mileage.
        /// </summary>
        /// <value>The mileage.</value>
        public string Mileage
        {
            get { return GetPropertyValue<string>(WellknownProperties.Contact.Mileage); }
            set { SetPropertyValue<string>(WellknownProperties.Contact.Mileage, value); }
        }

        /// <summary>
        /// Gets or sets the office location.
        /// </summary>
        /// <value>The office location.</value>
        public string OfficeLocation
        {
            get { return GetPropertyValue<string>(WellknownProperties.Contact.OfficeLocation); }
            set { SetPropertyValue<string>(WellknownProperties.Contact.OfficeLocation, value); }
        }

        /// <summary>
        /// Gets or sets the index of the postal address.
        /// </summary>
        /// <value>The index of the postal address.</value>
        public PhysicalAddressIndex? PostalAddressIndex
        {
            get
            {
                return
                    (PhysicalAddressIndex?)
                    GetNullablePropertyValue<int>(WellknownProperties.Contact.PostalAddressIndex);
            }
            set { SetNullablePropertyValue<int>(WellknownProperties.Contact.PostalAddressIndex, (int) value); }
        }

        /// <summary>
        /// Gets or sets the profession.
        /// </summary>
        /// <value>The profession.</value>
        public string Profession
        {
            get { return GetPropertyValue<string>(WellknownProperties.Contact.Profession); }
            set { SetPropertyValue<string>(WellknownProperties.Contact.Profession, value); }
        }

        /// <summary>
        /// Gets or sets the name of the spouse.
        /// </summary>
        /// <value>The name of the spouse.</value>
        public string SpouseName
        {
            get { return GetPropertyValue<string>(WellknownProperties.Contact.SpouseName); }
            set { SetPropertyValue<string>(WellknownProperties.Contact.SpouseName, value); }
        }

        /// <summary>
        /// Gets or sets the surname.
        /// </summary>
        /// <value>The surname.</value>
        public string Surname
        {
            get { return GetPropertyValue<string>(WellknownProperties.Contact.Surname); }
            set { SetPropertyValue<string>(WellknownProperties.Contact.Surname, value); }
        }

        /// <summary>
        /// Gets or sets the wedding anniversary.
        /// </summary>
        /// <value>The wedding anniversary.</value>
        public DateTime? WeddingAnniversary
        {
            get { return GetNullablePropertyValue<DateTime>(WellknownProperties.Contact.WeddingAnniversary); }
            set { SetNullablePropertyValue<DateTime>(WellknownProperties.Contact.WeddingAnniversary, value); }
        }
    }
}