using System.Collections.Generic;
using InfiniTec.Exchange.Security;
using InfiniTec.Exchange.WebDav.Request;
using InfiniTec.Threading;

namespace InfiniTec.Exchange.WebDav
{
    /// <summary>
    /// This class represents a public or private folder on an Exchange store.
    /// </summary>
    public abstract class FolderBase : ItemBase
    {
        private bool _FolderNeedsCreation;

        /// <summary>
        /// True, if the folder is created when the <see cref="ItemBase.Save"/> 
        /// operation is called. False, otherwise.
        /// </summary>
        public bool FolderNeedsCreation
        {
            get { return _FolderNeedsCreation; }
        }

        /// <summary>
        /// Creates a new instance of this class using the specified
        /// path connection and (optional) properties.
        /// </summary>
        /// <param name="path">The path of the folder on the server.</param>
        /// <param name="connection">The <see cref="Connection"/> to use.</param>
        /// <param name="propertiesToLoad">The properties to retrieve. If omitted, all properties are retrieved.</param>
        protected FolderBase(string path, Connection connection, params string[] propertiesToLoad) :
            this(path, connection, false, propertiesToLoad)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="FolderBase"/> class.
        /// </summary>
        /// <param name="path">The path.</param>
        /// <param name="connection">The connection.</param>
        /// <param name="isNew">if set to <c>true</c> [is new].</param>
        /// <param name="propertiesToLoad">The properties to load.</param>
        protected FolderBase(string path, Connection connection, bool isNew, params string[] propertiesToLoad) :
            base(path, connection, propertiesToLoad)
        {
            _FolderNeedsCreation = isNew;
        }

        /// <summary>
        /// Returns the <see cref="FolderSecurity"/> of the folder.
        /// </summary>
        /// <returns>The security descriptor of the folder</returns>
        public FolderSecurity GetSecurity()
        {
            return new FolderSecurity(GetSecurityDescriptorProperty().Value, 8, Connection);
        }

        /// <summary>
        /// Replaces the current security descriptor with the specified one.
        /// </summary>
        /// <param name="security">The new security descriptor</param>
        public void SetSecurity(FolderSecurity security)
        {
            SetSecurityInternal(security);
        }

        /// <summary>
        /// Returns the save operation used to save the changes
        /// made to this instance.
        /// </summary>
        /// <returns></returns>
        public override AsyncOperation<RequestStatus> GetSaveOperation()
        {
            if (!FolderNeedsCreation) return base.GetSaveOperation();
            else return new AsyncOperation<RequestStatus>(SaveInternal());
        }

        private IEnumerable<OperationAction> SaveInternal()
        {
            using (AsyncOperation op = Methods.MkCol(Url, Properties))
            {
                yield return new NestedOperation(op, true);

                if (op.Status == OperationStatus.Completed) _FolderNeedsCreation = false;
            }
        }
    }
}