using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Runtime.CompilerServices;
using System.Text;
using System.Text.RegularExpressions;
using InfiniTec.Exchange.Security;
using InfiniTec.Exchange.WebDav.Request;
using InfiniTec.Threading;

namespace InfiniTec.Exchange.WebDav
{
    /// <summary>
    /// This class provides access to properties common to all Exchange items.
    /// </summary>
    [SchemaClass(typeof (WellknownProperties.Item))]
    public class Item
    {
        private static string[] _ItemFields;

        private static Regex _EmailAddressParser =
            new Regex(
                @"(\""
                (?<name>.*?)\"")?  # Parse the mail address, encapsulated in quotation marks
                \s*?
                (\<(?<address>.*?)\>) # Parse the email address",
                RegexOptions.Compiled | RegexOptions.CultureInvariant | RegexOptions.ExplicitCapture |
                RegexOptions.IgnorePatternWhitespace);

        private IList<string> _Categories;


        /// <summary>
        /// Returns the security descriptor for the item
        /// </summary>
        /// <returns>An <see cref="ItemSecurity"/> instance.</returns>
        public ItemSecurity GetSecurity()
        {
            return BaseItem.GetSecurity();
        }

        /// <summary>
        /// Replaces the current security descriptor with the specified descriptor.
        /// </summary>
        /// <param name="security">The new security descriptor</param>
        public void SetSecurity(ItemSecurity security)
        {
            BaseItem.SetSecurity(security);
        }

        /// <summary>
        /// Synchronously copies the current item to the specified destination.
        /// </summary>
        /// <param name="path">The destination path.</param>
        /// <param name="option">A valid combination of the <see cref="CopyOption"/> enumeration.</param>
        /// <param name="scope">A valid member of the <see cref="Depth"/> enumeration.</param>
        /// <returns>The copied item.</returns>
        public SimpleItem Copy(string path, CopyOption option, Depth scope)
        {
            return BaseItem.Copy(path, option, scope);
        }

        /// <summary>
        /// Asynchronously copies the current item to the destination.
        /// </summary>
        /// <param name="path">The destination path.</param>
        /// <param name="option">A valid combination of the <see cref="CopyOption"/> enumeration.</param>
        /// <param name="scope">A valid member of the <see cref="Depth"/> enumeration.</param>
        public void CopyAsync(string path, CopyOption option, Depth scope)
        {
            BaseItem.CopyAsync(path, option, scope);
        }

        /// <summary>
        /// Synchronously moves this item to the specified destination.
        /// </summary>
        /// <param name="path">The destination path.</param>
        /// <param name="option">A valid combination of the <see cref="CopyOption"/> enumeration.</param>
        /// <param name="scope">A valid member of the <see cref="Depth"/> enumeration.</param>
        public void Move(string path, CopyOption option, Depth scope)
        {
            BaseItem.Move(path, option, scope);
        }

        /// <summary>
        /// Asynchronously moves this item to the specified destination.
        /// </summary>
        /// <param name="path">The destination path.</param>
        /// <param name="option">A valid combination of the <see cref="CopyOption"/> enumeration.</param>
        /// <param name="scope">A valid member of the <see cref="Depth"/> enumeration.</param>
        public void MoveAsync(string path, CopyOption option, Depth scope)
        {
            BaseItem.MoveAsync(path, option, scope);
        }

        /// <summary>
        /// This event is called during an async delete operation.
        /// </summary>
        public event ProgressChangedEventHandler<RequestStatus> DeleteProgressChanged
        {
            add { BaseItem.DeleteProgressChanged += value; }
            remove { BaseItem.DeleteProgressChanged -= value; }
        }

        /// <summary>
        /// This event is called during an async copy operation.
        /// </summary>
        public event ProgressChangedEventHandler<RequestStatus> CopyProgressChanged
        {
            add { BaseItem.CopyProgressChanged += value; }
            remove { BaseItem.CopyProgressChanged -= value; }
        }

        /// <summary>
        /// This event is called after an async copy operation has completed.
        /// </summary>
        public event EventHandler<CopyCompletedEventArgs> CopyCompleted
        {
            add { BaseItem.CopyCompleted += value; }
            remove { BaseItem.CopyCompleted -= value; }
        }

        /// <summary>
        /// This event is called during an async move operation.
        /// </summary>
        public event ProgressChangedEventHandler<RequestStatus> MoveProgressChanged
        {
            add { BaseItem.MoveProgressChanged += value; }
            remove { BaseItem.MoveProgressChanged -= value; }
        }

        /// <summary>
        /// This event is called after an async move operation.
        /// </summary>
        public event AsyncCompletedEventHandler MoveCompleted
        {
            add { BaseItem.MoveCompleted += value; }
            remove { BaseItem.MoveCompleted -= value; }
        }

        /// <summary>
        /// This event is called when an async refresh operation has completed.
        /// </summary>
        public event AsyncCompletedEventHandler RefreshCompleted
        {
            add { BaseItem.RefreshCompleted += value; }
            remove { BaseItem.RefreshCompleted -= value; }
        }


        /// <summary>
        /// This event is called when an async save operation has completed.
        /// </summary>
        public event AsyncCompletedEventHandler SaveCompleted
        {
            add { BaseItem.SaveCompleted += value; }
            remove { BaseItem.SaveCompleted -= value; }
        }

        /// <summary>
        /// This event is called when an async delete operation has completed.
        /// </summary>
        public event AsyncCompletedEventHandler DeleteCompleted
        {
            add { BaseItem.DeleteCompleted += value; }
            remove { BaseItem.DeleteCompleted -= value; }
        }

        private SimpleItem _BaseItem;
        private List<string> _RequiredFields;
        private Body _Body;

        /// <summary>
        /// Gets or sets the base <see cref="WebDav.Item"/>.
        /// </summary>
        /// <value>The base item.</value>
        protected internal SimpleItem BaseItem
        {
            [DebuggerStepThrough]
            get { return _BaseItem; }
            [DebuggerStepThrough]
            set { _BaseItem = value; }
        }

        /// <summary>
        /// Sets the content class of the item.
        /// </summary>
        /// <param name="outlookClass">The Outlook class for the item (e.g. IPM.Note)</param>
        /// <param name="contentClass">The WebDAV class for item (e.g. urn:content-classes:message)</param>
        public void SetContentClass(string outlookClass, string contentClass)
        {
            ItemClass = outlookClass;
            ContentClass = contentClass;
        }

        /// <summary>
        /// Proovides access to all properties of the item.
        /// </summary>
        /// <value>The properties.</value>
        public PropertyCollection Properties
        {
            get { return BaseItem.Properties; }
        }

        /// <summary>
        /// Gets the <see cref="WebDav.AttachmentCollection"/> for this item.
        /// </summary>
        /// <value>The attachments.</value>
        public AttachmentCollection Attachments
        {
            get { return BaseItem.Attachments; }
        }


        /// <summary>
        /// This method is called before the <see cref="Refresh"/> or <see cref="RefreshAsync"/>
        /// operation is called.
        /// </summary>
        protected virtual void PreRefresh()
        {
            _Body = null;

            foreach (string property in _RequiredFields)
            {
                if (!BaseItem.PropertiesToLoad.Contains(property))
                    BaseItem.PropertiesToLoad.Add(property);
            }
        }

        /// <summary>
        /// This method is called before the <see cref="Save"/> or <see cref="SaveAsync"/>
        /// operation is called.
        /// </summary>
        protected virtual void PreSave()
        {
            string outlookMessageClass;
            string contentClass;
            MultiValuedProperty<string> categories;

            if (BaseItem == null)
                throw new InvalidOperationException("There is no exchange item associated with this instance");

            if (
                !(Properties.Contains(WellknownProperties.Item.ContentClass) &&
                  Properties.Contains(WellknownProperties.Item.ItemClass)))
            {
                GetDefaultContentClass(out outlookMessageClass, out contentClass);
                SetContentClass(outlookMessageClass, contentClass);
            }

            if (_Categories != null)
            {
                if (
                    !BaseItem.Properties.TryGetMultiValuedProperty(WellknownProperties.Item.Categories, out categories) &&
                    _Categories.Count > 0)
                {
                    categories = new MultiValuedProperty<string>(WellknownProperties.Item.Categories);
                    foreach (string category in _Categories)
                    {
                        categories.Values.Add(category);
                    }
                    BaseItem.Properties.Add(categories);
                }
            }
        }

        /// <summary>
        /// Synchronously refreshes the properties of the curren item.
        /// </summary>
        public void Refresh()
        {
            PreRefresh();
            BaseItem.Refresh();
        }

        /// <summary>
        /// Asynchronously refreshes the properties of the curren item.
        /// </summary>
        public void RefreshAsync()
        {
            PreRefresh();
            BaseItem.RefreshAsync();
        }

        /// <summary>
        /// Synchronously saves the properties of the curren item.
        /// </summary>
        public void Save()
        {
            PreSave();
            BaseItem.Save();
        }

        /// <summary>
        /// Asynchronously saves the properties of the curren item.
        /// </summary>
        public void SaveAsync()
        {
            PreSave();
            BaseItem.SaveAsync();
        }

        /// <summary>
        /// Synchronously deletes the properties of the curren item.
        /// </summary>
        public void Delete()
        {
            BaseItem.Delete();
        }

        /// <summary>
        /// Asynchronously deletes the properties of the curren item.
        /// </summary>
        public void DeleteAsync()
        {
            BaseItem.DeleteAsync();
        }

        /// <summary>
        /// Initializes this class
        /// </summary>
        static Item()
        {
            _ItemFields = SchemaHelper.ExtractFieldValues(typeof (WellknownProperties.Item));
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ItemBase"/> class.
        /// </summary>
        /// <param name="path">The path.</param>
        /// <param name="connection">The connection.</param>
        public Item(string path, Connection connection) : this()
        {
            BaseItem = new SimpleItem(path, connection);
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ItemBase"/> class.
        /// </summary>
        public Item()
        {
            _RequiredFields = new List<string>();
            AddRequiredFields(_ItemFields);
        }

        /// <summary>
        /// Adds the required fields which to the list of fields which are
        /// loaded even if no other fields are specified.
        /// </summary>
        /// <param name="fields">The fields.</param>
        protected void AddRequiredFields(IEnumerable<string> fields)
        {
            _RequiredFields.AddRange(fields);
        }

        /// <summary>
        /// Gets the default content class for this class.
        /// </summary>
        /// <param name="outlookMessageClass">The outlook message class (e.g. IPM.NOTE).</param>
        /// <param name="contentClass">The WebDAV content class (e.g. urn:content-classes:message).</param>
        protected virtual void GetDefaultContentClass(out string outlookMessageClass, out string contentClass)
        {
            outlookMessageClass = "IPM.Note";
            contentClass = "urn:content-classes:message";
        }

        /// <summary>
        /// Gets the property value.
        /// </summary>
        /// <param name="name">The name of the property.</param>
        /// <typeparam name="T">The type of the requested property.</typeparam>
        /// <returns>The value of the requested property. If not found, the default
        /// value for the specified type is returned.</returns>
        protected internal T GetPropertyValue<T>(string name)
        {
            return GetPropertyValue(name, default(T));
        }

        /// <summary>
        /// Gets the property value.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <typeparam name="T">The type of the requested property.</typeparam>
        /// <returns>The value of the requested property. If not found, the default
        /// value for the specified type is returned.</returns>
        protected internal T? GetNullablePropertyValue<T>(string name)
            where T : struct
        {
            Property<T> prop;

            if (!BaseItem.Properties.TryGetProperty(name, out prop)) return null;
            else return prop.Value;
        }

        /// <summary>
        /// Gets the multi-valued property value.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <typeparam name="T">The type of the requested property.</typeparam>
        /// <returns>The value of the requested property. If not found, the default
        /// value for the specified type is returned.</returns>
        protected internal IList<T> GetMultivaluedPropertyValue<T>(string name)
        {
            return GetMultivaluedPropertyValue<T>(name, null);
        }

        /// <summary>
        /// Gets the value of the property.
        /// </summary>
        /// <param name="name">The name of the property.</param>
        /// <param name="defaultValue">The default value for the property</param>
        /// <typeparam name="T">The type of the requested property.</typeparam>
        /// <returns>The value of the requested property. If not found, specified
        /// default value is returned.</returns>
        protected internal T GetPropertyValue<T>(string name, T defaultValue)
        {
            Property<T> prop;

            if (!BaseItem.Properties.TryGetProperty(name, out prop)) return defaultValue;
            else return prop.Value;
        }

        /// <summary>
        /// Gets the property value.
        /// </summary>
        /// <param name="name">The name of the property.</param>
        /// <param name="defaultValue">The default value for the property</param>
        /// <typeparam name="T">The type of the requested property.</typeparam>
        /// <returns>The value of the requested property. If not found, the specified default
        /// value is returned.</returns>
        protected internal IList<T> GetMultivaluedPropertyValue<T>(string name, T[] defaultValue)
        {
            MultiValuedProperty<T> prop;

            if (!BaseItem.Properties.TryGetMultiValuedProperty(name, out prop)) return defaultValue;
            else return prop.Values;
        }

        /// <summary>
        /// Gets the property value.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <param name="value">The vlaue of the value.</param>
        /// <typeparam name="T">The type of the requested property.</typeparam>
        /// <returns>The value of the requested property. If not found, the default
        /// value for the specified type is returned.</returns>
        protected internal void SetPropertyValue<T>(string name, T value)
        {
            Property<T> prop;

            if (value == null) BaseItem.Properties.Remove(name);

            if (!BaseItem.Properties.TryGetProperty(name, out prop))
            {
                prop = new Property<T>(name, value);
                BaseItem.Properties.Add(prop);
            }
            else
            {
                prop.Value = value;
            }
        }

        /// <summary>
        /// Sets the nullable property value.
        /// </summary>
        /// <typeparam name="T">The type of the property</typeparam>
        /// <param name="name">The name of the property.</param>
        /// <param name="value">The value of the property.</param>
        protected internal void SetNullablePropertyValue<T>(string name, T? value)
            where T : struct
        {
            Property<T> prop;

            if (value == null)
            {
                BaseItem.Properties.Remove(name);
                return;
            }

            if (!BaseItem.Properties.TryGetProperty(name, out prop))
            {
                prop = new Property<T>(name, value.Value);
                BaseItem.Properties.Add(prop);
            }
            else
            {
                prop.Value = value.Value;
            }
        }


        /// <summary>
        /// Ensures that the property with the specified name and type is present in the 
        /// and can be accessed.
        /// </summary>
        /// <param name="name">The name of the property.</param>
        protected void EnsureMultiValuedProperty<T>(string name)
        {
            if (Properties.Contains(name)) return;

            Properties.AddMultiValued<T>(name, null);
        }


        [MethodImpl(MethodImplOptions.Synchronized)]
        private void EnsureBodyIsLoaded()
        {
            if (_Body == null) _Body = new Body(this);
        }

        /// <summary>
        /// Returns wheter the specified flags are set in the given value.
        /// </summary>
        /// <param name="value">The value to check.</param>
        /// <param name="flags">The flags to check</param>
        /// <returns>Returns true, if the flags are set in the specified value. False otherwise. If 
        /// the specified value is null, null is returned.</returns>
        protected internal static bool? CheckFlagValue(int? value, int flags)
        {
            if (value == null) return null;

            return ((value.Value & flags) == flags);
        }

        [MethodImpl(MethodImplOptions.Synchronized)]
        private void EnsureMultivaluedPropertyExists<T>(string name)
        {
            if (BaseItem.Properties.Contains(name)) return;

            BaseItem.Properties.Add(new MultiValuedProperty<T>(name));
        }

        internal void LoadAddress(out EmailAddress address, string propertyName)
        {
            List<EmailAddress> addresses;

            addresses = ParseAddresses(GetPropertyValue<string>(propertyName));

            address = (addresses.Count > 0) ? addresses[0] : null;
        }

        private static List<EmailAddress> ParseAddresses(string addresses)
        {
            List<EmailAddress> result;
            string name;
            string address;

            result = new List<EmailAddress>();

            if (string.IsNullOrEmpty(addresses)) return result;

            foreach (Match match in _EmailAddressParser.Matches(addresses))
            {
                name = (match.Groups["name"].Success) ? match.Groups["name"].Value : "";
                address = (match.Groups["address"].Success) ? match.Groups["address"].Value : "";

                if (address == null) continue;

                result.Add(new EmailAddress(name, address));
            }
            return result;
        }

        internal void LoadAddressList(out ChangeNotifyCollection<EmailAddress> recipients, string propertyName)
        {
            recipients = new ChangeNotifyCollection<EmailAddress>();

            recipients.AddRange(ParseAddresses(GetPropertyValue<string>(propertyName)));
        }

        internal void CommitAddressList(ChangeNotifyCollection<EmailAddress> list, string property)
        {
            StringBuilder sb;

            if (list == null || !list.HasChanged) return;

            sb = new StringBuilder();
            foreach (EmailAddress address in list)
            {
                sb.Append(address.ToString());
                sb.Append(", ");
            }

            SetPropertyValue(property, sb.ToString());
        }

        /// <summary>
        /// Gets or sets the subject.
        /// </summary>
        /// <value>The subject.</value>
        public string Subject
        {
            get { return GetPropertyValue<string>(WellknownProperties.Item.Subject); }
            set { SetPropertyValue(WellknownProperties.Item.Subject, value); }
        }

        /// <summary>
        /// Gets the item class of the item (e.g. IPM.Note).
        /// </summary>
        /// <value>The subject.</value>
        public string ItemClass
        {
            get { return GetPropertyValue<string>(WellknownProperties.Item.ItemClass); }
            private set { SetPropertyValue(WellknownProperties.Item.ItemClass, value); }
        }

        /// <summary>
        /// Gets the content class of the item (e.g. IPM.Note).
        /// </summary>
        /// <value>The subject.</value>
        public string ContentClass
        {
            get { return GetPropertyValue<string>(WellknownProperties.Item.ContentClass); }
            private set { SetPropertyValue(WellknownProperties.Item.ContentClass, value); }
        }

        /// <summary>
        /// Gets or sets the sensitivity.
        /// </summary>
        /// <value>The sensitivity.</value>
        public Sensitivity? Sensitivity
        {
            get { return (Sensitivity?) GetNullablePropertyValue<int>(WellknownProperties.Item.Sensitivity); }
            set { SetNullablePropertyValue<int>(WellknownProperties.Item.Sensitivity, (int) value); }
        }

        /// <summary>
        /// Gets the body.
        /// </summary>
        /// <value>The body.</value>
        public Body Body
        {
            get
            {
                EnsureBodyIsLoaded();
                return _Body;
            }
        }

        /// <summary>
        /// Gets the received date and time.
        /// </summary>
        /// <value>The received.</value>
        public DateTime? Received
        {
            get { return GetNullablePropertyValue<DateTime>(WellknownProperties.Item.Received); }
        }

        /// <summary>
        /// Gets the size.
        /// </summary>
        /// <value>The size.</value>
        public int? Size
        {
            get { return GetNullablePropertyValue<int>(WellknownProperties.Item.MessageSize); }
        }

        /// <summary>
        /// Gets the categories.
        /// </summary>
        /// <value>The categories.</value>
        public IList<string> Categories
        {
            [MethodImpl(MethodImplOptions.Synchronized)]
            get
            {
                EnsureCategoryList();

                return _Categories;
            }
        }

        private void EnsureCategoryList()
        {
            if (_Categories != null) return;

            _Categories = GetMultivaluedPropertyValue<string>(WellknownProperties.Item.Categories);
            if (_Categories == null) _Categories = new List<string>();
        }

        /// <summary>
        /// Gets or sets the importance.
        /// </summary>
        /// <value>The importance.</value>
        public Importance? Importance
        {
            get { return (Importance?) GetNullablePropertyValue<int>(WellknownProperties.Item.Importance); }
            set { SetNullablePropertyValue<int>(WellknownProperties.Item.Importance, (int) value); }
        }

        /// <summary>
        /// Gets or sets the in reply to.
        /// </summary>
        /// <value>The in reply to.</value>
        public string InReplyTo
        {
            get { return GetPropertyValue<string>(WellknownProperties.Item.InReplyTo); }
            set { SetPropertyValue(WellknownProperties.Item.InReplyTo, value); }
        }


        /// <summary>
        /// Gets a value indicating whether this instance has been submitted.
        /// </summary>
        /// <value>
        /// 	<see langword="true"/> if this instance has been submitted; otherwise, <see langword="false"/>.
        /// </value>
        public bool? IsSubmitted
        {
            get
            {
                return
                    CheckFlagValue(GetNullablePropertyValue<int>(WellknownProperties.Item.MessageFlags),
                                   (int) MessageFlags.HasBeenSubmitted);
            }
        }


        /// <summary>
        /// Gets a value indicating whether this instance is a draft.
        /// </summary>
        /// <value>
        /// 	<see langword="true"/> if this instance is a draft; otherwise, <see langword="false"/>.
        /// </value>
        public bool? IsDraft
        {
            get
            {
                return
                    CheckFlagValue(GetNullablePropertyValue<int>(WellknownProperties.Item.MessageFlags),
                                   (int) MessageFlags.IsDraft);
            }
        }


        /// <summary>
        /// Returns true, if the item has been resent.
        /// </summary>
        /// <value>True, if the item has been resent. False, otherwise</value>
        public bool? IsResend
        {
            get
            {
                return
                    CheckFlagValue(GetNullablePropertyValue<int>(WellknownProperties.Item.MessageFlags),
                                   (int) MessageFlags.IsResend);
            }
        }

        /// <summary>
        /// Gets a value indicating whether this instance is unmodified.
        /// </summary>
        /// <value>
        /// 	<see langword="true"/> if this instance is unmodified; otherwise, <see langword="false"/>.
        /// </value>
        public bool? IsUnmodified
        {
            get
            {
                return
                    CheckFlagValue(GetNullablePropertyValue<int>(WellknownProperties.Item.MessageFlags),
                                   (int) MessageFlags.IsUnmodified);
            }
        }

        /// <summary>
        /// Gets the message headers.
        /// </summary>
        /// <value>The message headers.</value>
        public string MessageHeaders
        {
            get { return GetPropertyValue<string>(WellknownProperties.Item.MessageHeaders); }
        }

        /// <summary>
        /// Gets the sent date and time.
        /// </summary>
        /// <value>The sent date and time.</value>
        public DateTime? Sent
        {
            get { return GetNullablePropertyValue<DateTime>(WellknownProperties.Item.SentDate); }
        }

        /// <summary>
        /// Gets the created date and time.
        /// </summary>
        /// <value>The created date and time.</value>
        public DateTime? Created
        {
            get { return GetNullablePropertyValue<DateTime>(WellknownProperties.Item.CreationDate); }
        }

        /// <summary>
        /// Gets or sets the reminder due by.
        /// </summary>
        /// <value>The reminder due by.</value>
        public DateTime? ReminderDueBy
        {
            get { return GetNullablePropertyValue<DateTime>(WellknownProperties.Item.ReminderDueBy); }
            set { SetNullablePropertyValue(WellknownProperties.Item.ReminderDueBy, value); }
        }


        /// <summary>
        /// Gets a value indicating whether a reminder is set.
        /// </summary>
        /// <value>
        /// 	<see langword="true"/> if a reminder is set; otherwise, <see langword="false"/>.
        /// </value>
        public bool? ReminderIsSet
        {
            get { return GetNullablePropertyValue<bool>(WellknownProperties.Item.ReminderIsSet); }
        }

        /// <summary>
        /// Gets or sets the reminder minutes before start.
        /// </summary>
        /// <value>The reminder minutes before start.</value>
        public int? ReminderMinutesBeforeStart
        {
            get { return GetNullablePropertyValue<int>(WellknownProperties.Item.ReminderMinutesBeforeState); }
            set { SetNullablePropertyValue(WellknownProperties.Item.ReminderMinutesBeforeState, value); }
        }


        /// <summary>
        /// Gets a value indicating whether this instance has attachments.
        /// </summary>
        /// <value>
        /// 	<see langword="true"/> if this instance has attachments; otherwise, <see langword="false"/>.
        /// </value>
        public bool? HasAttachments
        {
            get { return GetNullablePropertyValue<bool>(WellknownProperties.Item.HasAttachments); }
        }


        /// <summary>
        /// Gets or sets the last verb executed.
        /// </summary>
        /// <value>The last verb executed.</value>
        public LastVerbExcecuted? LastVerbExecuted
        {
            get { return (LastVerbExcecuted?) GetNullablePropertyValue<int>(WellknownProperties.Item.LastVerbExecuted); }
            set { SetNullablePropertyValue(WellknownProperties.Item.LastVerbExecuted, (int?) value); }
        }

        /// <summary>
        /// Gets or sets the last verb execution time.
        /// </summary>
        /// <value>The last verb execution time.</value>
        public DateTime? LastVerbExecutionTime
        {
            get { return GetNullablePropertyValue<DateTime>(WellknownProperties.Item.LastVerbExecutedOn); }
            set { SetNullablePropertyValue(WellknownProperties.Item.LastVerbExecutedOn, value); }
        }

        /// <summary>
        /// Gets or sets the index of the icon.
        /// </summary>
        /// <value>The index of the icon.</value>
        public IconIndex? IconIndex
        {
            get { return (IconIndex?) GetNullablePropertyValue<int>(WellknownProperties.Item.IconIndex); }
            set { SetNullablePropertyValue(WellknownProperties.Item.IconIndex, (int?) value); }
        }
    }
}