using System;
using System.Collections.Generic;
using System.ComponentModel;
using InfiniTec.Exchange.Security;
using InfiniTec.Exchange.WebDav.Request;
using InfiniTec.Threading;
using AsyncOperation=InfiniTec.Threading.AsyncOperation;

namespace InfiniTec.Exchange.WebDav
{
    /// <summary>B
    /// This class simplifies the access to items stored in an Exchange 2000/2003 folder.
    /// </summary>
    public abstract class ItemBase : ObjectBase
    {
        private PropertyCollection _Properties;

        /// <summary>
        /// This event is called during an async refresh operation.
        /// </summary>
        public event ProgressChangedEventHandler<RequestStatus> RefreshProgressChanged;

        /// <summary>
        /// This event is called when an async refresh operation completes.
        /// </summary>
        public event AsyncCompletedEventHandler RefreshCompleted;

        /// <summary>
        /// This event is called during an async save operation.
        /// </summary>
        public event ProgressChangedEventHandler<RequestStatus> SaveProgressChanged;

        /// <summary>
        /// This event is called when an async save operation finishes.
        /// </summary>
        public event AsyncCompletedEventHandler SaveCompleted;

        /// <summary>
        /// This event is called during an async delete operation.
        /// </summary>
        public event ProgressChangedEventHandler<RequestStatus> DeleteProgressChanged;

        /// <summary>
        /// This event is called when an async delete operation is finished.
        /// </summary>
        public event AsyncCompletedEventHandler DeleteCompleted;

        /// <summary>
        /// This event is called during an async copy operation.
        /// </summary>
        public event ProgressChangedEventHandler<RequestStatus> CopyProgressChanged;

        /// <summary>
        /// This event is called after an async copy operation has completed.
        /// </summary>
        public event EventHandler<CopyCompletedEventArgs> CopyCompleted;

        /// <summary>
        /// This event is called during an async move operation.
        /// </summary>
        public event ProgressChangedEventHandler<RequestStatus> MoveProgressChanged;

        /// <summary>
        /// This event is called after an async move operation.
        /// </summary>
        public event AsyncCompletedEventHandler MoveCompleted;

        private List<string> _PropertiesToLoad;

        /// <summary>
        /// Gets list of properties which are loaded by the Refresh operation.
        /// </summary>
        /// <remarks>If the collection is empty, all properties are requested.</remarks>
        public IList<string> PropertiesToLoad
        {
            get { return _PropertiesToLoad; }
        }

        private DefaultProperties _DefaultPropertiesToLoad;

        /// <summary>
        /// Gets or sets the <see cref="DefaultProperties"/> to load.
        /// These properties are loaded regardless of the <see cref="PropertiesToLoad"/>
        /// contents.
        /// </summary>
        public DefaultProperties DefaultPropertiesToLoad
        {
            get { return _DefaultPropertiesToLoad; }
            set { _DefaultPropertiesToLoad = value; }
        }

        /// <summary>
        /// Returns the url of the item.
        /// </summary>
        /// <returns>A string containing the url of the item.</returns>
        public override string ToString()
        {
            return Url.ToString();
        }

        /// <summary>
        /// Gets the security descriptor property for this item.
        /// </summary>
        /// <returns>The security descriptor property for this item</returns>
        protected Property<byte[]> GetSecurityDescriptorProperty()
        {
            Property<byte[]> property;

            if (!Properties.TryGetProperty(WellknownProperties.Security.NTSecurityDescriptor, out property) &&
                !Properties.TryGetProperty(WellknownProperties.Security.PR_NT_Security_Descriptor, out property) &&
                !Properties.TryGetProperty(WellknownProperties.Security.PR_ACL_Table, out property))
                throw new SecurityInformationMissingException(
                    "The security information was not loaded during the last refresh operation! Ensure that Item.DefaultProperties is correctly set.");
            return property;
        }

        /// <summary>
        /// Replaces the current security descriptor of the item with the specified item.
        /// </summary>
        /// <param name="security"></param>
        protected void SetSecurityInternal(ExchangeSecurity security)
        {
            Property<byte[]> property;
            byte[] buffer;

            property = GetSecurityDescriptorProperty();

            buffer = new byte[8 + security.BinaryLength];

            Buffer.BlockCopy(property.Value, 0, buffer, 0, 8);
            security.GetBinaryForm(buffer, 8);
            property.Value = buffer;
        }

        /// <summary>
        /// Initializes a new instance of this class using the specified settings.
        /// </summary>
        /// <param name="path">The path of the item.</param>
        /// <param name="connection">The <see cref="Connection"/> to use.</param>
        /// <param name="propertiesToLoad">The properties to load.</param>
        protected ItemBase(string path, Connection connection, params string[] propertiesToLoad)
            :
                base(path, connection)
        {
            _Properties = new PropertyCollection();
            _PropertiesToLoad = new List<string>(propertiesToLoad);
        }

        /// <summary>
        /// Creates a new instance of this class
        /// </summary>
        /// <param name="url">The url of the item</param>
        /// <param name="connection">The <see cref="Connection"/> to use.</param>
        /// <param name="properties">The properties to load.</param>
        protected ItemBase(Uri url, Connection connection, IEnumerable<Property> properties)
            :
                base(url, connection)
        {
            _PropertiesToLoad = new List<string>();
            _Properties = new PropertyCollection();
            _Properties.AddRange(properties);
        }

        /// <summary>
        /// Returns the <see cref="InfiniTec.Threading.AsyncOperation"/> used to save the pending changes of this item.
        /// </summary>
        /// <returns>An <see cref="InfiniTec.Threading.AsyncOperation"/>.</returns>
        public virtual AsyncOperation<RequestStatus> GetSaveOperation()
        {
            return Methods.PropPatch(Url, Properties);
        }

        /// <summary>
        /// Synchronously saves all pending changes back to the Exchange server.
        /// </summary>
        public void Save()
        {
            using (AsyncOperation<RequestStatus> op = GetSaveOperation())
            {
                op.Run();
                op.WaitForCompletion();
            }
        }

        /// <summary>
        /// Asynchronously saves all pending changes back to the Exchange server.
        /// </summary>
        public void SaveAsync()
        {
            AsyncOperation<RequestStatus> operation;

            AsyncOperationContext.InitiatePendingAsyncOperation();

            operation = GetSaveOperation();

            AsyncOperationContext.RegisterPendingAsyncOperation(operation);

            operation.ProgressChanged += SaveProgressChanged;
            operation.Completed += SaveAsyncCompleted;
            operation.Error += SaveAsyncError;

            operation.Run();
        }

        private void SaveAsyncError(object sender, ExceptionEventArgs e)
        {
            AsyncOperationContext.UnregisterPendingAsyncOperation();
            OnSaveCompleted(new AsyncCompletedEventArgs(e.Exception, false, null));
        }

        private void SaveAsyncCompleted(object sender, CompletedEventArgs e)
        {
            AsyncOperationContext.UnregisterPendingAsyncOperation();

            switch (e.Status)
            {
                case OperationStatus.Completed:
                    OnSaveCompleted(new AsyncCompletedEventArgs(null, false, null));
                    break;
                case OperationStatus.Aborted:
                    OnSaveCompleted(new AsyncCompletedEventArgs(null, true, null));
                    break;
            }
        }

        /// <summary>
        /// This method is called when an async save operation completes.
        /// </summary>
        /// <param name="e">An <see cref="AsyncCompletedEventArgs"/> instance, indicating
        /// success or failure of the save operation.</param>
        protected virtual void OnSaveCompleted(AsyncCompletedEventArgs e)
        {
            if (SaveCompleted != null) SaveCompleted(this, e);
        }

        /// <summary>
        /// Returns an <see cref="InfiniTec.Threading.AsyncOperation"/> used to refresh this item.
        /// </summary>
        /// <returns>An <see cref="InfiniTec.Threading.AsyncOperation"/>.</returns>
        public virtual AsyncOperation<RequestStatus> GetRefreshOperation()
        {
            return new AsyncOperation<RequestStatus>(RefreshOperationInternal());
        }

        private IEnumerable<OperationAction> RefreshOperationInternal()
        {
            ProgressUpdate<RequestStatus> progress;
            List<string> propertiesToLoad;

            progress = new ProgressUpdate<RequestStatus>();

            propertiesToLoad = new List<string>(PropertiesToLoad);
            if (propertiesToLoad.Count > 0)
            {
                if ((DefaultPropertiesToLoad & DefaultProperties.FolderInformation) != 0)
                    propertiesToLoad.Add(WellknownProperties.Dav.IsFolder);
                if ((DefaultPropertiesToLoad & DefaultProperties.SecurityInformation) != 0)
                {
                    propertiesToLoad.Add(WellknownProperties.Security.NTSecurityDescriptor);
                    propertiesToLoad.Add(WellknownProperties.Security.PR_ACL_Table);
                    propertiesToLoad.Add(WellknownProperties.Security.PR_NT_Security_Descriptor);
                }
            }

            using (AsyncOperation<RequestStatus, PropertyCollection> operation =
                Methods.PropFind(Url, propertiesToLoad.ToArray()))
            {
                yield return new NestedOperation<RequestStatus>(operation, true,
                                                                progress);

                if (operation.Status != OperationStatus.Completed) yield break;

                _Properties = operation.Result;
            }
        }

        /// <summary>
        /// Returns an <see cref="InfiniTec.Threading.AsyncOperation"/> used to delete this item.
        /// </summary>
        /// <returns>An <see cref="InfiniTec.Threading.AsyncOperation"/>.</returns>
        public virtual AsyncOperation<RequestStatus> GetDeleteOperation()
        {
            return Methods.Delete(Url);
        }

        /// <summary>
        /// Synchronously deletes the current item from the Exchange folder.
        /// </summary>
        public void Delete()
        {
            using (AsyncOperation op = GetDeleteOperation())
            {
                op.Run();
                op.WaitForCompletion();
            }
        }

        /// <summary>
        /// Asynchronously deletes the item from the Exchange server.
        /// </summary>
        public void DeleteAsync()
        {
            AsyncOperation<RequestStatus> op;

            AsyncOperationContext.InitiatePendingAsyncOperation();

            op = GetDeleteOperation();

            AsyncOperationContext.RegisterPendingAsyncOperation(op);

            op.Completed += DeleteAsyncCompleted;
            op.Error += DeleteAsyncError;
            op.ProgressChanged += DeleteProgressChanged;

            op.Run();
        }

        private void DeleteAsyncError(object sender, ExceptionEventArgs e)
        {
            AsyncOperationContext.UnregisterPendingAsyncOperation();
            OnDeleteCompleted(new AsyncCompletedEventArgs(e.Exception, false, null));
        }

        private void DeleteAsyncCompleted(object sender, CompletedEventArgs e)
        {
            AsyncOperationContext.UnregisterPendingAsyncOperation();
            OnDeleteCompleted(new AsyncCompletedEventArgs(null, e.Status != OperationStatus.Completed, null));
        }

        /// <summary>
        /// Synchronously refreshes the properties of the curren item.
        /// </summary>
        public void Refresh()
        {
            using (AsyncOperation<RequestStatus> operation = GetRefreshOperation())
            {
                operation.Run();
                operation.WaitForCompletion();
            }
        }

        /// <summary>
        /// Asynchronously refreshes the properties of this instance.
        /// </summary>
        public void RefreshAsync()
        {
            AsyncOperation<RequestStatus> operation;

            AsyncOperationContext.InitiatePendingAsyncOperation();

            operation = GetRefreshOperation();

            AsyncOperationContext.RegisterPendingAsyncOperation(operation);
            operation.ProgressChanged += RefreshProgressChanged;
            operation.Completed += RefreshAsyncCompleted;
            operation.Error += RefreshAsyncError;

            operation.Run();
        }


        private void RefreshAsyncError(object sender, ExceptionEventArgs e)
        {
            AsyncOperationContext.UnregisterPendingAsyncOperation();

            OnRefreshCompleted(new AsyncCompletedEventArgs(e.Exception, false, null));
        }

        private void RefreshAsyncCompleted(object sender, CompletedEventArgs e)
        {
            AsyncOperationContext.UnregisterPendingAsyncOperation();

            OnRefreshCompleted(new AsyncCompletedEventArgs(null, e.Status == OperationStatus.Aborted, null));
        }

        /// <summary>
        /// This method is called after an async refresh operation has completed.
        /// </summary>
        /// <param name="e">An <see cref="AsyncCompletedEventArgs"/> instance indicating the success
        /// or failure of the refresh operation.</param>
        protected virtual void OnRefreshCompleted(AsyncCompletedEventArgs e)
        {
            if (RefreshCompleted != null)
            {
                RefreshCompleted(this, e);
            }
        }

        /// <summary>
        /// This method is called during an async delete operation.
        /// </summary>
        /// <param name="e">A ProgressChangedEventArgs info containing the current progress of the operation.</param>
        protected virtual void OnDeleteProgress(ProgressChangedEventArgs<RequestStatus> e)
        {
            if (DeleteProgressChanged != null) DeleteProgressChanged(this, e);
        }

        /// <summary>
        /// This method is called after an async delete operation has completed.
        /// </summary>
        /// <param name="e">An <see cref="AsyncCompletedEventArgs"/> instance indicating the success
        /// or failure of the delete operation.</param>
        protected virtual void OnDeleteCompleted(AsyncCompletedEventArgs e)
        {
            if (DeleteCompleted != null) DeleteCompleted(this, e);
        }

        /// <summary>
        /// Returns a <see cref="PropertyCollection"/> which allow the access and modification of the properties of this
        /// item.
        /// </summary>
        public PropertyCollection Properties
        {
            get { return _Properties; }
        }

        /// <summary>
        /// Returns an <see cref="InfiniTec.Threading.AsyncOperation"/> which is used to copy this item to the specified destination.
        /// </summary>
        /// <param name="path">The destination path.</param>
        /// <param name="option">A valid combination of the <see cref="CopyOption"/> enumeration.</param>
        /// <param name="scope">A valid member of the <see cref="Depth"/> enumeration.</param>
        /// <returns></returns>
        public AsyncOperation<RequestStatus, SimpleItem> GetCopyOperation(string path,
                                                                          CopyOption option, Depth scope)
        {
            return Methods.Copy(Url, Connection.CreateUrl(path), option, scope);
        }

        /// <summary>
        /// This method is called when an async copy operation has completed.
        /// </summary>
        /// <param name="e">An <see cref="CopyCompletedEventArgs"/> instance indicating the success
        /// or failure of the copy operation.</param>
        protected virtual void OnCopyCompleted(CopyCompletedEventArgs e)
        {
            if (CopyCompleted != null) CopyCompleted(this, e);
        }

        /// <summary>
        /// Synchronously copies the current item to the specified destination.
        /// </summary>
        /// <param name="path">The destination path.</param>
        /// <param name="option">A valid combination of the <see cref="CopyOption"/> enumeration.</param>
        /// <param name="scope">A valid member of the <see cref="Depth"/> enumeration.</param>
        /// <returns>The copied item.</returns>
        public SimpleItem Copy(string path, CopyOption option, Depth scope)
        {
            using (AsyncOperation<RequestStatus, SimpleItem> operation =
                GetCopyOperation(path, option, scope))
            {
                operation.Run();
                operation.WaitForCompletion();

                return operation.Result;
            }
        }

        /// <summary>
        /// Asynchronously copies the current item to the destination.
        /// </summary>
        /// <param name="path">The destination path.</param>
        /// <param name="option">A valid combination of the <see cref="CopyOption"/> enumeration.</param>
        /// <param name="scope">A valid member of the <see cref="Depth"/> enumeration.</param>
        public void CopyAsync(string path, CopyOption option, Depth scope)
        {
            AsyncOperation<RequestStatus, SimpleItem> operation;

            AsyncOperationContext.InitiatePendingAsyncOperation();

            operation = GetCopyOperation(path, option, scope);
            operation.ProgressChanged += CopyProgressChanged;
            operation.Error += CopyInternalError;
            operation.Completed += CopyInternalCompleted;

            operation.Run();
        }

        private void CopyInternalCompleted(object sender, CompletedEventArgs e)
        {
            AsyncOperation<RequestStatus, SimpleItem> operation;

            AsyncOperationContext.UnregisterPendingAsyncOperation();

            operation = (AsyncOperation<RequestStatus, SimpleItem>) sender;

            if (e.Status == OperationStatus.Completed)
                OnCopyCompleted(new CopyCompletedEventArgs(operation.Result));
            else
                OnCopyCompleted(new CopyCompletedEventArgs(true));
        }

        private void CopyInternalError(object sender, ExceptionEventArgs e)
        {
            AsyncOperationContext.UnregisterPendingAsyncOperation();

            OnCopyCompleted(new CopyCompletedEventArgs(e.Exception));
        }

        /// <summary>
        /// Returns an <see cref="InfiniTec.Threading.AsyncOperation"/> which is used to move the current item to the destination folder.
        /// </summary>
        /// <param name="path">The destination path.</param>
        /// <param name="option">A valid combination of the <see cref="CopyOption"/> enumeration.</param>
        /// <param name="scope">A valid member of the <see cref="Depth"/> enumeration.</param>
        /// <returns>An <see cref="InfiniTec.Threading.AsyncOperation"/></returns>
        public AsyncOperation<RequestStatus> GetMoveOperation(string path,
                                                              CopyOption option, Depth scope)
        {
            return Methods.Move(Url, Connection.CreateUrl(path), option, scope);
        }

        /// <summary>
        /// This method is called, when an async move operation has completed.
        /// </summary>
        /// <param name="e">An <see cref="AsyncCompletedEventArgs"/> indicating success or failure of the operation</param>
        protected virtual void OnMoveCompleted(AsyncCompletedEventArgs e)
        {
            if (MoveCompleted != null) MoveCompleted(this, e);
        }

        /// <summary>
        /// Synchronously moves this item to the specified destination.
        /// </summary>
        /// <param name="path">The destination path.</param>
        /// <param name="option">A valid combination of the <see cref="CopyOption"/> enumeration.</param>
        /// <param name="scope">A valid member of the <see cref="Depth"/> enumeration.</param>
        public void Move(string path, CopyOption option, Depth scope)
        {
            using (AsyncOperation<RequestStatus> operation =
                GetMoveOperation(path, option, scope))
            {
                operation.Run();
                operation.WaitForCompletion();
            }
        }

        /// <summary>
        /// Asynchronously moves this item to the specified destination.
        /// </summary>
        /// <param name="path">The destination path.</param>
        /// <param name="option">A valid combination of the <see cref="CopyOption"/> enumeration.</param>
        /// <param name="scope">A valid member of the <see cref="Depth"/> enumeration.</param>
        public void MoveAsync(string path, CopyOption option, Depth scope)
        {
            AsyncOperation<RequestStatus> operation;

            AsyncOperationContext.InitiatePendingAsyncOperation();

            operation = GetMoveOperation(path, option, scope);
            AsyncOperationContext.RegisterPendingAsyncOperation(operation);
            operation.ProgressChanged += MoveProgressChanged;
            operation.Error += MoveInternalError;
            operation.Completed += MoveInternalCompleted;

            operation.Run();
        }

        private void MoveInternalCompleted(object sender, CompletedEventArgs e)
        {
            AsyncOperationContext.UnregisterPendingAsyncOperation();

            if (e.Status == OperationStatus.Completed)
                OnMoveCompleted(new AsyncCompletedEventArgs(null, false, null));
            else
                OnMoveCompleted(new AsyncCompletedEventArgs(null, true, null));
        }

        private void MoveInternalError(object sender, ExceptionEventArgs e)
        {
            AsyncOperationContext.UnregisterPendingAsyncOperation();

            OnMoveCompleted(new AsyncCompletedEventArgs(e.Exception, false, null));
        }
    }
}