using System;

namespace InfiniTec.Exchange.WebDav
{
    /// <summary>
    /// Provides access to the wellknown folders, like Inbox, Calendar, Contacts, etc.
    /// </summary>
    public class Mailbox : Folder<Item>
    {
        private static string[] _ItemFields;

        /// <summary>
        /// Initializes the <see cref="Mailbox"/> class.
        /// </summary>
        static Mailbox()
        {
            _ItemFields = SchemaHelper.ExtractFieldValues(typeof (WellknownProperties.Mailbox));
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Mailbox"/> class.
        /// </summary>
        /// <param name="path">The path of the folder on the server.</param>
        /// <param name="connection">The <see cref="Connection"/> to use.</param>
        /// <param name="propertiesToLoad">The properties to retrieve. If omitted, all properties are retrieved.</param>
        public Mailbox(string path, Connection connection, params string[] propertiesToLoad) :
            base(path, connection, propertiesToLoad)
        {
            foreach (string property in _ItemFields)
            {
                PropertiesToLoad.Add(property);
            }
        }

        /// <summary>
        /// Returns a folder object representing the language-dependant default calendar folder of the current mailbox.
        /// </summary>
        /// <value>The default calendar of the current mailbox</value>
        public Folder<CalendarItem> Calendar
        {
            get { return OpenWellknownFolder<CalendarItem>(WellknownProperties.Mailbox.CalendarFolderUrl); }
        }

        /// <summary>
        /// Returns a folder object representing the language-dependant default contact folder of the current mailbox.
        /// </summary>
        /// <value>The default contact of the current mailbox</value>
        public Folder<ContactItem> Contacts
        {
            get { return OpenWellknownFolder<ContactItem>(WellknownProperties.Mailbox.ContactsUrl); }
        }

        /// <summary>
        /// Returns a folder object representing the language-dependant default deleted items folder of the current mailbox.
        /// </summary>
        /// <value>The default deleted items folder of the current mailbox</value>
        public Folder<Item> DeletedItems
        {
            get { return OpenWellknownFolder<Item>(WellknownProperties.Mailbox.DeletedItemsFolderUrl); }
        }

        /// <summary>
        /// Returns a folder object representing the language-dependant default drafts folder of the current mailbox.
        /// </summary>
        /// <value>The default drafts folder of the current mailbox</value>
        public Folder<MessageItem> Drafts
        {
            get { return OpenWellknownFolder<MessageItem>(WellknownProperties.Mailbox.DraftsFolderUrl); }
        }

        /// <summary>
        /// Returns a folder object representing the language-dependant default inbox folder of the current mailbox.
        /// </summary>
        /// <value>The default inbox folder of the current mailbox</value>
        public Folder<MessageItem> Inbox
        {
            get { return OpenWellknownFolder<MessageItem>(WellknownProperties.Mailbox.InboxFolderUrl); }
        }

        /// <summary>
        /// Returns a folder object representing the language-dependant default journal folder of the current mailbox.
        /// </summary>
        /// <value>The default journal folder of the current mailbox</value>
        public Folder<Item> Journal
        {
            get { return OpenWellknownFolder<Item>(WellknownProperties.Mailbox.JournalFolderUrl); }
        }

        /// <summary>
        /// Returns a folder object representing the language-dependant default junkmail folder of the current mailbox.
        /// </summary>
        /// <value>The default junkmail folder of the current mailbox</value>
        public Folder<MessageItem> JunkMail
        {
            get { return OpenWellknownFolder<MessageItem>(WellknownProperties.Mailbox.JunkMailFolderUrl); }
        }

        /// <summary>
        /// Returns a folder object representing the language-dependant default notes folder of the current mailbox.
        /// </summary>
        /// <value>The default notes folder of the current mailbox</value>
        public Folder<Item> Notes
        {
            get { return OpenWellknownFolder<Item>(WellknownProperties.Mailbox.NotesFolderUrl); }
        }

        /// <summary>
        /// Returns a folder object representing the language-dependant default outbox folder of the current mailbox.
        /// </summary>
        /// <value>The default outbox folder of the current mailbox</value>
        public Folder<MessageItem> Outbox
        {
            get { return OpenWellknownFolder<MessageItem>(WellknownProperties.Mailbox.OutboxFolderUrl); }
        }

        /// <summary>
        /// Returns a folder object representing the language-dependant default sent items folder of the current mailbox.
        /// </summary>
        /// <value>The default sent items folder of the current mailbox</value>
        public Folder<MessageItem> SentItems
        {
            get { return OpenWellknownFolder<MessageItem>(WellknownProperties.Mailbox.SentItemsFolderUrl); }
        }

        /// <summary>
        /// Returns a folder object representing the language-dependant default task folder of the current mailbox.
        /// </summary>
        /// <value>The default task folder of the current mailbox</value>
        public Folder<TaskItem> Tasks
        {
            get { return OpenWellknownFolder<TaskItem>(WellknownProperties.Mailbox.TasksFolderUrl); }
        }

        private Folder<ItemType> OpenWellknownFolder<ItemType>(string folderId)
            where ItemType : Item, new()
        {
            if (!Properties.Contains(folderId))
                throw new InvalidOperationException(
                    "The refresh method has to be called before opening a wellknown folder.");

            return new Folder<ItemType>(Properties.GetProperty<string>(folderId).Value, Connection);
        }
    }
}