using System;
using System.Collections.Generic;
using System.IO;
using System.Net;
using System.Text;
using System.Xml;
using System.Xml.XPath;
using InfiniTec.Exchange.WebDav.Request;
using InfiniTec.Threading;

namespace InfiniTec.Exchange.WebDav
{
    /// <summary>
    /// This class implements the various WebDav Methods like PropFind, PropPatch, Copy, Move, etc.
    /// </summary>
    public class Methods
    {
        private Connection _Connection;

        /// <summary>
        /// Returns the <see cref="Connection"/> to use for the requests.
        /// </summary>
        public Connection Connection
        {
            get { return _Connection; }
        }

        /// <summary>
        /// Initiates a new instance of this class using the specified connection.
        /// </summary>
        /// <param name="connection">The <see cref="Connection"/> to use.</param>
        public Methods(Connection connection)
        {
            _Connection = connection;
        }

        /// <summary>
        /// Returns an <see cref="AsyncOperation"/> implementing a PropFind request.
        /// </summary>
        /// <param name="url">The url to query</param>
        /// <param name="propertiesToLoad">The properties to load</param>
        /// <returns>An <see cref="AsyncOperation"/></returns>
        public AsyncOperation<RequestStatus, PropertyCollection> PropFind(Uri url, params string[] propertiesToLoad)
        {
            return new AsyncOperation<RequestStatus, PropertyCollection>(
                PropFindInternal(url, propertiesToLoad));
        }

        /// <summary>
        /// Returns an <see cref="AsyncOperation"/> implementing a PropPatch request.
        /// </summary>
        /// <param name="url">The url to query</param>
        /// <param name="properties">A <see cref="PropertyCollection"/> containing the properties to set on the specified url</param>
        /// <remarks>Only properties, with a HasChanged value of true are saved. Additionally, all 
        /// deleted properties are removed from the resource.</remarks>
        /// <returns>An <see cref="AsyncOperation"/></returns>
        public AsyncOperation<RequestStatus> PropPatch(Uri url, PropertyCollection properties)
        {
            return new AsyncOperation<RequestStatus>(PropPatchInternal(url, properties));
        }

        private IEnumerable<OperationAction> PropPatchInternal(Uri url, PropertyCollection properties)
        {
            RequestSettings info;

            using (Stream stream = PreparePropertyUpdate(properties))
            {
                info = new RequestSettings(url, "PROPPATCH");
                info.Request = stream;
                info.Headers.Add("Depth", "0");
                using (RequestOperation request = _Connection.CreateRequest(info))
                {
                    yield return new NestedOperation(request, true);
                }
            }
        }

        /// <summary>
        /// Returns an <see cref="AsyncOperation"/> implementing a Delete request.
        /// </summary>
        /// <param name="url">The url to query</param>
        /// <returns>An <see cref="AsyncOperation"/></returns>
        public AsyncOperation<RequestStatus> Delete(Uri url)
        {
            return new AsyncOperation<RequestStatus>(DeleteInternal(url));
        }

        private IEnumerable<OperationAction> DeleteInternal(Uri url)
        {
            RequestSettings settings;

            settings = new RequestSettings(url, "DELETE");
            using (RequestOperation request = _Connection.CreateRequest(settings))
            {
                yield return new NestedOperation(request, true);
            }
        }

        /// <summary>
        /// Returns an <see cref="AsyncOperation"/> implementing a MkCol request.
        /// </summary>
        /// <param name="url">The url of the folder to create</param>
        /// <param name="properties">The properties to set on the newly created folder.</param>
        /// <returns>An <see cref="AsyncOperation"/></returns>
        public AsyncOperation<RequestStatus> MkCol(Uri url, PropertyCollection properties)
        {
            return new AsyncOperation<RequestStatus>(MkColInternal(url, properties));
        }

        private IEnumerable<OperationAction> MkColInternal(Uri url, PropertyCollection properties)
        {
            RequestSettings info;
            Stream stream;

            if (properties.Count > 0)
            {
                stream = PreparePropertyUpdate(properties);
            }
            else stream = null;

            using (stream)
            {
                info = new RequestSettings(url, "MKCOL");
                info.Request = stream;
                info.Headers.Add("Depth", "0");

                using (RequestOperation request = _Connection.CreateRequest(info))
                {
                    yield return new NestedOperation(request, true);

                    // TODO: add exception handling
                }
            }
        }

        private static Stream PreparePropertyUpdate(PropertyCollection properties)
        {
            XmlWriterSettings settings;
            Stream stream;
            MemoryStream requestStream;
            string line;


            settings = new XmlWriterSettings();
            settings.Encoding = Encoding.UTF8;
            settings.Indent = true;

            stream = new MemoryStream();

            ProcessEmptyProperties(properties);

            using (XmlWriter writer = XmlWriter.Create(stream, settings))
            {
                writer.WriteProcessingInstruction("xml", "version=\"1.0\" encoding=\"UTF-8\"");

                writer.WriteStartElement("propertyupdate", Namespaces.Dav);
                writer.WriteAttributeString("xmlns", "dt", null, Namespaces.DataType);
                writer.WriteStartElement("set", Namespaces.Dav);
                writer.WriteStartElement("prop", Namespaces.Dav);

                foreach (Property p in properties)
                {
                    if (!p.HasChanged) continue;

                    p.SaveTo(writer);
                }

                writer.WriteEndElement();
                writer.WriteEndElement();

                if (properties.RemovedProperties.Count > 0)
                {
                    writer.WriteStartElement("remove", Namespaces.Dav);
                    writer.WriteStartElement("prop", Namespaces.Dav);

                    foreach (string entry in properties.RemovedProperties)
                    {
                        Property.WriteEncodedStartElementForPropertyName(writer, entry);
                        writer.WriteEndElement();
                    }
                    writer.WriteEndElement();

                    writer.WriteEndElement();
                }
                writer.WriteEndElement();

                writer.WriteEndDocument();
            }
            stream.Seek(0, SeekOrigin.Begin);

            requestStream = new MemoryStream();

            using (TextReader reader = new StreamReader(stream))
            using (StreamWriter writer = new StreamWriter(requestStream))
            {
                do
                {
                    line = reader.ReadLine();
                    if (line == null) break;

                    writer.WriteLine(line.Replace(Request.Request.NullPrefix, "0"));
                } while (true);

                writer.Flush();

                stream = new MemoryStream(requestStream.GetBuffer(), 0, (int) requestStream.Length, false);
            }

            stream.Seek(0, SeekOrigin.Begin);
            return stream;
        }

        private static void ProcessEmptyProperties(PropertyCollection properties)
        {
            List<Property> propertiesToRemove;

            propertiesToRemove = new List<Property>();

            foreach (Property property in properties)
            {
                if (property.IsPropertyValueNull() && property.HasChanged)
                {
                    propertiesToRemove.Add(property);
                }
            }

            foreach (Property property in propertiesToRemove)
            {
                properties.Remove(property);
            }
        }

        private IEnumerable<OperationAction> PropFindInternal(Uri url, string[] propertiesToLoad)
        {
            RequestSettings info;
            Response response;
            MemoryStream stream;
            XmlWriterSettings settings;
            XPathNavigator navi;
            XmlNamespaceManager nsmgr;
            PropertyCollection result;
            string ns;
            string localName;

            settings = new XmlWriterSettings();
            settings.Encoding = Encoding.GetEncoding("ISO-8859-1");

            stream = new MemoryStream();

            using (XmlWriter writer = XmlWriter.Create(stream, settings))
            {
                //writer.WriteProcessingInstruction("xml", "version=\"1.0\" encoding=\"ISO-8859-1\"");

                writer.WriteStartElement("propfind", Namespaces.Dav);

                if (propertiesToLoad == null || propertiesToLoad.Length == 0)
                {
                    writer.WriteStartElement("allprop", Namespaces.Dav);
                    writer.WriteElementString("allprop", Namespaces.Exchange, null);
                    writer.WriteEndElement();
                }
                else
                {
                    writer.WriteStartElement("prop", Namespaces.Dav);
                    foreach (string p in propertiesToLoad)
                    {
                        Property.ParsePropertyName(p, out ns, out localName);

                        if (char.IsDigit(localName[0]))
                        {
                            writer.WriteRaw(string.Format(@"<x:{0} xmlns:x=""{1}"" />", localName, ns));
                        }
                        else writer.WriteElementString(localName, ns, "");
                    }
                    writer.WriteEndElement();
                }
                writer.WriteEndElement();
                writer.WriteEndDocument();
            }
            stream.Seek(0, SeekOrigin.Begin);

            //ns = settings.Encoding.GetString(stream.GetBuffer(), 0, (int) stream.Length);
            //stream = new MemoryStream();

            //stream.Write(settings.Encoding.GetBytes(ns), 0, settings.Encoding.GetByteCount(ns));
            //stream.Seek(0, SeekOrigin.Begin);

            info = new RequestSettings(url, "PROPFIND");
            info.Request = stream;
            info.Headers.Add("Depth", "0");

            using (RequestOperation request = _Connection.CreateRequest(info))
            {
                yield return new NestedOperation(request, true);

                // TODO: add exception handling

                response = request.Result;
            }

            navi = response.Content.CreateNavigator();
            nsmgr = new XmlNamespaceManager(navi.NameTable);
            nsmgr.AddNamespace("dav", Namespaces.Dav);

            result = new PropertyCollection();

            foreach (XPathNavigator navigator in navi.Select("dav:multistatus/dav:response/dav:propstat", nsmgr))
            {
                result.AddRange(DecodeSingleResponse(navigator));
            }

            yield return new OperationResult<PropertyCollection>(result);
        }

        /// <summary>
        /// Returns an <see cref="AsyncOperation"/> implementing a Search request.
        /// </summary>
        /// <param name="url">The url to query</param>
        /// <param name="settings">The configuration for the search request</param>
        /// <returns>An <see cref="AsyncOperation"/></returns>
        public AsyncOperation<RequestStatus, IList<SimpleItem>> Search(Uri url, SearchSettings settings)
        {
            return new AsyncOperation<RequestStatus, IList<SimpleItem>>(SearchInternal(url, settings));
        }

        private IEnumerable<OperationAction> SearchInternal(Uri url, SearchSettings settings)
        {
            RequestSettings rs;
            Response response;
            XPathNavigator navi;
            XmlNamespaceManager nsmgr;
            List<SimpleItem> items;
            SimpleItem item;
            PropertyContainer container;

            using (Stream stream = PrepareSearchRequest(url, settings))
            {
                rs = new RequestSettings(url, "SEARCH");
                rs.Request = stream;
                rs.Headers.Add("Depth", "0");
                foreach (Range range in settings.Ranges)
                {
                    rs.Ranges.Add(range);
                }

                using (RequestOperation request = _Connection.CreateRequest(rs))
                {
                    yield return new NestedOperation(request, true);

                    response = request.Result;
                }

                navi = response.Content.CreateNavigator();
                nsmgr = new XmlNamespaceManager(navi.NameTable);
                nsmgr.AddNamespace("dav", Namespaces.Dav);

                items = new List<SimpleItem>();
                foreach (XPathNavigator entry in navi.Select("dav:multistatus/dav:response", nsmgr))
                {
                    container = ParseResponse(entry);
                    item = new SimpleItem(container.Url, Connection, container.Properties);
                    items.Add(item);
                }
                yield return new OperationResult<IList<SimpleItem>>(items.AsReadOnly());
            }
        }

        /// <summary>
        /// Parses a dav:propstat node returned from a PROPFIND or SEARCH operation and
        /// returns a <see cref="PropertyContainer"/> containing the url of the item
        /// as well as the properties of that item.
        /// </summary>
        /// <param name="node">An <see cref="XPathNavigator"/> positioned on a dav:response node.</param>
        /// <returns>A <see cref="PropertyContainer"/> instance.</returns>
        public PropertyContainer ParseResponse(XPathNavigator node)
        {
            PropertyContainer result;
            XmlNamespaceManager nsmgr;

            nsmgr = new XmlNamespaceManager(node.NameTable);
            nsmgr.AddNamespace("dav", Namespaces.Dav);

            result = new PropertyContainer();

            result.Url = new Uri(node.SelectSingleNode("dav:href", nsmgr).Value);
            result.Properties = DecodeSingleResponse(node.SelectSingleNode("dav:propstat", nsmgr));

            return result;
        }

        private static Stream PrepareSearchRequest(Uri url, SearchSettings settings)
        {
            MemoryStream stream;
            XmlWriterSettings writerSettings;
            StringBuilder sb;

            writerSettings = new XmlWriterSettings();
            writerSettings.Encoding = Encoding.UTF8;

            stream = new MemoryStream();

            using (XmlWriter writer = XmlWriter.Create(stream, writerSettings))
            {
                writer.WriteStartElement("searchrequest", Namespaces.Dav);

                sb = new StringBuilder();
                sb.AppendFormat("SELECT ");

                if (settings.PropertiesToLoad.Count == 0)
                {
                    sb.Append("*");
                }
                else
                {
                    foreach (string property in settings.PropertiesToLoad)
                    {
                        sb.AppendFormat("\"{0}\", ", property);
                    }
                    sb.Remove(sb.Length - 2, 2);
                }
                sb.Append(" FROM SCOPE (");

                sb.AppendFormat("'{0} TRAVERSAL OF \"{1}\"'", settings.Scope, url);
                foreach (Uri root in settings.AdditionalSearchRoots)
                {
                    sb.AppendFormat(", '{0} TRAVERSAL OF \"{1}\"'", settings.Scope, root.AbsolutePath);
                }

                sb.Append(") ");
                if (!string.IsNullOrEmpty(settings.Constraints))
                {
                    sb.Append("WHERE ");
                    sb.Append(settings.Constraints);
                    sb.Append(" ");
                }
                if (!string.IsNullOrEmpty(settings.OrderBy))
                {
                    sb.AppendFormat("ORDER BY {0}", settings.OrderBy);
                }

                writer.WriteElementString("sql", Namespaces.Dav, sb.ToString());

                writer.WriteEndElement();
                writer.WriteEndDocument();
            }
            stream.Seek(0, SeekOrigin.Begin);
            return stream;
        }

        private static Property[] DecodeSingleResponse(XPathNavigator result)
        {
            XmlNamespaceManager nsmgr;
            XPathNodeIterator iterator;
            List<Property> properties;
            HttpStatus status;

            nsmgr = new XmlNamespaceManager(result.NameTable);
            nsmgr.AddNamespace("dav", Namespaces.Dav);

            status = new HttpStatus(result.SelectSingleNode("dav:status", nsmgr).Value);
            iterator = result.Select("dav:prop/*", nsmgr);

            properties = new List<Property>();

            foreach (XPathNavigator property in iterator)
            {
                properties.Add(Property.Create(property, status));
            }

            return properties.ToArray();
        }

        /// <summary>
        /// Returns an <see cref="AsyncOperation"/> implementing a Copy request.
        /// </summary>
        /// <param name="source">The <see cref="Uri"/> of the source element.</param>
        /// <param name="destination">The <see cref="Uri"/> of the destination.</param>
        /// <param name="option">A valid combination of the <see cref="CopyOption"/> enumeration.</param>
        /// <param name="scope">A valid member of the <see cref="Depth"/> enumeration.</param>
        /// <returns>An <see cref="AsyncOperation"/></returns>
        public AsyncOperation<RequestStatus, SimpleItem> Copy(Uri source,
                                                              Uri destination, CopyOption option, Depth scope)
        {
            return new AsyncOperation<RequestStatus, SimpleItem>(CopyInternal(source,
                                                                              destination, option, scope));
        }

        private IEnumerable<OperationAction> CopyInternal(Uri source, Uri destination,
                                                          CopyOption option, Depth scope)
        {
            RequestSettings settings;

            settings = CreateCopyOrMoveRequest("COPY", source, destination, option, scope);

            using (AsyncOperation<RequestStatus, Response> operation =
                Connection.CreateRequest(settings))
            {
                yield return new NestedOperation(operation, true);
            }

            yield return new OperationResult<SimpleItem>(
                new SimpleItem(destination, Connection, new PropertyCollection()));
        }

        private static string GetDepthHeader(Depth scope)
        {
            switch (scope)
            {
                case Depth.Deep:
                    return "infinity";
                case Depth.OneLevel:
                    return "1";
                case Depth.OneLevelNoRoot:
                    return "1, noroot";
                case Depth.DeepNoRoot:
                    return "infinity, noroot";
                default:
                    return "0";
            }
        }

        /// <summary>
        /// Returns an <see cref="AsyncOperation"/> implementing a Move request.
        /// </summary>
        /// <param name="source">The <see cref="Uri"/> of the source element.</param>
        /// <param name="destination">The <see cref="Uri"/> of the destination.</param>
        /// <param name="option">A valid combination of the <see cref="CopyOption"/> enumeration.</param>
        /// <param name="scope">A valid member of the <see cref="Depth"/> enumeration.</param>
        /// <returns>An <see cref="AsyncOperation"/></returns>
        public AsyncOperation<RequestStatus> Move(Uri source,
                                                  Uri destination, CopyOption option, Depth scope)
        {
            return new AsyncOperation<RequestStatus>(MoveInternal(source,
                                                                  destination, option, scope));
        }

        private IEnumerable<OperationAction> MoveInternal(Uri source, Uri destination,
                                                          CopyOption option, Depth scope)
        {
            RequestSettings settings;

            settings = CreateCopyOrMoveRequest("MOVE", source, destination, option, scope);

            using (AsyncOperation<RequestStatus, Response> operation =
                Connection.CreateRequest(settings))
            {
                yield return new NestedOperation(operation, true);
            }
        }

        private static RequestSettings CreateCopyOrMoveRequest(string operation,
                                                               Uri source, Uri destination, CopyOption options,
                                                               Depth scope)
        {
            RequestSettings settings;
            settings = new RequestSettings(source, operation);
            settings.Headers.Add("Destination", destination.ToString());
            settings.Headers.Add("Allow-Rename",
                                 ((options & CopyOption.AllowRename) == CopyOption.AllowRename) ? "T" : "F");
            settings.Headers.Add("Overwrite",
                                 ((options & CopyOption.Overwrite) == CopyOption.Overwrite) ? "T" : "F");
            settings.Headers.Add("Depth", GetDepthHeader(scope));
            return settings;
        }

        /// <summary>
        /// Issues a PUT operation against the specified url and uploads the content from the specified stream
        /// </summary>
        /// <param name="url">The target url of the request</param>
        /// <param name="stream">The stream containing the data to upload</param>
        /// <returns>An <see cref="AsyncOperation"/> which executes the request.</returns>
        public AsyncOperation<RequestStatus> Put(Uri url, Stream stream)
        {
            return new AsyncOperation<RequestStatus>(PutInternal(url, stream));
        }

        private IEnumerable<OperationAction> PutInternal(Uri url, Stream stream)
        {
            RequestSettings settings;

            settings = new RequestSettings(url, WebRequestMethods.Http.Put);

            settings.Request = stream;

            using (AsyncOperation<RequestStatus, Response> operation =
                Connection.CreateRequest(settings))
            {
                yield return new NestedOperation(operation, true);
            }
        }

        /// <summary>
        /// Issues an X-MS-ENUMATTS request against the specified url and returns an array of <see cref="PropertyContainer"/> objects
        /// containing information about the attachments of the item.
        /// </summary>
        /// <param name="url">The target url of the request.</param>
        /// <returns>An <see cref="AsyncOperation"/> which executes the request.</returns>
        public AsyncOperation<RequestStatus, PropertyContainer[]> X_MS_EnumAtts(Uri url)
        {
            return new AsyncOperation<RequestStatus, PropertyContainer[]>(X_MS_EnumAttsInternal(url));
        }

        private IEnumerable<OperationAction> X_MS_EnumAttsInternal(Uri url)
        {
            RequestSettings settings;
            XPathNavigator response;
            XmlNamespaceManager nsmgr;
            List<PropertyContainer> result;

            settings = new RequestSettings(url, "X-MS-ENUMATTS");

            using (RequestOperation operation = Connection.CreateRequest(settings))
            {
                yield return new NestedOperation(operation, true);

                response = operation.Result.Content.CreateNavigator();
            }

            nsmgr = new XmlNamespaceManager(response.NameTable);
            nsmgr.AddNamespace("dav", Namespaces.Dav);

            result = new List<PropertyContainer>();

            foreach (XPathNavigator navigator in response.Select("dav:multistatus/dav:response", nsmgr))
            {
                result.Add(ParseResponse(navigator));
            }

            yield return new OperationResult<PropertyContainer[]>(result.ToArray());
        }

        /// <summary>
        /// Issues a GET operation against the specified url and writes the result of the request into the output stream.
        /// </summary>
        /// <param name="url">The target url of the request</param>
        /// <param name="outputStream">The target stream which receives the response of the request</param>
        /// <returns>An <see cref="AsyncOperation"/> which executes the request.</returns>
        public AsyncOperation<RequestStatus> Get(Uri url, Stream outputStream)
        {
            return new AsyncOperation<RequestStatus>(GetInternal(url, outputStream));
        }

        private IEnumerable<OperationAction> GetInternal(Uri Url, Stream outputStream)
        {
            RequestSettings settings;

            settings = new RequestSettings(Url, WebRequestMethods.Http.Get);
            settings.OutputStream = outputStream;

            using (RequestOperation operation = Connection.CreateRequest(settings))
            {
                yield return new NestedOperation(operation, true);
            }
        }
    }
}