using System;
using System.Diagnostics;
using InfiniTec.Threading;

namespace InfiniTec.Exchange.WebDav
{
    /// <summary>
    /// This class provided base functionality and properties for all WebDav related helper objects.
    /// </summary>
    public abstract class ObjectBase
    {
        private Connection _Connection;
        private Methods _Methods;
        private Uri _Url;
        private AsyncOperationContext _AsyncOperationContext;

        /// <summary>
        /// Gets the async operation support object, to prevent concurrent asynchronous operations on this instance.
        /// </summary>
        /// <value>The async operation support.</value>
        protected AsyncOperationContext AsyncOperationContext
        {
            [DebuggerStepThrough]
            get { return _AsyncOperationContext; }
        }

        /// <summary>
        /// Returns the <see cref="Uri"/> of the current instance.
        /// </summary>
        public Uri Url
        {
            get { return _Url; }
        }

        /// <summary>
        /// Returns an instance of a <see cref="Methods"/> class which can be used for WebDav requests.
        /// </summary>
        protected Methods Methods
        {
            get { return _Methods; }
        }

        /// <summary>
        /// Returns the connection associated with this instance.
        /// </summary>
        public Connection Connection
        {
            get { return _Connection; }
        }

        /// <summary>
        /// Initiates a new instance of this class using the specified settings.
        /// </summary>
        /// <param name="path">The path of the item on the Exchange folder.</param>
        /// <param name="connection">The connection to use</param>
        protected ObjectBase(string path, Connection connection) : this(
            connection.CreateUrl(path), connection)
        {
        }

        /// <summary>
        /// Initiates a new instance of this class using the specified settings.
        /// </summary>
        /// <param name="url">The url of the item on the Exchange folder.</param>
        /// <param name="connection">The connection to use</param>
        protected ObjectBase(Uri url, Connection connection)
        {
            if (connection == null) throw new ArgumentNullException("connection");

            if (url == null) throw new ArgumentNullException("url");

            if (!connection.IsValidUrl(url))
                throw new InvalidConnectionException("The specified url is not suitable for the given connection");

            _Connection = connection;
            _Url = url;
            _Methods = new Methods(connection);
            _AsyncOperationContext = new AsyncOperationContext();
        }
    }
}