using System;
using System.Collections.Generic;
using System.IO;
using System.Net;
using System.Text;
using System.Web;
using System.Xml.XPath;
using InfiniTec.Threading;

namespace InfiniTec.Exchange.WebDav.Request
{
    internal static class Request
    {
        public const string NullPrefix = "null__________________";

        public static RequestOperation Create(RequestSettings info, Connection connection)
        {
            return new RequestOperation(DoRequestAsync(info, connection));
        }

        private static AsyncOperation<RequestStatus, CookieCollection> DoFormBasedAuthentication(RequestSettings info,
                                                                                                 Connection connection)
        {
            return new AsyncOperation<RequestStatus, CookieCollection>(DoFormbasedAuthenticationAsync(info, connection));
        }

        private static IEnumerable<OperationAction> DoFormbasedAuthenticationAsync(RequestSettings info,
                                                                                   Connection connection)
        {
            HttpWebRequest request;
            byte[] body;
            IAsyncResult ar;
            ProgressUpdate<RequestStatus> action;
            Uri authUrl;

            authUrl = connection.CreateUrl("/exchweb/bin/auth/owaauth.dll");

            request = GetRequestObject(new RequestSettings(authUrl, "POST"), connection);

            request.CookieContainer = new CookieContainer();
            request.ContentType = "application/x-www-form-urlencoded";
            request.UserAgent = "Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.2)";
            request.AllowAutoRedirect = false;

            body =
                Encoding.UTF8.GetBytes(
                    string.Format(
                        "destination={0}&flags=0&username={1}&password={2}&SubmitCreds=Log+On&forcedownlevel=0&trusted=0",
                        HttpUtility.UrlEncode(info.Url.ToString()),
                        HttpUtility.UrlEncode(connection.Credential.Username),
                        HttpUtility.UrlEncode(connection.Credential.Password)));

            request.ContentLength = body.Length;

            ar = request.BeginGetRequestStream(null, null);
            action =
                new ProgressUpdate<RequestStatus>(RequestStatus.PerformingFormbasedAuthentication, ar.AsyncWaitHandle);
            yield return action;

            using (Stream stream = request.EndGetRequestStream(ar))
            {
                ar = stream.BeginWrite(body, 0, body.Length, null, null);
                action.ProgressPercentage = 50;
                yield return action;

                stream.EndWrite(ar);
            }

            ar = request.BeginGetResponse(null, null);
            action.ProgressPercentage = 75;
            action.AsyncWaitHandle = ar.AsyncWaitHandle;
            yield return action;

            using (HttpWebResponse response = (HttpWebResponse) request.EndGetResponse(ar))
            {
                if (response.Cookies.Count < 2)
                    throw new UnauthorizedAccessException("Could not login to Exchange server");

                yield return new OperationResult<CookieCollection>(response.Cookies);
            }
        }

        internal static HttpWebRequest GetRequestObject(RequestSettings info, Connection connection)
        {
            HttpWebRequest webrequest;
            string[] credentials;
            CredentialCache cache;

            // Initiate a new WebRequest to the given URI.

            webrequest = (HttpWebRequest) WebRequest.Create(info.Url);

            webrequest.CookieContainer = new CookieContainer();
            if (connection.FBACache != null) webrequest.CookieContainer.Add(connection.FBACache);

            if (connection.Credential != null &&
                connection.Credential.AuthenticationType != AuthenticationType.FormBasedAuthentication)
            {
                cache = new CredentialCache();

                if (connection.Credential.Username != "")
                {
                    credentials = connection.Credential.Username.Split(new char[] {'\\'}, 2);
                    if (credentials.Length == 2)
                        cache.Add(info.Url, connection.Credential.AuthenticationType.ToString(),
                                  new NetworkCredential(credentials[1], connection.Credential.Password, credentials[0]));
                    else
                        cache.Add(info.Url, connection.Credential.AuthenticationType.ToString(),
                                  new NetworkCredential(credentials[0], connection.Credential.Password, ""));
                    webrequest.Credentials = cache;
                }
                else
                    webrequest.Credentials = CredentialCache.DefaultCredentials;
            }

            webrequest.Method = info.Method;
            webrequest.PreAuthenticate = true;
            webrequest.KeepAlive = false;
            webrequest.ServicePoint.UseNagleAlgorithm = false;

            return webrequest;
        }

        /// <summary>
        /// Does the request async.
        /// </summary>
        /// <param name="info">The info.</param>
        /// <param name="connection">The connection.</param>
        /// <returns></returns>
        private static IEnumerable<OperationAction> DoRequestAsync(RequestSettings info, Connection connection)
        {
            HttpWebRequest webrequest;
            IAsyncResult ar;
            ProgressUpdate<RequestStatus> action;
            Stream output;
            XPathDocument doc;
            bool finished;
            HttpWebResponse response;

            do
            {
                webrequest = null;
                try
                {
                    finished = true;
                    action = new ProgressUpdate<RequestStatus>();

                    if (connection.Credential.AuthenticationType == AuthenticationType.FormBasedAuthentication &&
                        connection.FBACache == null)
                    {
                        using (
                            AsyncOperation<RequestStatus, CookieCollection> ao =
                                DoFormBasedAuthentication(info, connection))
                        {
                            yield return new NestedOperation(ao, true);

                            connection.FBACache = ao.Result;
                        }
                    }

                    webrequest = GetRequestObject(info, connection);

                    // If headers should be send along, add them to the
                    // collection.
                    if (info.Headers.Count > 0)
                    {
                        webrequest.Headers = info.Headers;
                    }
                    webrequest.Headers.Add("Translate", "F");
                    // Set the length of the request.
                    foreach (Range range in info.Ranges)
                    {
                        webrequest.AddRange("rows", range.From, range.To);
                    }

                    if (info.Request != null)
                    {
                        // Send the Create to the server.

                        webrequest.ContentType = "text/xml";
                        webrequest.ContentLength = info.Request.Length;

                        ar = webrequest.BeginGetRequestStream(null, null);

                        action = new ProgressUpdate<RequestStatus>(RequestStatus.SendingRequest,
                                                                   ar.AsyncWaitHandle);
                        yield return action;

                        using (Stream stream = webrequest.EndGetRequestStream(ar))
                        using (AsyncOperation op = AsyncOperations.Copy(info.Request, stream))
                        {
                            yield return new NestedOperation(op, true);

                            stream.Flush();
                        }
                    }

                    ar = webrequest.BeginGetResponse(null, null);

                    yield return action.Update(RequestStatus.ReceivingResponse, 0,
                                               ar.AsyncWaitHandle);

                    try
                    {
                        response = (HttpWebResponse) webrequest.EndGetResponse(ar);
                    }
                    catch (WebException ex)
                    {
                        if (ex.Response == null || (int) ((HttpWebResponse) ex.Response).StatusCode != 440) throw;

                        if (connection.Credential.AuthenticationType != AuthenticationType.FormBasedAuthentication)
                        {
                            throw new FormbasedAuthenticationRequiredException(
                                "The requested url requires formbased authentication");
                        }
                        else
                        {
                            connection.FBACache = null;
                            finished = false;
                            continue;
                        }
                    }

                    using (response)
                    {
                        doc = null;

                        if (connection.Credential.AuthenticationType == AuthenticationType.FormBasedAuthentication &&
                            response.Cookies != null && response.Cookies.Count > 2)
                        {
                            connection.FBACache = response.Cookies;
                        }
                        if (response.ContentLength != 0)
                        {
                            if (info.OutputStream != null) output = info.OutputStream;
                            else output = new MemoryStream((int) Math.Max(response.ContentLength, 0));

                            using (Stream stream = response.GetResponseStream())
                            using (AsyncOperation op = AsyncOperations.Copy(stream, output))
                            {
                                yield return new NestedOperation(op, true);

                                stream.Flush();
                            }

                            output.Seek(0, SeekOrigin.Begin);

                            if (info.OutputStream == null)
                            {
                                doc = ParseResponse(output);
                            }
                        }

                        webrequest = null;
                        yield return new OperationResult<Response>(new Response(doc, response.Headers));
                    }
                }
                finally
                {
                    if (webrequest != null) webrequest.Abort();
                }
            } while (!finished);
        }

        private static XPathDocument ParseResponse(Stream stream)
        {
            StringBuilder sb;
            StreamReader reader;

            sb = new StringBuilder();

            reader = new StreamReader(stream);

            sb.Append(reader.ReadToEnd());
            sb.Replace(":0x", ":" + NullPrefix + "_x");
            return new XPathDocument(new StringReader(sb.ToString()));
        }
    }
}