using System;
using System.Collections.Generic;
using System.Diagnostics;

namespace InfiniTec.Exchange.WebDav
{
    /// <summary>
    /// This class contains information about a search operation.
    /// </summary>
    public class SearchSettings
    {
        private SearchScope _Scope;
        private IList<Range> _Ranges;
        private string _Constraints;
        private string _OrderBy;
        private List<string> _PropertiesToLoad;
        private List<Uri> _AdditionalSearchRoots;

        /// <summary>
        /// This list contains additional search roots which should be searched along with
        /// the base url.
        /// </summary>
        public IList<Uri> AdditionalSearchRoots
        {
            [DebuggerStepThrough]
            get { return _AdditionalSearchRoots; }
        }


        /// <summary>
        /// Gets or sets the <see cref="SearchScope"/> of the search operation.
        /// </summary>
        public SearchScope Scope
        {
            get { return _Scope; }
            set { _Scope = value; }
        }

        /// <summary>
        /// Provides access to a collection of <see cref="Range"/> objects.
        /// </summary>
        public IList<Range> Ranges
        {
            get { return _Ranges; }
        }


        /// <summary>
        /// Gets or sets the constraints for the search operation.
        /// </summary>
        public string Constraints
        {
            get { return _Constraints; }
            set { _Constraints = value; }
        }

        /// <summary>
        /// Gets or sets the sort order for the search operation.
        /// </summary>
        public string OrderBy
        {
            get { return _OrderBy; }
            set { _OrderBy = value; }
        }

        /// <summary>
        /// Specifies which properties to load by the search operation.
        /// If this list is empty, all properties defined by the schema
        /// of the searchroot folder.
        /// </summary>
        public IList<string> PropertiesToLoad
        {
            get { return _PropertiesToLoad; }
        }

        /// <summary>
        /// Creates a new instance of this class.
        /// </summary>
        /// <param name="propertiesToLoad">The properties to load.</param>
        public SearchSettings(params string[] propertiesToLoad) :
            this((IEnumerable<string>) propertiesToLoad)
        {
        }

        /// <summary>
        /// Creates a new instance of this class.
        /// </summary>
        /// <param name="propertiesToLoad">The properties to load</param>
        public SearchSettings(IEnumerable<string> propertiesToLoad)
        {
            _AdditionalSearchRoots = new List<Uri>();
            _Ranges = new List<Range>();
            Scope = SearchScope.Shallow;
            _PropertiesToLoad = new List<string>(propertiesToLoad);
        }
    }
}