using System;
using System.Collections.Generic;
using System.Diagnostics;
using InfiniTec.Threading;

namespace InfiniTec.Exchange.WebDav
{
    /// <summary>
    /// This class performs a search of an Exchange public or private
    /// folder.
    /// </summary>
    public class Searcher : SearcherBase
    {
        /// <summary>
        /// This event is invoked when the progress of the search operation
        /// changed.
        /// </summary>
        public event EventHandler<SearchProgressChangedEventArgs> ProgressChanged;

        private List<SimpleItem> _SearchResult;

        /// <summary>
        /// Gets the search result.
        /// </summary>
        /// <value>The search result.</value>
        public IList<SimpleItem> SearchResult
        {
            [DebuggerStepThrough]
            get { return _SearchResult.AsReadOnly(); }
        }

        /// <summary>
        /// Synchronously starts the search operation.
        /// </summary>
        public override void Find()
        {
            _SearchResult = new List<SimpleItem>();
            base.Find();
        }

        /// <summary>
        /// Asynchronously starts the search operation.
        /// </summary>
        public override void FindAsync()
        {
            _SearchResult = new List<SimpleItem>();
            base.FindAsync();
        }

        /// <summary>
        /// Calls the <see cref="M:ProgressChanged"/> event.
        /// </summary>
        /// <param name="e"></param>
        protected virtual void OnProgressChanged(SearchProgressChangedEventArgs e)
        {
            if (ProgressChanged != null) ProgressChanged(this, e);
        }

        private void SearchProgressChanged(object sender, ProgressChangedEventArgs<SearchProgress> e)
        {
            SearchProgressChangedEventArgs args;

            args = new SearchProgressChangedEventArgs(e.Status.Range, e.Status.Items);
            OnProgressChanged(args);

            e.Cancel = args.Cancel;
        }

        /// <summary>
        /// Allows the implementors to register events on the async operation which is about to execute.
        /// </summary>
        /// <param name="operation">The operation.</param>
        protected override void RegisterEvents(AsyncOperation operation)
        {
            ((AsyncOperation<SearchProgress>) operation).ProgressChanged += SearchProgressChanged;
        }

        /// <summary>
        /// Gets the asynchronous operation which is used to search the configured search roots.
        /// </summary>
        /// <returns></returns>
        public virtual AsyncOperation<SearchProgress> GetSearchOperation()
        {
            return (AsyncOperation<SearchProgress>) GetInternalSearchOperation();
        }

        /// <summary>
        /// Processes the partial resultset from the last search operation
        /// </summary>
        /// <param name="items">The items of the last search operation</param>
        protected override void ProcessPartialResult(IList<SimpleItem> items)
        {
            _SearchResult.AddRange(items);
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Searcher"/> class.
        /// </summary>
        /// <param name="searchRoot">The root of the search operation.</param>
        /// <param name="connection">The <see cref="Connection"/> to use.</param>
        /// <param name="propertiesToLoad">The properties to load. This parameter is optional.</param>
        public Searcher(string searchRoot, Connection connection, params string[] propertiesToLoad)
            :
                base(searchRoot, connection, propertiesToLoad)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Searcher"/> class.
        /// </summary>
        /// <param name="searchRoot">The search root.</param>
        /// <param name="connection">The connection.</param>
        /// <param name="propertiesToLoad">The properties to load.</param>
        public Searcher(Uri searchRoot, Connection connection, params string[] propertiesToLoad)
            :
                base(searchRoot, connection, propertiesToLoad)
        {
        }
    }
}