using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using InfiniTec.Exchange.WebDav.Request;
using InfiniTec.Threading;
using AsyncOperation=InfiniTec.Threading.AsyncOperation;

namespace InfiniTec.Exchange.WebDav
{
    /// <summary>
    /// This class performs a search of an Exchange public or private
    /// folder.
    /// </summary>
    public abstract class SearcherBase : ObjectBase
    {
        /// <summary>
        /// This event is invoked when the the search operation has completed.
        /// </summary>
        public event AsyncCompletedEventHandler Completed;

        private int _PageSize;
        private SearchScope _SearchScope;
        private IList<string> _PropertiesToLoad;
        private string _SearchOrder;
        private string _Constraints;
        private SearchScope _Scope;
        private List<Uri> _AdditionalSearchRoots;

        /// <summary>
        /// This list contains additional search roots which should be searched along with
        /// the base url.
        /// </summary>
        public IList<Uri> AdditionalSearchRoots
        {
            [DebuggerStepThrough]
            get { return _AdditionalSearchRoots; }
        }

        /// <summary>
        /// Gets or sets the <see cref="SearchScope"/> of the search operation.
        /// </summary>
        public SearchScope Scope
        {
            get { return _Scope; }
            set { _Scope = value; }
        }

        /// <summary>
        /// Gets or sets the constraints of the operation.
        /// </summary>
        /// <remarks>
        /// The constraints is something like
        /// <code>"DAV:ishidden" = false</code>
        /// </remarks>
        public string Constraints
        {
            get { return _Constraints; }
            set { _Constraints = value; }
        }

        /// <summary>
        /// Gets or sets the sorting key of the search.
        /// </summary>
        /// <remarks>The search order can be something like
        /// <code>"DAV:displayname" ASC</code>
        /// </remarks>
        public string SearchOrder
        {
            get { return _SearchOrder; }
            set { _SearchOrder = value; }
        }

        /// <summary>
        /// Provides access to a collection containing the properties to search for.
        /// If this collection is empty, all properties are returned.
        /// </summary>
        public IList<string> PropertiesToLoad
        {
            get { return _PropertiesToLoad; }
        }

        /// <summary>
        /// Gets or sets the <see cref="SearchScope"/> of the search operation.
        /// </summary>
        public SearchScope SearchScope
        {
            get { return _SearchScope; }
            set { _SearchScope = value; }
        }

        /// <summary>
        /// Gets or sets the number of items returned by each iteration.
        /// </summary>
        public int PageSize
        {
            get { return _PageSize; }
            set { _PageSize = value; }
        }

        /// <summary>
        /// Calls the <see cref="M:ProgressCompleted"/> event.
        /// </summary>
        /// <param name="e"></param>
        protected virtual void OnCompleted(AsyncCompletedEventArgs e)
        {
            if (Completed != null) Completed(this, e);
        }

        /// <summary>
        /// Synchronously starts the search operation.
        /// </summary>
        /// <returns></returns>
        public virtual void Find()
        {
            using (AsyncOperation op = GetInternalSearchOperation())
            {
                op.Run();
                op.WaitForCompletion();
            }
        }

        /// <summary>
        /// Asynchronously starts the search operation.
        /// </summary>
        public virtual void FindAsync()
        {
            AsyncOperation operation;

            AsyncOperationContext.InitiatePendingAsyncOperation();

            operation = GetInternalSearchOperation();

            AsyncOperationContext.RegisterPendingAsyncOperation(operation);

            operation.Completed += SearchAsyncCompleted;
            operation.Error += SearchAsyncError;

            RegisterEvents(operation);

            operation.Run();
        }

        /// <summary>
        /// Allows the implementors to register events on the async operation which is about to execute.
        /// </summary>
        /// <param name="operation">The operation.</param>
        protected abstract void RegisterEvents(AsyncOperation operation);

        private void SearchAsyncError(object sender, ExceptionEventArgs e)
        {
            AsyncOperationContext.UnregisterPendingAsyncOperation();
            OnCompleted(new AsyncCompletedEventArgs(e.Exception, false, null));
        }

        private void SearchAsyncCompleted(object sender, CompletedEventArgs e)
        {
            AsyncOperation<SearchProgress> operation;

            AsyncOperationContext.UnregisterPendingAsyncOperation();

            operation = (AsyncOperation<SearchProgress>) sender;

            if (e.Status == OperationStatus.Completed)
                OnCompleted(new AsyncCompletedEventArgs(null, false, null));
            else
                OnCompleted(new AsyncCompletedEventArgs(null, true, null));
        }

        /// <summary>
        /// Returns the search operation used to perform the search based
        /// on the specified settings.
        /// </summary>
        /// <returns></returns>
        protected virtual AsyncOperation GetInternalSearchOperation()
        {
            return new AsyncOperation<SearchProgress>(SearchOperationInternal());
        }

        private IEnumerable<OperationAction> SearchOperationInternal()
        {
            SearchSettings settings;
            int itemsFound;
            int page;
            SearchProgress progress;
            List<SimpleItem> items;
            ProgressUpdate<SearchProgress> update;

            progress = new SearchProgress();
            update = new ProgressUpdate<SearchProgress>(progress);
            items = new List<SimpleItem>();

            settings = new SearchSettings(_PropertiesToLoad);
            settings.Constraints = Constraints;
            settings.OrderBy = SearchOrder;
            settings.Scope = Scope;
            settings.Ranges.Add(new Range(0, PageSize - 1));
            foreach (Uri url in AdditionalSearchRoots)
            {
                settings.AdditionalSearchRoots.Add(url);
            }
            page = 0;

            do
            {
                using (AsyncOperation<RequestStatus, IList<SimpleItem>> op = Methods.Search(Url, settings))
                {
                    yield return new NestedOperation(op, true);

                    itemsFound = op.Result.Count;

                    if (itemsFound > 0)
                    {
                        progress.Range = new Range(page*PageSize, page*PageSize + itemsFound - 1);
                    }
                    else
                    {
                        progress.Range = null;
                    }
                    progress.Items = op.Result;

                    ProcessPartialResult(op.Result);

                    yield return update;
                }

                page++;

                settings.Ranges.Clear();
                settings.Ranges.Add(new Range(page*PageSize, (page + 1)*PageSize - 1));
            } while (itemsFound == PageSize);
        }

        /// <summary>
        /// Processes the partial resultset from the last search operation
        /// </summary>
        /// <param name="items">The items of the last search operation</param>
        protected abstract void ProcessPartialResult(IList<SimpleItem> items);

        /// <summary>
        /// Creates a new instance of this class using the specified settings and
        /// a PageSize of 200.
        /// </summary>
        /// <param name="searchRoot">The root of the search operation.</param>
        /// <param name="connection">The <see cref="Connection"/> to use.</param>
        /// <param name="propertiesToLoad">The properties to load. This parameter is optional.</param>
        public SearcherBase(string searchRoot, Connection connection, params string[] propertiesToLoad)
            :
                base(searchRoot, connection)
        {
            Initialize(propertiesToLoad);
        }


        internal SearcherBase(Uri searchRoot, Connection connection, string[] propertiesToLoad)
            :
                base(searchRoot, connection)
        {
            Initialize(propertiesToLoad);
        }

        private void Initialize(string[] propertiesToLoad)
        {
            _PageSize = 200;
            _PropertiesToLoad = new List<string>(propertiesToLoad);
            _AdditionalSearchRoots = new List<Uri>();
        }
    }
}