using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.IO;
using InfiniTec.Exchange.Security;
using InfiniTec.Exchange.WebDav.Request;
using InfiniTec.Threading;

namespace InfiniTec.Exchange.WebDav
{
    /// <summary>
    /// This class simplifies the access to items stored in an Exchange 2000/2003 folder.
    /// </summary>
    public class SimpleItem : ItemBase
    {
        /// <summary>
        /// This event is executed when the status of the <see cref="GetContent"/> operation changes.
        /// </summary>
        public event ProgressChangedEventHandler<RequestStatus> GetContentProgressChanged;

        /// <summary>
        /// This event is executed when the <see cref="GetContent"/> operation is completed.
        /// </summary>
        public event AsyncCompletedEventHandler GetContentCompleted;

        /// <summary>
        /// This event is executed when the status of the <see cref="SetContent"/> operation changes.
        /// </summary>
        public event ProgressChangedEventHandler<RequestStatus> SetContentProgressChanged;

        /// <summary>
        /// This event is executed when the <see cref="SetContent"/> operation is completed.
        /// </summary>
        public event AsyncCompletedEventHandler SetContentCompleted;

        private AttachmentCollection _Attachments;

        /// <summary>
        /// Initializes a new instance of this class using the specified settings.
        /// </summary>
        /// <param name="path">The path of the item.</param>
        /// <param name="connection">The <see cref="Connection"/> to use.</param>
        /// <param name="propertiesToLoad">The properties to load.</param>
        public SimpleItem(string path, Connection connection, params string[] propertiesToLoad) :
            base(path, connection, propertiesToLoad)
        {
            Initialize();
        }

        private void Initialize()
        {
            _Attachments = new AttachmentCollection(Url, Connection);
        }

        internal SimpleItem(Uri url, Connection connection, IEnumerable<Property> propertiesToLoad) :
            base(url, connection, propertiesToLoad)
        {
            Initialize();
        }

        /// <summary>
        /// Returns an <see cref="AttachmentCollection"/> containing information about the attachments of the item.
        /// </summary>
        public AttachmentCollection Attachments
        {
            get { return _Attachments; }
        }

        /// <summary>
        /// Returns the security descriptor for the item
        /// </summary>
        /// <returns>An <see cref="ItemSecurity"/> instance.</returns>
        public ItemSecurity GetSecurity()
        {
            return new ItemSecurity(GetSecurityDescriptorProperty().Value, 8, Connection);
        }

        /// <summary>
        /// Replaces the current security descriptor with the specified descriptor.
        /// </summary>
        /// <param name="security">The new security descriptor</param>
        public void SetSecurity(ItemSecurity security)
        {
            SetSecurityInternal(security);
        }

        /// <summary>
        /// Returns an <see cref="InfiniTec.Threading.AsyncOperation"/> which can be used to GetContent the attachment.
        /// </summary>
        /// <param name="outputStream">The target stream for the content of the item.</param>
        /// <returns>An <see cref="InfiniTec.Threading.AsyncOperation"/>.</returns>
        public AsyncOperation<RequestStatus> GetGetContentOperation(Stream outputStream)
        {
            return Methods.Get(Url, outputStream);
        }


        /// <summary>
        /// GetContents the content of the attachment into the specified output stream. The current thread is blocked
        /// until the operation completes.
        /// </summary>
        /// <param name="outputStream">A <see cref="Stream"/> used to store the contents of the attachment.</param>
        /// <exception cref="AsyncOperationFailedException">Thrown, when the operation fails. See <see cref="Exception.InnerException"/> for details.</exception>
        public void GetContent(Stream outputStream)
        {
            using (AsyncOperation<RequestStatus> operation = GetGetContentOperation(outputStream))
            {
                operation.Run();

                operation.WaitForCompletion();
            }
        }

        /// <summary>
        /// Asynchronously GetContents the content of the attachment into the specified output stream.
        /// </summary>
        /// <param name="outputStream">A <see cref="Stream"/> used to store the contents of the attachment.</param>
        public void GetContentAsync(Stream outputStream)
        {
            AsyncOperation<RequestStatus> operation;

            AsyncOperationContext.InitiatePendingAsyncOperation();

            operation = GetGetContentOperation(outputStream);
            AsyncOperationContext.RegisterPendingAsyncOperation(operation);

            operation.ProgressChanged += GetContentProgressChanged;
            operation.Completed += GetContent_Completed;
            operation.Error += GetContent_Error;

            operation.Run();
        }

        private void GetContent_Error(object sender, ExceptionEventArgs e)
        {
            AsyncOperationContext.UnregisterPendingAsyncOperation();

            OnGetContentCompleted(new AsyncCompletedEventArgs(e.Exception, false, null));
        }

        private void GetContent_Completed(object sender, CompletedEventArgs e)
        {
            AsyncOperationContext.UnregisterPendingAsyncOperation();

            OnGetContentCompleted(new AsyncCompletedEventArgs(null, e.Status == OperationStatus.Aborted, null));
        }

        /// <summary>
        /// Executes the <see cref="GetContentCompleted"/> event.
        /// </summary>
        /// <param name="e">The event parameters.</param>
        protected virtual void OnGetContentCompleted(AsyncCompletedEventArgs e)
        {
            if (GetContentCompleted != null) GetContentCompleted(this, e);
        }


        /// <summary>
        /// Returns an <see cref="InfiniTec.Threading.AsyncOperation"/> which can be used to SetContent the attachment.
        /// </summary>
        /// <param name="stream">The target stream for the content of the item.</param>
        /// <returns>An <see cref="InfiniTec.Threading.AsyncOperation"/>.</returns>
        public AsyncOperation<RequestStatus> GetSetContentOperation(Stream stream)
        {
            return Methods.Put(Url, stream);
        }

        /// <summary>
        /// SetContents the content of the attachment into the specified output stream. The current thread is blocked
        /// until the operation completes.
        /// </summary>
        /// <param name="stream">A <see cref="Stream"/> used to store the contents of the attachment.</param>
        /// <exception cref="AsyncOperationFailedException">Thrown, when the operation fails. See <see cref="Exception.InnerException"/> for details.</exception>
        public void SetContent(Stream stream)
        {
            using (AsyncOperation<RequestStatus> operation = GetSetContentOperation(stream))
            {
                operation.Run();

                operation.WaitForCompletion();
            }
        }

        /// <summary>
        /// Asynchronously SetContents the content of the attachment into the specified output stream.
        /// </summary>
        /// <param name="stream">A <see cref="Stream"/> used to store the contents of the attachment.</param>
        public void SetContentAsync(Stream stream)
        {
            AsyncOperation<RequestStatus> operation;

            AsyncOperationContext.InitiatePendingAsyncOperation();

            operation = GetSetContentOperation(stream);
            AsyncOperationContext.RegisterPendingAsyncOperation(operation);

            operation.ProgressChanged += SetContentProgressChanged;
            operation.Completed += SetContent_Completed;
            operation.Error += SetContent_Error;

            operation.Run();
        }

        private void SetContent_Error(object sender, ExceptionEventArgs e)
        {
            AsyncOperationContext.UnregisterPendingAsyncOperation();

            OnSetContentCompleted(new AsyncCompletedEventArgs(e.Exception, false, null));
        }

        private void SetContent_Completed(object sender, CompletedEventArgs e)
        {
            AsyncOperationContext.UnregisterPendingAsyncOperation();

            OnSetContentCompleted(new AsyncCompletedEventArgs(null, e.Status == OperationStatus.Aborted, null));
        }

        /// <summary>
        /// Executes the <see cref="SetContentCompleted"/> event.
        /// </summary>
        /// <param name="e">The event parameters.</param>
        protected virtual void OnSetContentCompleted(AsyncCompletedEventArgs e)
        {
            if (SetContentCompleted != null) SetContentCompleted(this, e);
        }
    }
}